/*
   (c) Copyright 2001-2009  The world wide DirectFB Open Source Community (directfb.org)
   (c) Copyright 2000-2004  Convergence (integrated media) GmbH

   All rights reserved.

   Written by Denis Oliver Kropp <dok@directfb.org>,
              Andreas Hundt <andi@fischlustig.de>,
              Sven Neumann <neo@directfb.org>,
              Ville Syrjälä <syrjala@sci.fi> and
              Claudio Ciccani <klan@users.sf.net>.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

#include <directfb.h>

#include <display/idirectfbsurface.h>

#include <media/idirectfbimageprovider.h>

#include <core/coredefs.h>
#include <core/coretypes.h>

#include <core/layers.h>
#include <core/surface.h>

#include <misc/gfx_util.h>
#include <misc/util.h>
#include <direct/interface.h>
#include <direct/mem.h>
#include <direct/memcpy.h>
#include <direct/messages.h>

#include <setjmp.h>
#include <math.h>

#undef HAVE_STDLIB_H
#include <jpeglib.h>


static DFBResult
Probe( IDirectFBImageProvider_ProbeContext *ctx );

static DFBResult
Construct( IDirectFBImageProvider *thiz,
           ... );

#include <direct/interface_implementation.h>

DIRECT_INTERFACE_IMPLEMENTATION( IDirectFBImageProvider, JPEG )

D_DEBUG_DOMAIN( JPEG, "image/jpeg", "ImageProvider: JPEG");

/*
 * private data struct of IDirectFBImageProvider_JPEG
 */
typedef struct {
     int                  ref;      /* reference counter */

     IDirectFBDataBuffer *buffer;

     DIRenderCallback     render_callback;
     void                *render_callback_context;

     int                  width;    /*  width of the JPEG image   */
     int                  height;   /*  height of the JPEG image  */

     u32                 *image;        /*  decoded image data    */
     int                  image_width;  /*  width of image data   */
     int                  image_height; /*  height of image data  */

     CoreDFB             *core;

     CoreSurface         *decode_surface;
} IDirectFBImageProvider_JPEG_data;

static DirectResult
IDirectFBImageProvider_JPEG_AddRef  ( IDirectFBImageProvider *thiz );

static DirectResult
IDirectFBImageProvider_JPEG_Release ( IDirectFBImageProvider *thiz );

static DFBResult
IDirectFBImageProvider_JPEG_RenderTo( IDirectFBImageProvider *thiz,
                                      IDirectFBSurface       *destination,
                                      const DFBRectangle     *destination_rect );

static DFBResult
IDirectFBImageProvider_JPEG_SetRenderCallback( IDirectFBImageProvider *thiz,
                                               DIRenderCallback        callback,
                                               void                   *context );

static DFBResult
IDirectFBImageProvider_JPEG_GetSurfaceDescription( IDirectFBImageProvider *thiz,
                                                   DFBSurfaceDescription  *dsc);

static DFBResult
IDirectFBImageProvider_JPEG_GetImageDescription( IDirectFBImageProvider *thiz,
                                                 DFBImageDescription    *dsc );


#define JPEG_PROG_BUF_SIZE    0x10000

typedef struct {
     struct jpeg_source_mgr  pub; /* public fields */

     JOCTET                 *data;       /* start of buffer */

     IDirectFBDataBuffer    *buffer;

     int                     peekonly;
     int                     peekoffset;
} buffer_source_mgr;

typedef buffer_source_mgr * buffer_src_ptr;

static void
buffer_init_source (j_decompress_ptr cinfo)
{
     buffer_src_ptr src          = (buffer_src_ptr) cinfo->src;
     IDirectFBDataBuffer *buffer = src->buffer;

     buffer->SeekTo( buffer, 0 ); /* ignore return value */
}

static boolean
buffer_fill_input_buffer (j_decompress_ptr cinfo)
{
     DFBResult            ret;
     unsigned int         nbytes = 0;
     buffer_src_ptr       src    = (buffer_src_ptr) cinfo->src;
     IDirectFBDataBuffer *buffer = src->buffer;

     buffer->WaitForDataWithTimeout( buffer, JPEG_PROG_BUF_SIZE, 1, 0 );

     if (src->peekonly) {
          ret = buffer->PeekData( buffer, JPEG_PROG_BUF_SIZE,
                                  src->peekoffset, src->data, &nbytes );
          src->peekoffset += MAX( nbytes, 0 );
     }
     else {
          ret = buffer->GetData( buffer, JPEG_PROG_BUF_SIZE, src->data, &nbytes );
     }

     if (ret || nbytes <= 0) {
          /* Insert a fake EOI marker */
          src->data[0] = (JOCTET) 0xFF;
          src->data[1] = (JOCTET) JPEG_EOI;
          nbytes = 2;

          if (ret && ret != DFB_EOF)
               DirectFBError( "(DirectFB/ImageProvider_JPEG) GetData failed", ret );
     }

     src->pub.next_input_byte = src->data;
     src->pub.bytes_in_buffer = nbytes;

     return TRUE;
}

static void
buffer_skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
     buffer_src_ptr src = (buffer_src_ptr) cinfo->src;

     if (num_bytes > 0) {
          while (num_bytes > (long) src->pub.bytes_in_buffer) {
               num_bytes -= (long) src->pub.bytes_in_buffer;
               (void)buffer_fill_input_buffer(cinfo);
          }
          src->pub.next_input_byte += (size_t) num_bytes;
          src->pub.bytes_in_buffer -= (size_t) num_bytes;
     }
}

static void
buffer_term_source (j_decompress_ptr cinfo)
{
}

static void
jpeg_buffer_src (j_decompress_ptr cinfo, IDirectFBDataBuffer *buffer, int peekonly)
{
     buffer_src_ptr src;

     cinfo->src = (struct jpeg_source_mgr *)
                  cinfo->mem->alloc_small ((j_common_ptr) cinfo, JPOOL_PERMANENT,
                                           sizeof (buffer_source_mgr));

     src = (buffer_src_ptr) cinfo->src;

     src->data = (JOCTET *)
                  cinfo->mem->alloc_small ((j_common_ptr) cinfo, JPOOL_PERMANENT,
                                           JPEG_PROG_BUF_SIZE * sizeof (JOCTET));

     src->buffer = buffer;
     src->peekonly = peekonly;
     src->peekoffset = 0;

     src->pub.init_source       = buffer_init_source;
     src->pub.fill_input_buffer = buffer_fill_input_buffer;
     src->pub.skip_input_data   = buffer_skip_input_data;
     src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
     src->pub.term_source       = buffer_term_source;
     src->pub.bytes_in_buffer   = 0; /* forces fill_input_buffer on first read */
     src->pub.next_input_byte   = NULL; /* until buffer loaded */
}

struct my_error_mgr {
     struct jpeg_error_mgr pub;     /* "public" fields */
     jmp_buf  setjmp_buffer;          /* for return to caller */
};

static void
jpeglib_panic(j_common_ptr cinfo)
{
     struct my_error_mgr *myerr = (struct my_error_mgr*) cinfo->err;
     longjmp(myerr->setjmp_buffer, 1);
}

static inline void
copy_line32( u32 *argb, const u8 *rgb, int width )
{
     while (width--) {
          *argb++ = 0xFF000000 | (rgb[0] << 16) | (rgb[1] << 8) | rgb[2];

          rgb += 3;
     }
}

static inline void
copy_line_nv16( u16 *yy, u16 *cbcr, const u8 *src_ycbcr, int width )
{
     int x;

     for (x=0; x<width/2; x++) {
#ifdef WORDS_BIGENDIAN
          yy[x] = (src_ycbcr[0] << 8) | src_ycbcr[3];

          cbcr[x] = (((src_ycbcr[1] + src_ycbcr[4]) << 7) & 0xff00) |
                     ((src_ycbcr[2] + src_ycbcr[5]) >> 1);
#else
          yy[x] = (src_ycbcr[3] << 8) | src_ycbcr[0];

          cbcr[x] = (((src_ycbcr[2] + src_ycbcr[5]) << 7) & 0xff00) |
                     ((src_ycbcr[1] + src_ycbcr[4]) >> 1);
#endif

          src_ycbcr += 6;
     }

     if (width & 1) {
          u8 *y = (u8*) yy;

          y[width-1] = src_ycbcr[0];

#ifdef WORDS_BIGENDIAN
          cbcr[x] = (src_ycbcr[1] << 8) | src_ycbcr[2];
#else
          cbcr[x] = (src_ycbcr[2] << 8) | src_ycbcr[1];
#endif
     }
}


static DFBResult
Probe( IDirectFBImageProvider_ProbeContext *ctx )
{
     if (ctx->header[0] == 0xff && ctx->header[1] == 0xd8) {
          if (strncmp ((char*) ctx->header + 6, "JFIF", 4) == 0 ||
              strncmp ((char*) ctx->header + 6, "Exif", 4) == 0)
               return DFB_OK;

          if (ctx->filename && strchr (ctx->filename, '.' ) &&
             (strcasecmp ( strchr (ctx->filename, '.' ), ".jpg" ) == 0 ||
              strcasecmp ( strchr (ctx->filename, '.' ), ".jpeg") == 0))
               return DFB_OK;
     }

     return DFB_UNSUPPORTED;
}

static DFBResult
Construct( IDirectFBImageProvider *thiz,
           ... )
{
     struct jpeg_decompress_struct cinfo;
     struct my_error_mgr jerr;

     IDirectFBDataBuffer *buffer;
     CoreDFB             *core;
     va_list              tag;

     DIRECT_ALLOCATE_INTERFACE_DATA(thiz, IDirectFBImageProvider_JPEG)

     va_start( tag, thiz );
     buffer = va_arg( tag, IDirectFBDataBuffer * );
     core = va_arg( tag, CoreDFB * );
     va_end( tag );

     data->ref    = 1;
     data->buffer = buffer;
     data->core   = core;

     buffer->AddRef( buffer );

     cinfo.err = jpeg_std_error(&jerr.pub);
     jerr.pub.error_exit = jpeglib_panic;

     if (setjmp(jerr.setjmp_buffer)) {
          D_ERROR( "ImageProvider/JPEG: Error while reading headers!\n" );

          jpeg_destroy_decompress(&cinfo);
          buffer->Release( buffer );
          DIRECT_DEALLOCATE_INTERFACE( thiz );
          return DFB_FAILURE;
     }

     jpeg_create_decompress(&cinfo);
     jpeg_buffer_src(&cinfo, buffer, 1);
     jpeg_read_header(&cinfo, TRUE);
     jpeg_start_decompress(&cinfo);

     data->width = cinfo.output_width;
     data->height = cinfo.output_height;

     jpeg_abort_decompress(&cinfo);
     jpeg_destroy_decompress(&cinfo);

     thiz->AddRef = IDirectFBImageProvider_JPEG_AddRef;
     thiz->Release = IDirectFBImageProvider_JPEG_Release;
     thiz->RenderTo = IDirectFBImageProvider_JPEG_RenderTo;
     thiz->SetRenderCallback = IDirectFBImageProvider_JPEG_SetRenderCallback;
     thiz->GetImageDescription =IDirectFBImageProvider_JPEG_GetImageDescription;
     thiz->GetSurfaceDescription =
     IDirectFBImageProvider_JPEG_GetSurfaceDescription;

     return DFB_OK;
}

static void
IDirectFBImageProvider_JPEG_Destruct( IDirectFBImageProvider *thiz )
{
     IDirectFBImageProvider_JPEG_data *data =
                              (IDirectFBImageProvider_JPEG_data*)thiz->priv;

     data->buffer->Release( data->buffer );

     if (data->image)
          D_FREE( data->image );

     if (data->decode_surface)
          dfb_surface_unref( data->decode_surface );

     DIRECT_DEALLOCATE_INTERFACE( thiz );
}

static DirectResult
IDirectFBImageProvider_JPEG_AddRef( IDirectFBImageProvider *thiz )
{
     DIRECT_INTERFACE_GET_DATA(IDirectFBImageProvider_JPEG)

     data->ref++;

     return DFB_OK;
}

static DirectResult
IDirectFBImageProvider_JPEG_Release( IDirectFBImageProvider *thiz )
{
     DIRECT_INTERFACE_GET_DATA(IDirectFBImageProvider_JPEG)

     if (--data->ref == 0) {
          IDirectFBImageProvider_JPEG_Destruct( thiz );
     }

     return DFB_OK;
}

static void
JPEG_stretchblit (CardState    *state,
                  DFBRectangle *src_rect,
                  DFBRectangle *dst_rect)
{
  D_DEBUG_AT (JPEG, "StretchBlit %dx%d -> %dx%d (%s -> %s)\n",
              src_rect->w, src_rect->h, dst_rect->w, dst_rect->h,
              dfb_pixelformat_name (state->source->config.format),
              dfb_pixelformat_name (state->destination->config.format));

  /* thankfully this is intelligent enough to do a simple blit if possible */
  dfb_gfxcard_stretchblit (src_rect, dst_rect, state);
}

#define _ROUND_UP_8(x)  (((x)+ 7) & ~ 7)
#define _ROUND_UP_16(x) (((x)+15) & ~15)
static void
JPEG_setup_yuv444p( const struct jpeg_decompress_struct *cinfo,
                    CoreSurfaceConfig                   *config,
                    int                                 *pitch_y,
                    int                                 *pitch_cb,
                    int                                 *pitch_cr,
                    int                                 *offset_cb,
                    int                                 *offset_cr )
{
     config->format = DSPF_YUV444P;

     config->size.w = _ROUND_UP_8 (cinfo->cur_comp_info[0]
                                                      ->downsampled_width);
     config->size.h = _ROUND_UP_8 (cinfo->cur_comp_info[0]
                                                     ->downsampled_height);

     *pitch_y  = config->size.w;
     *pitch_cb = _ROUND_UP_8 (cinfo->cur_comp_info[1]->downsampled_width);
     *pitch_cr = _ROUND_UP_8 (cinfo->cur_comp_info[2]->downsampled_width);

     *offset_cb
          = *offset_cr
          = *pitch_y * config->size.h;
}

static DFBResult
JPEG_raw_decode( IDirectFBImageProvider_JPEG_data *data,
                 CoreSurface                      *dst_surface,
                 DFBRectangle                     *rect,
                 struct jpeg_decompress_struct    *cinfo,
                 CardState                        *state,
                 CoreSurfaceBufferLock            *lock )
{
     /* can we do a raw decode? */
     if (cinfo->num_components != 3
         || cinfo->jpeg_color_space != JCS_YCbCr
         || cinfo->data_precision != 8)
          return DFB_UNSUPPORTED;

     CoreSurfaceConfig config;
     int pitch_y, pitch_cb, pitch_cr;
     int offset_cb; /* cb from y */
     int offset_cr; /* cr from cb */
     int y_v_sampl = 1;
     int cb_v_sampl = 1;
     int cr_v_sampl = 1;

     const int v_samp[3] = { cinfo->cur_comp_info[0]->v_samp_factor,
                             cinfo->cur_comp_info[1]->v_samp_factor,
                             cinfo->cur_comp_info[2]->v_samp_factor };

     /* Worst case, is 4x4 chroma subsampling where an MCU is 32 lines. libjpeg
        won't decode more than one MCU in any one go. */
#undef _NO_STACK
#ifndef _NO_STACK
     JSAMPROW Yrows [cinfo->max_v_samp_factor * DCTSIZE];
     JSAMPROW Cbrows[cinfo->max_v_samp_factor * DCTSIZE];
     JSAMPROW Crrows[cinfo->max_v_samp_factor * DCTSIZE];
     JSAMPARRAY jpeg_buffer[3] = {
          [0] = Yrows,
          [1] = Cbrows,
          [2] = Crrows,
     };
#else
     /* valgrind can't check stack variables */
#warning this is not for production use, as we do not clean up on errors
     JSAMPROW *Yrows, *Cbrows, *Crrows;
     JSAMPARRAY *jpeg_buffer;

     /* over allocate, least trouble */
     Yrows = malloc (sizeof (JSAMPROW) * cinfo->max_v_samp_factor * DCTSIZE);
     Cbrows = malloc (sizeof (JSAMPROW) * cinfo->max_v_samp_factor * DCTSIZE);
     Crrows = malloc (sizeof (JSAMPROW) * cinfo->max_v_samp_factor * DCTSIZE);

     jpeg_buffer = malloc (3 * sizeof (JSAMPARRAY));
     jpeg_buffer[0] = Yrows;
     jpeg_buffer[1] = Cbrows;
     jpeg_buffer[2] = Crrows;
#endif
     void *yaddr, *cbaddr, *craddr;
     unsigned int i;

     DIRenderCallbackResult cb_result = DIRCR_OK;

     D_DEBUG_AT (JPEG, "trying raw decode\n");

     if (cinfo->max_h_samp_factor == 1
         && cinfo->max_v_samp_factor == 1
         && cinfo->cur_comp_info[0]->h_samp_factor == 1
         && v_samp[0] == 1
         && cinfo->cur_comp_info[1]->h_samp_factor == 1
         && v_samp[1] == 1
         && cinfo->cur_comp_info[2]->h_samp_factor == 1
         && v_samp[2] == 1) {
          /* The image is YCbCr 4:4:4 */
          D_DEBUG_AT (JPEG, "  -> 4:4:4 image\n");

          JPEG_setup_yuv444p (cinfo, &config, &pitch_y, &pitch_cb, &pitch_cr,
                              &offset_cb, &offset_cr);
     }
     else if (cinfo->max_h_samp_factor == 2
              && cinfo->max_v_samp_factor == 2
              && cinfo->cur_comp_info[0]->h_samp_factor == 2
              && v_samp[0] == 2
              && cinfo->cur_comp_info[1]->h_samp_factor == 1
              && v_samp[1] == 1
              && cinfo->cur_comp_info[2]->h_samp_factor == 1
              && v_samp[2] == 1) {
          /* The image is YCbCr 4:2:0 ... */
          D_DEBUG_AT (JPEG, "  -> 4:2:0 (I420) image\n");

          if ((cinfo->cur_comp_info[1]->downsampled_width
               == cinfo->cur_comp_info[0]->downsampled_width)
              && (cinfo->cur_comp_info[2]->downsampled_width
                  == cinfo->cur_comp_info[0]->downsampled_width)) {
               /* ... but because we are iDCT scaling during decompression it
                  might end up scaled to 4:4:4, as libjpeg can (and will) apply
                  different scaling factors for the three components. */
               D_DEBUG_AT (JPEG, "    -> iDCT %u/%u scaled to 4:4:4\n",
                           cinfo->scale_num, cinfo->scale_denom);

               JPEG_setup_yuv444p (cinfo, &config, &pitch_y, &pitch_cb, &pitch_cr,
                                   &offset_cb, &offset_cr);
          }
          else if ((cinfo->cur_comp_info[1]->downsampled_width
                    == ((cinfo->cur_comp_info[0]->downsampled_width + 1)
                        / 2))
                   && (cinfo->cur_comp_info[2]->downsampled_width
                       == ((cinfo->cur_comp_info[0]->downsampled_width + 1)
                           / 2))) {
               /* ... no iDCT scaling, which means it really ends up
                  as 4:2:0. */
               config.format = DSPF_I420;

               config.size.w = _ROUND_UP_16 (cinfo->cur_comp_info[0]
                                                           ->downsampled_width);
               config.size.h = _ROUND_UP_16 (cinfo->cur_comp_info[0]
                                                          ->downsampled_height);

               pitch_y = config.size.w;
               pitch_cb
                    = pitch_cr
                    = pitch_y / 2;

               offset_cb = pitch_y  * config.size.h;
               offset_cr = pitch_cb * config.size.h / 2;

               cb_v_sampl
                    = cr_v_sampl
                    = 2;
          }
          else {
               /* shouldn't happen, really */
               D_DEBUG_AT (JPEG, "  -> unsupported (unexpected)!\n");
               return DFB_UNSUPPORTED;
          }
     }
     else if (cinfo->max_h_samp_factor == 2
              && cinfo->max_v_samp_factor == 1
              && cinfo->cur_comp_info[0]->h_samp_factor == 2
              && v_samp[0] == 1
              && cinfo->cur_comp_info[1]->h_samp_factor == 1
              && v_samp[1] == 1
              && cinfo->cur_comp_info[2]->h_samp_factor == 1
              && v_samp[2] == 1) {
          /* The image is YCbCr 4:2:2 ... */
          D_DEBUG_AT (JPEG, "  -> 4:2:2 (YV16) image\n");

          if ((cinfo->cur_comp_info[1]->downsampled_width
               == cinfo->cur_comp_info[0]->downsampled_width)
              && (cinfo->cur_comp_info[2]->downsampled_width
                  == cinfo->cur_comp_info[0]->downsampled_width)) {
               /* ... iDCT scaling again, as above. */
               D_DEBUG_AT (JPEG, "    -> iDCT %u/%u scaled to 4:4:4\n",
                           cinfo->scale_num, cinfo->scale_denom);

               JPEG_setup_yuv444p (cinfo, &config, &pitch_y, &pitch_cb, &pitch_cr,
                                   &offset_cb, &offset_cr);
          }
          else if ((cinfo->cur_comp_info[1]->downsampled_width
                    == ((cinfo->cur_comp_info[0]->downsampled_width + 1)
                        / 2))
                   && (cinfo->cur_comp_info[2]->downsampled_width
                       == ((cinfo->cur_comp_info[0]->downsampled_width + 1)
                           / 2))) {
               /* ... no iDCT scaling, as above. */
               config.format = DSPF_YV16;

               config.size.w = _ROUND_UP_16 (cinfo->cur_comp_info[0]
                                                           ->downsampled_width);
               config.size.h = _ROUND_UP_16 (cinfo->cur_comp_info[0]
                                                          ->downsampled_height);

               pitch_y = config.size.w;
               pitch_cb
                    = pitch_cr
                    = pitch_y / 2;

               /* YV16 has Cr/Cb planes not Cb/Cr */
               offset_cb = (pitch_y + pitch_cr) * config.size.h;
               offset_cr = -pitch_cr * config.size.h;
          }
          else {
               /* shouldn't happen, really */
               D_DEBUG_AT (JPEG, "  -> unsupported (unexpected)!\n");
               return DFB_UNSUPPORTED;
          }
     }
     else {
          D_DEBUG_AT (JPEG, "  -> unsupported!\n");
          return DFB_UNSUPPORTED;
     }

     /* yes, we can handle this raw format! */

     config.flags = CSCONF_SIZE | CSCONF_FORMAT | CSCONF_CAPS;
     config.caps  = DSCAPS_VIDEOONLY;
     #ifdef DIRECT_BUILD_DEBUG
     config.caps |= DSCAPS_SHARED;
     #endif
     if (dfb_surface_create (data->core, &config, CSTF_NONE, 0, NULL,
                             &data->decode_surface)) {
          D_ERROR ("failed to create temporary decode surface\n");
          return DFB_UNSUPPORTED;
     }

     if (dfb_surface_lock_buffer (data->decode_surface,
                                  CSBR_BACK, CSAID_CPU, CSAF_WRITE, lock)) {
          dfb_surface_unref (data->decode_surface);
          data->decode_surface = NULL;
          return DFB_UNSUPPORTED;
     }

     cinfo->raw_data_out = true;
     jpeg_start_decompress (cinfo);

     dfb_state_set_source (state, data->decode_surface);

     /* Initialize the various pointers to build a planar YUV buffer. */
     yaddr  = lock->addr;
     cbaddr = yaddr  + offset_cb;
     craddr = cbaddr + offset_cr;

     /* init. */
     for (i = 0; i < cinfo->max_v_samp_factor * DCTSIZE; ++i) {
          Yrows[i] = yaddr + i * pitch_y;
          Cbrows[i] = cbaddr + i * pitch_cb;
          Crrows[i] = craddr + i * pitch_cr;
     }

     while (cinfo->output_scanline < data->image_height
            && cb_result == DIRCR_OK) {
          int x = jpeg_read_raw_data (cinfo, jpeg_buffer,
                                      cinfo->max_v_samp_factor * DCTSIZE);
          if (x <= 0)
               /* Actually, x == 0 means that we don't have enough data to
                  continue decoding the picture. */
               break;

          D_DEBUG_AT (JPEG, "  -> decoded %d scanlines (out of %d (%d))\n",
                      cinfo->output_scanline, data->image_height,
                      config.size.h);

          /* increment the pointers by the number of decoded (luma)
             scanlines. */
          for (i = 0; i < cinfo->max_v_samp_factor * DCTSIZE; ++i) {
               Yrows[i] += (x * pitch_y) / y_v_sampl;
               Cbrows[i] += (x * pitch_cb) / cb_v_sampl;
               Crrows[i] += (x * pitch_cr) / cr_v_sampl;
          }

          if (data->render_callback) {
               DFBRectangle src_rect = {
                    .x = 0,
                    .y = cinfo->output_scanline - x,
                    .w = data->image_width,
                    .h = x,
               };
               DFBRectangle r = src_rect;
               float factor = (rect->h / (float) data->image_height);
               DFBRectangle dst_rect = {
                    .x = rect->x,
                    .y = (int) (src_rect.y * factor),
                    .w = rect->w,
                    .h = (int) (src_rect.h * factor),
               };

               D_DEBUG_AT (JPEG,
                           "  -> render callback %d,%d %dx%d -> %d,%d %dx%d\n",
                           src_rect.x, src_rect.y, src_rect.w,
                           src_rect.h, dst_rect.x, dst_rect.y,
                           dst_rect.w, dst_rect.h);

               JPEG_stretchblit (state, &src_rect, &dst_rect);

               cb_result = data->render_callback (&r,
                                                  data->render_callback_context);
          }
     }

     D_DEBUG_AT (JPEG, "  -> decoded %d scanlines (out of %d)\n",
                 cinfo->output_scanline, data->image_height);

     if (cinfo->output_scanline < data->image_height
         || cb_result != DIRCR_OK) {
          if (cb_result != DIRCR_OK)
               D_ERROR ("raw decode failed after %d of %d scanlines, "
                        "trying pure software\n",
                        cinfo->output_scanline, cinfo->output_height);
          jpeg_abort_decompress (cinfo);
          jpeg_destroy_decompress (cinfo);
          dfb_surface_unlock_buffer (data->decode_surface, lock);
          lock->pitch = 0;
          dfb_surface_unref (data->decode_surface);
          data->decode_surface = NULL;

          if (cb_result != DIRCR_OK)
               return DFB_INTERRUPTED;

          return DFB_FAILURE; /* restart */
     }

     jpeg_finish_decompress (cinfo);
     jpeg_destroy_decompress (cinfo);

     dfb_surface_unlock_buffer (data->decode_surface, lock);
     lock->pitch = 0;

     /* use DFB to convert raw YUV to destination format, and apply any
        necessary additional clip/stretch */
     {
     DFBRectangle src_rect = {
          .x = 0,
          .y = 0,
          .w = data->image_width,
          .h = data->image_height
     };

     JPEG_stretchblit (state, &src_rect, rect);

     if (data->render_callback)
          data->render_callback (&src_rect, data->render_callback_context);
     }

#ifdef _NO_STACK
     free (Yrows);
     free (Cbrows);
     free (Crrows);
     free (jpeg_buffer);
#endif
     return DFB_OK;
}

static DFBResult
IDirectFBImageProvider_JPEG_RenderTo( IDirectFBImageProvider *thiz,
                                      IDirectFBSurface       *destination,
                                      const DFBRectangle     *dest_rect )
{
     DFBResult              ret;
     bool                   direct = false;
     DFBRegion              clip;
     DFBRectangle           rect;
     DFBSurfacePixelFormat  format;
     IDirectFBSurface_data *dst_data;
     CoreSurface           *dst_surface;
     CoreSurfaceBufferLock  lock = { .pitch = 0 };
     DIRenderCallbackResult cb_result = DIRCR_OK;
     bool                   try_raw = true;

     DIRECT_INTERFACE_GET_DATA(IDirectFBImageProvider_JPEG)

     dst_data = (IDirectFBSurface_data*) destination->priv;
     if (!dst_data)
          return DFB_DEAD;

     dst_surface = dst_data->surface;
     if (!dst_surface)
          return DFB_DESTROYED;

     ret = destination->GetPixelFormat( destination, &format );
     if (ret)
          return ret;

     dfb_region_from_rectangle( &clip, &dst_data->area.current );

     if (dest_rect) {
          if (dest_rect->w < 1 || dest_rect->h < 1)
               return DFB_INVARG;

          rect = *dest_rect;
          rect.x += dst_data->area.wanted.x;
          rect.y += dst_data->area.wanted.y;

          if (!dfb_rectangle_region_intersects( &rect, &clip ))
               return DFB_OK;
     }
     else {
          rect = dst_data->area.wanted;
     }

     if (data->image &&
         (rect.x || rect.y || rect.w != data->image_width || rect.h != data->image_height)) {
           D_FREE( data->image );
           data->image        = NULL;
           data->image_width  = 0;
           data->image_height = 0;
     }

     /* actual loading and rendering */
     if (!data->image
         && !data->decode_surface) {
          struct jpeg_decompress_struct cinfo;
          struct my_error_mgr jerr;
          JSAMPARRAY buffer;      /* Output row buffer */
          int row_stride;         /* physical row width in output buffer */
          u32 *row_ptr;
          int y = 0;
          int uv_offset = 0;
          CardState state;
          CoreSurfaceBufferLock lock_raw = { .pitch = 0 };

          cinfo.err = jpeg_std_error(&jerr.pub);
          jerr.pub.error_exit = jpeglib_panic;

          if (setjmp(jerr.setjmp_buffer)) {
               D_ERROR( "ImageProvider/JPEG: Error during decoding!\n" );

               jpeg_destroy_decompress(&cinfo);

               if (try_raw) {
                    dfb_state_set_source (&state, NULL);
                    dfb_state_set_destination (&state, NULL);
                    dfb_state_destroy (&state);
               }
               if (data->decode_surface) {
                    if (lock_raw.pitch)
                         dfb_surface_unlock_buffer (data->decode_surface,
                                                    &lock_raw);
                    dfb_surface_unref (data->decode_surface);
                    data->decode_surface = NULL;
               }

               if (data->image) {
                    dfb_scale_linear_32( data->image, data->image_width, data->image_height,
                                         lock.addr, lock.pitch, &rect, dst_surface, &clip );
                    if (lock.pitch)
                         dfb_surface_unlock_buffer( dst_surface, &lock );
                    if (data->render_callback) {
                         DFBRectangle r = { 0, 0, data->image_width, data->image_height };

                         if (data->render_callback( &r, data->render_callback_context ) != DIRCR_OK)
                              return DFB_INTERRUPTED;
                    }

                    return DFB_INCOMPLETE;
               }
               else if (lock.pitch)
                    dfb_surface_unlock_buffer( dst_surface, &lock );

               return DFB_FAILURE;
          }

restart:
          jpeg_create_decompress(&cinfo);
          jpeg_buffer_src(&cinfo, data->buffer, 0);
          jpeg_read_header(&cinfo, TRUE);

#if JPEG_LIB_VERSION >= 70
          cinfo.scale_num = 8;
          cinfo.scale_denom = 8;
#else
          cinfo.scale_num = 1;
          cinfo.scale_denom = 1;
#endif
          jpeg_calc_output_dimensions(&cinfo);

          if (rect.x == 0 && rect.y == 0) {
#if JPEG_LIB_VERSION >= 70
               cinfo.scale_num = 16;
               while (cinfo.scale_num > 1) {
                    jpeg_calc_output_dimensions (&cinfo);
                    if (cinfo.output_width <= rect.w
                        || cinfo.output_height <= rect.h)
                         break;
                    --cinfo.scale_num;
               }
               jpeg_calc_output_dimensions (&cinfo);
#else
               while (cinfo.scale_denom < 8
                      && ((cinfo.output_width >> 1) >= rect.w)
                      && ((cinfo.output_height >> 1) >= rect.h)) {
                    cinfo.scale_denom <<= 1;
                    jpeg_calc_output_dimensions (&cinfo);
               }
#endif
          }

          cinfo.output_components = 3;

          data->image_width = cinfo.output_width;
          data->image_height = cinfo.output_height;

          cinfo.do_fancy_upsampling = FALSE;
          cinfo.do_block_smoothing = FALSE;

          if (try_raw) {
               /* init a state, so that we can use gfxcard/blit to convert YUV to
                  requested destination format */
               dfb_state_init (&state, data->core);
               dfb_state_set_destination (&state, dst_surface);
               dfb_state_set_clip (&state, &clip);

               ret = JPEG_raw_decode (data, dst_surface, &rect, &cinfo,
                                      &state, &lock_raw);

               /* remove the state */
               dfb_state_set_source (&state, NULL);
               dfb_state_set_destination (&state, NULL);
               dfb_state_destroy (&state);

               switch (ret) {
                    case DFB_OK:
                    case DFB_INTERRUPTED:
                         /* All ok, or callback signalled abort. */
                         return ret;

                    case DFB_UNSUPPORTED:
                         /* Not enough video memory or source image format not
                            supported. jpeg_start_decompress() was not yet
                            called, so we can just break. */
                         break;

                    default:
                         D_BUG ("JPEG_raw_decode() returned unknown "
                                "result %d\n", ret);
                    case DFB_FAILURE:
                         /* General failure during compression, we restart
                            trying a software decoding and in for that have to
                            reinitialise cinfo. */
                         try_raw = false;
                         goto restart;
               }
          }


          /* start using the full software decode path */
          D_WARN ("slow JPEG decode path");

          ret = dfb_surface_lock_buffer( dst_surface, CSBR_BACK, CSAID_CPU, CSAF_WRITE, &lock );
          if (ret) {
               jpeg_abort_decompress(&cinfo);
               jpeg_destroy_decompress(&cinfo);
               return ret;
          }

          if (cinfo.output_width == rect.w && cinfo.output_height == rect.h)
               direct = true;

          switch (dst_surface->config.format) {
               case DSPF_NV16:
                    uv_offset = dst_surface->config.size.h * lock.pitch;

                    if (direct && !rect.x && !rect.y) {
                         D_INFO( "JPEG: Using YCbCr color space directly! (%dx%d)\n",
                                 cinfo.output_width, cinfo.output_height );
                         cinfo.out_color_space = JCS_YCbCr;
                         break;
                    }
                    D_INFO( "JPEG: Going through RGB color space! (%dx%d -> %dx%d @%d,%d)\n",
                            cinfo.output_width, cinfo.output_height, rect.w, rect.h, rect.x, rect.y );

               default:
                    cinfo.out_color_space = JCS_RGB;
                    break;
          }

          jpeg_start_decompress(&cinfo);

          row_stride = cinfo.output_width * 3;

          buffer = (*cinfo.mem->alloc_sarray)((j_common_ptr) &cinfo,
                                              JPOOL_IMAGE, row_stride, 1);

          data->image = D_CALLOC( data->image_height, data->image_width * 4 );
          if (!data->image) {
               dfb_surface_unlock_buffer( dst_surface, &lock );
               return D_OOM();
          }
          row_ptr = data->image;

          while (cinfo.output_scanline < cinfo.output_height && cb_result == DIRCR_OK) {
               jpeg_read_scanlines(&cinfo, buffer, 1);

               switch (dst_surface->config.format) {
                    case DSPF_NV16:
                         if (direct) {
                              copy_line_nv16( lock.addr, lock.addr + uv_offset, *buffer, rect.w );

                              lock.addr += lock.pitch;

                              if (data->render_callback) {
                                   DFBRectangle r = { 0, y, data->image_width, 1 };

                                   cb_result = data->render_callback( &r, 
                                                                      data->render_callback_context );
                              }
                              break;
                         }

                    default:
                         copy_line32( row_ptr, *buffer, data->image_width );

                         if (direct) {
                              DFBRectangle r = { rect.x, rect.y+y, rect.w, 1 };
                              dfb_copy_buffer_32( row_ptr, lock.addr, lock.pitch,
                                                  &r, dst_surface, &clip );
                              if (data->render_callback) {
                                   r = (DFBRectangle){ 0, y, data->image_width, 1 };
                                   cb_result = data->render_callback( &r, 
                                                                      data->render_callback_context );
                              }
                         }
                         break;
               }

               row_ptr += data->image_width;
               y++;
          }

          if (!direct) {
               dfb_scale_linear_32( data->image, data->image_width, data->image_height,
                                    lock.addr, lock.pitch, &rect, dst_surface, &clip );
               if (data->render_callback) {
                    DFBRectangle r = { 0, 0, data->image_width, data->image_height };
                    cb_result = data->render_callback( &r, data->render_callback_context );
               }
          }

          if (cb_result != DIRCR_OK) {
               jpeg_abort_decompress(&cinfo);
               D_FREE( data->image );
               data->image = NULL;
          }
          else {
               jpeg_finish_decompress(&cinfo);
          }
          jpeg_destroy_decompress(&cinfo);

          dfb_surface_unlock_buffer( dst_surface, &lock );
     }
     else if (data->decode_surface) {
          CardState    state;
          DFBRectangle src_rect = {
               .x = 0,
               .y = 0,
               .w = data->image_width,
               .h = data->image_height
          };

          /* use DFB to convert raw YUV to destination format, and
             apply any necessary additional clip/stretch */
          dfb_state_init (&state, data->core);
          dfb_state_set_source (&state, data->decode_surface);
          dfb_state_set_destination (&state, dst_surface);
          dfb_state_set_clip (&state, &clip);

          JPEG_stretchblit (&state, &src_rect, &rect);

          /* remove the state */
          dfb_state_set_source (&state, NULL);
          dfb_state_set_destination (&state, NULL);
          dfb_state_destroy (&state);

          if (data->render_callback) {
               DFBRectangle r = { 0, 0, data->image_width, data->image_height };
               data->render_callback (&r, data->render_callback_context);
          }
     }
     else {
          ret = dfb_surface_lock_buffer( dst_surface, CSBR_BACK, CSAID_CPU, CSAF_WRITE, &lock );
          if (ret)
            return ret;

          dfb_scale_linear_32( data->image, data->image_width, data->image_height,
                               lock.addr, lock.pitch, &rect, dst_surface, &clip );
          if (data->render_callback) {
               DFBRectangle r = { 0, 0, data->image_width, data->image_height };
               data->render_callback( &r, data->render_callback_context );
          }

          dfb_surface_unlock_buffer( dst_surface, &lock );
     }

     if (cb_result != DIRCR_OK)
         return DFB_INTERRUPTED;

     return DFB_OK;
}

static DFBResult
IDirectFBImageProvider_JPEG_SetRenderCallback( IDirectFBImageProvider *thiz,
                                               DIRenderCallback        callback,
                                               void                   *context )
{
     DIRECT_INTERFACE_GET_DATA (IDirectFBImageProvider_JPEG)

     data->render_callback         = callback;
     data->render_callback_context = context;

     return DFB_OK;
}

static DFBResult
IDirectFBImageProvider_JPEG_GetSurfaceDescription( IDirectFBImageProvider *thiz,
                                                   DFBSurfaceDescription  *dsc )
{
     DIRECT_INTERFACE_GET_DATA(IDirectFBImageProvider_JPEG)

     dsc->flags  = DSDESC_WIDTH |  DSDESC_HEIGHT | DSDESC_PIXELFORMAT;
     dsc->height = data->height;
     dsc->width  = data->width;
     dsc->pixelformat = dfb_primary_layer_pixelformat();

     return DFB_OK;
}

static DFBResult
IDirectFBImageProvider_JPEG_GetImageDescription( IDirectFBImageProvider *thiz,
                                                 DFBImageDescription    *dsc )
{
     DIRECT_INTERFACE_GET_DATA(IDirectFBImageProvider_JPEG)

     if (!dsc)
          return DFB_INVARG;

     dsc->caps = DICAPS_NONE;

     return DFB_OK;
}
