/**
 * This module shows the status of front end adapters in sysfs as:
 * /sys/dvb-frontend
 */

#include <linux/kobject.h>
#include <linux/string.h>
#include <linux/sysfs.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/init.h>
#include <frontend_status.h>

#define to_fe_kroot_attr(x) container_of(x, struct fe_kroot_attribute, attr)

static struct kobject *fe_kroot;

struct fe_kroot_attribute {
	struct attribute attr;
	ssize_t (*show)(struct kobject *fe_kroot, struct fe_kroot_attribute *attr, char *buf);
	/* frontend number */
	int fe_id;
};

static ssize_t fe_kroot_attr_show(struct kobject *kobj, struct attribute *attr, char *buf);
static ssize_t dmod_show(struct kobject *fe_kroot_obj, struct fe_kroot_attribute *attr, char *buf);
static ssize_t tuner_show(struct kobject *fe_kroot_obj, struct fe_kroot_attribute *attr, char *buf);
static void fe_kroot_release(struct kobject *kobj);

extern int dvb_frontend_get_status(int frontend_num, frontend_query_t query, int *result);
extern int dvb_frontend_get_count(int *count);

static struct sysfs_ops fe_kroot_sysfs_ops = {
	.show = fe_kroot_attr_show,
	.store = NULL,
};

static struct kobj_type fe_kroot_ktype = {
	.sysfs_ops = &fe_kroot_sysfs_ops,
	.release = fe_kroot_release,
	.default_attrs = NULL,
};

static ssize_t fe_kroot_attr_show(struct kobject *kobj, struct attribute *attr, char *buf)
{
	struct fe_kroot_attribute *attribute;

	attribute = to_fe_kroot_attr(attr);
	if (!attribute->show) {
		return -EIO;
	}
	return attribute->show(kobj, attribute, buf);
}

/**
 * fe_kroot_release - Releases the memory held by given object
 * Do not call manually.
 * @kobj
 *
 * REQUIRED by the kernel.
 */
static void fe_kroot_release(struct kobject *kobj)
{
	int i;

	for (i = 0; kobj->ktype->default_attrs[i] != NULL; i++) {
		kfree(kobj->ktype->default_attrs[i]->name);
		kfree(kobj->ktype->default_attrs[i]);
	}
	kfree(kobj->ktype->default_attrs);
	kfree(kobj);
}

/**
 * dmod_show - Shows the status of the dmod for given frontend
 * @fe_kroot_obj
 * @attr
 * @buf
 */
static ssize_t dmod_show(struct kobject *fe_kroot_obj, struct fe_kroot_attribute *attr, char *buf)
{
	int status;

	if (dvb_frontend_get_status(attr->fe_id, STATUS_DMOD, &status)) {
		return 0;
	}
	return sprintf(buf, "%d\n", status);
}

static ssize_t tuner_show(struct kobject *fe_kroot_obj, struct fe_kroot_attribute *attr, char *buf)
{
	int status;

	if (dvb_frontend_get_status(attr->fe_id, STATUS_TUNER, &status)) {
		return 0;
	}
	return sprintf(buf, "%d\n", status);
}

static void destroy_fe_kroot_obj(struct kobject *fe_kroot)
{
	kobject_del(fe_kroot);
	kobject_put(fe_kroot);
}

static int __init frontend_status_init(void)
{
	struct fe_kroot_attribute *attribute = NULL;
	int retval;
	char *file_entry_name = NULL;
	unsigned int fe_count;
	int fe_count_max = 0;

	/* allocate the memory for the whole object */
	fe_kroot = kzalloc(sizeof(*fe_kroot), GFP_KERNEL);
	if (fe_kroot == NULL) {
		return -EINVAL;
	}
	fe_kroot->kset = NULL;
	kobject_init(fe_kroot);
	if (dvb_frontend_get_count(&fe_count_max)) {
		return -EINVAL;
	}
	/* for each frontend, add the tuner and dmod status */
	fe_kroot_ktype.default_attrs = kmalloc((fe_count_max + 1) * sizeof(struct attribute *), GFP_KERNEL);
	if (fe_kroot_ktype.default_attrs == NULL) {
		return -ENOMEM;
	}
	for (fe_count = 0; fe_count < fe_count_max; fe_count++) {
		/* sets the dmod file entry attribute */
		attribute =  kmalloc(sizeof(struct fe_kroot_attribute), GFP_KERNEL);
		if (attribute == NULL) {
			return -ENOMEM;
		}
		file_entry_name = kasprintf(GFP_KERNEL, "fe%d-dmod-status", fe_count);
		if (file_entry_name == NULL) {
			return -ENOMEM;
		}
		attribute->attr.name = file_entry_name;
		attribute->attr.mode = 0444;
		attribute->show = dmod_show;
		attribute->fe_id = fe_count;
		fe_kroot_ktype.default_attrs[fe_count * 2] = &attribute->attr;
		/* sets the tuner file entry attribute */
		attribute =  kmalloc(sizeof(struct fe_kroot_attribute), GFP_KERNEL);
		if (attribute == NULL) {
			return -ENOMEM;
		}
		file_entry_name = kasprintf(GFP_KERNEL, "fe%d-tuner-status", fe_count);
		if (file_entry_name == NULL) {
			return -ENOMEM;
		}
		attribute->attr.name = file_entry_name;
		attribute->attr.mode = 0444;
		attribute->show = tuner_show;
		attribute->fe_id = fe_count;
		fe_kroot_ktype.default_attrs[fe_count * 2 + 1] = &attribute->attr;
	}
	/* need to null terminate the list of attributes */
	fe_kroot_ktype.default_attrs[fe_count * 2] = NULL;
	fe_kroot->ktype = &fe_kroot_ktype;
	kobject_set_name(fe_kroot, "dvb-frontend");
	fe_kroot->parent = NULL;
	retval = kobject_add(fe_kroot);
	/* it is responsible for sending the uevent that the kobject was added to the
	   system.  */
	kobject_uevent(fe_kroot, KOBJ_ADD);
	return retval;
}

static void __exit frontend_status_exit(void)
{
	destroy_fe_kroot_obj(fe_kroot);
}

module_init(frontend_status_init);
module_exit(frontend_status_exit);
MODULE_LICENSE("GPL");

