/*
 *
 *  D-Bus++ - C++ bindings for D-Bus
 *
 *  Copyright (C) 2005-2007  Paolo Durante <shackan@gmail.com>
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include <dbus-c++/debug.h>
#include <dbus-c++/object.h>
#include "internalerror.h"

#include <map>
#include <dbus/dbus.h>

#include "message_p.h"
#include "server_p.h"
#include "connection_p.h"

extern "C" {

#include <sys/queue.h>
#include <setjmp.h>
#include <semaphore.h>

//#define dbg printf
static void dbg(const char *fmt, ...) { }

typedef struct pool_s pool_t;

typedef struct thd_s {
	LIST_ENTRY(thd_s) entry;
	sem_t sem;
	sem_t fini_sem;
	pthread_t tid;
	void *(*func) (void *);
	void *arg, *ret;
	pool_t *pool;
	int detached;
	jmp_buf exit;
} thd_t;

struct pool_s {
	LIST_HEAD(, thd_s) pool;
};

static pool_t pool;
static pthread_mutex_t pool_mutex = PTHREAD_MUTEX_INITIALIZER;
static LIST_HEAD(, thd_s) thd_running[256];

static inline uint8_t hash(pthread_t tid)
{
	uint32_t x = (uint32_t) tid;
	//printf("thd hash %d\n", (uint8_t) (x ^ (x >> 8) ^ (x >> 16) ^ (x >> 24)));
	return x ^ (x >> 8) ^ (x >> 16) ^ (x >> 24);
}

static void *reusable_thread(thd_t *t)
{
	int loop = 0;
	for ( ; ; ) {
		dbg("_pool_ thd %lx: waiting START\n", t->tid);
		sem_wait(&t->sem);
		dbg("_pool_ thd %lx: awaken with task %p %p (thread reused #%d)\n", t->tid, t->func, t->arg, loop++);
		if (setjmp(t->exit))
			dbg("_pool_ jump !!\n");
		else
			t->ret = t->func(t->arg);
		dbg("_pool_ thd %lx: task finished, returns %p\n", t->tid, t->ret);

		sem_post(&t->fini_sem);
		dbg("_pool_ thd %lx: waiting JOIN\n", t->tid);
		sem_wait(&t->sem);
		dbg("_pool_ thd %lx: JOIN done\n", t->tid);

		if (t->detached) {
			sem_wait(&t->fini_sem);
			dbg("_pool_ thd %lx: DETACHED\n", t->tid);
		}

		/* put back the thread in pool */
		pthread_mutex_lock(&pool_mutex);
		LIST_REMOVE(t, entry);
		LIST_INSERT_HEAD(&t->pool->pool, t, entry);
		pthread_mutex_unlock(&pool_mutex);
	}

	return NULL;
}

static thd_t *new_reusable_thread()
{
	thd_t *t = (thd_t*)malloc(sizeof(thd_t));
	static int nb;

	dbg("NEW REAL THREAD #%d\n", ++nb);

	sem_init(&t->sem, 0, 0);
	sem_init(&t->fini_sem, 0, 0);
	if (pthread_create(&t->tid, NULL, (void* (*)(void*))reusable_thread, t)) {
		free(t);
		return NULL;
	}

	return t;
}
static thd_t *get_reusable_thread()
{
	thd_t *res;

	pthread_mutex_lock(&pool_mutex);
	res = LIST_FIRST(&pool.pool);
	if (res) {
		LIST_REMOVE(res, entry);
	} else
		res = new_reusable_thread();

	LIST_INSERT_HEAD(&thd_running[hash(res->tid)], res, entry);
	res->pool = &pool;
	pthread_mutex_unlock(&pool_mutex);

	return res;
}

static thd_t *find_thread(pthread_t tid)
{
	thd_t *res;
	pthread_mutex_lock(&pool_mutex);
	LIST_FOREACH(res, &thd_running[hash(tid)], entry)
		if (res->tid == tid)
			break;
	pthread_mutex_unlock(&pool_mutex);
	return res;
}

static int _pool_pthread_create (pthread_t *__restrict __newthread,
			     __const pthread_attr_t *__restrict __attr,
			     void *(*__start_routine) (void *),
			     void *__restrict __arg)
{
	dbg("_pool_pthread_create\n");

	if (__attr)
	{
		dbg("_pool_pthread_create ERROR : non NULL __attr\n");
		return 1; // TODO : Should be EINVAL
	}

	thd_t *t = get_reusable_thread();
	t->func = __start_routine;
	t->arg = __arg;
	t->detached = 0;

	sem_post(&t->sem);
	dbg("_pool_ --> %lx %p\n", t->tid, t);

	*__newthread = t->tid;
	return 0;
}

static int _pool_pthread_detach (pthread_t __th)
{
	thd_t *t = find_thread(__th);
	dbg("_pool_ pthread_detach %lx (%p)\n", __th, t);
	if (t->detached)
		return -1;
	t->detached = 1;
	sem_post(&t->sem);
	return 0;
}

static void _pool_pthread_exit(void *__retval)
{
	thd_t *t = find_thread(pthread_self());
	dbg("_pool_ pthread_exit !!!!!!!!!! (%p)\n", t);
	if (t) {
		t->ret = __retval;
		longjmp(t->exit, 1);
	} else
		pthread_exit(__retval);
}

} // extern "C"

using namespace DBus;

Object::Object( Connection& conn, const Path& path, const char* service )
:	_conn(conn), _path(path), _service(service ? service : "")
{
}

Object::~Object()
{
}

struct ObjectAdaptor::Private
{
	static void unregister_function_stub( DBusConnection*, void* );
	static DBusHandlerResult message_function_stub( DBusConnection*, DBusMessage*, void* );
};

static DBusObjectPathVTable _vtable =
{	
	ObjectAdaptor::Private::unregister_function_stub,
	ObjectAdaptor::Private::message_function_stub,
	NULL, NULL, NULL, NULL
};

void ObjectAdaptor::Private::unregister_function_stub( DBusConnection* conn, void* data )
{
 	//TODO: what do we have to do here ?
}

DBusHandlerResult ObjectAdaptor::Private::message_function_stub( DBusConnection*, DBusMessage* dmsg, void* data )
{
	ObjectAdaptor* o = static_cast<ObjectAdaptor*>(data);

	if( o )
	{
		Message msg(new Message::Private(dmsg));

		debug_log("in object %s", o->path().c_str());
		debug_log(" got message #%d from %s to %s",
			msg.serial(),
			msg.sender(),
			msg.destination()
		);

		return o->handle_message(msg)
			? DBUS_HANDLER_RESULT_HANDLED
			: DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	}
	else
	{
		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	}
}

typedef std::map<Path, ObjectAdaptor*> ObjectAdaptorTable;
static ObjectAdaptorTable _adaptor_table;

ObjectAdaptor* ObjectAdaptor::from_path( const Path& path )
{
	ObjectAdaptorTable::iterator ati = _adaptor_table.find(path);

	if(ati != _adaptor_table.end())
		return ati->second;

	return NULL;
}

ObjectAdaptorPList ObjectAdaptor::from_path_prefix( const std::string& prefix )
{
	ObjectAdaptorPList ali;

	ObjectAdaptorTable::iterator ati = _adaptor_table.begin();

	size_t plen = prefix.length();

	while(ati != _adaptor_table.end())
	{
		if(!strncmp(ati->second->path().c_str(), prefix.c_str(), plen))
			ali.push_back(ati->second);

		++ati;
	}

	return ali;
}

ObjectAdaptor::ObjectAdaptor( Connection& conn, const Path& path )
:	Object(conn, path, conn.unique_name())
{
	register_obj();
}

ObjectAdaptor::~ObjectAdaptor()
{
	unregister_obj();
}

void ObjectAdaptor::register_obj()
{
	debug_log("registering local object %s", path().c_str());

	if(!dbus_connection_register_object_path(conn()._pvt->conn, path().c_str(), &_vtable, this))
	{
 		throw ErrorNoMemory("unable to register object path");
	}
	else
	{
		InterfaceAdaptorTable::const_iterator ii = _interfaces.begin();
		while( ii != _interfaces.end() )
		{
			std::string im = "type='method_call',interface='"+ii->first+"',path='"+path()+"'";
			conn().add_match(im.c_str());
			++ii;
		}
	}

	_adaptor_table[path()] = this;
}

void ObjectAdaptor::unregister_obj()
{
	_adaptor_table.erase(path());

	debug_log("unregistering local object %s", path().c_str());

	dbus_connection_unregister_object_path(conn()._pvt->conn, path().c_str());

	InterfaceAdaptorTable::const_iterator ii = _interfaces.begin();
	while( ii != _interfaces.end() )
	{
		std::string im = "type='method_call',interface='"+ii->first+"',path='"+path()+"'";
		conn().remove_match(im.c_str());
		++ii;
	}
}

void ObjectAdaptor::_emit_signal( SignalMessage& sig )
{
	sig.path(path().c_str());

	conn().send(sig);
}

struct ReturnLaterError
{
	const Tag* tag;
};

class MethodHandlerParam
{
public:
	MethodHandlerParam( ObjectAdaptor* a,
			    InterfaceAdaptor* i, 
			    const CallMessage& c ):
		adaptor(a), iface(i), cmsg(c)
	{}

	ObjectAdaptor*    adaptor;
	InterfaceAdaptor* iface;
	CallMessage       cmsg;
};

void* ObjectAdaptor::method_handler_launcher_thread( void* p )
{
	MethodHandlerParam* param = reinterpret_cast<MethodHandlerParam*>(p);

	debug_log("launching method handler");

	param->adaptor->method_handler_launcher(param->iface, param->cmsg);

	debug_log("method handler exiting ...");

	delete param;

	_pool_pthread_exit(NULL);
}

void ObjectAdaptor::method_handler_launcher( InterfaceAdaptor* ii, const CallMessage& cmsg )
{
	try
	{
		Message ret = ii->dispatch_method(cmsg);
		conn().send(ret);
	}
	catch(Error& e)
	{
		ErrorMessage em(cmsg, e.name(), e.message());
		conn().send(em);
	}
	catch(ReturnLaterError& rle)
	{
		_continuations[rle.tag] = new Continuation(conn(), cmsg, rle.tag);
	}
}

bool ObjectAdaptor::handle_message( const Message& msg )
{
	switch( msg.type() )
	{
		case DBUS_MESSAGE_TYPE_METHOD_CALL:
		{
			const CallMessage& cmsg = reinterpret_cast<const CallMessage&>(msg);
			const char* member      = cmsg.member();
			const char* interface   = cmsg.interface();

			debug_log(" invoking method %s.%s", interface, member);

			InterfaceAdaptor* ii = find_interface(interface);
			if( ii )
			{
				pthread_t handler;

				MethodHandlerParam* param = new MethodHandlerParam(this, ii, cmsg);

				if (_pool_pthread_create(&handler, NULL,
						   method_handler_launcher_thread, param)) {

					// TODO: exception?
					debug_log("failed to create method handler");
					return false;
				}

				if (_pool_pthread_detach(handler)) {
					debug_log("failed to detach method handler");
					return false;
				}

				// TODO: return code?
				return true;
			}
			else
			{
				return false;
			}
		}
		default:
		{
			return false;
		}
	}
}

void ObjectAdaptor::return_later( const Tag* tag )
{
	ReturnLaterError rle = { tag };
	throw rle;
}

void ObjectAdaptor::return_now( Continuation* ret )
{
	ret->_conn.send(ret->_return);

	ContinuationMap::iterator di = _continuations.find(ret->_tag);

	delete di->second;

	_continuations.erase(di);
}

void ObjectAdaptor::return_error( Continuation* ret, const Error error )
{
	ret->_conn.send(ErrorMessage(ret->_call, error.name(), error.message()));

	ContinuationMap::iterator di = _continuations.find(ret->_tag);

	delete di->second;

	_continuations.erase(di);
}

ObjectAdaptor::Continuation* ObjectAdaptor::find_continuation( const Tag* tag )
{
	ContinuationMap::iterator di = _continuations.find(tag);

	return di != _continuations.end() ? di->second : NULL;
}

ObjectAdaptor::Continuation::Continuation( Connection& conn, const CallMessage& call, const Tag* tag )
: _conn(conn), _call(call), _return(_call), _tag(tag)
{
	_writer = _return.writer(); //todo: verify
}

/*
*/

ObjectProxy::ObjectProxy( Connection& conn, const Path& path, const char* service )
:	Object(conn, path, service)
{
	register_obj();
}

ObjectProxy::~ObjectProxy()
{
	unregister_obj();
}

void ObjectProxy::register_obj()
{
	debug_log("registering remote object %s", path().c_str());

	_filtered = new Callback<ObjectProxy, bool, const Message&>(this, &ObjectProxy::handle_message);
	
	conn().add_filter(_filtered);

	InterfaceProxyTable::const_iterator ii = _interfaces.begin();
	while( ii != _interfaces.end() )
	{
		std::string im = "type='signal',interface='"+ii->first+"',path='"+path()+"'";
		conn().add_match(im.c_str());
		++ii;
	}
}

void ObjectProxy::unregister_obj()
{
	debug_log("unregistering remote object %s", path().c_str());
	
	InterfaceProxyTable::const_iterator ii = _interfaces.begin();
	while( ii != _interfaces.end() )
	{
		std::string im = "type='signal',interface='"+ii->first+"',path='"+path()+"'";
		conn().remove_match(im.c_str());
		++ii;
	}
	conn().remove_filter(_filtered);
}

Message ObjectProxy::_invoke_method( CallMessage& call )
{
	call.path(path().c_str());
	call.destination(service().c_str());

	return conn().send_blocking(call, 20000); // set timeout to 20 second
}

class SignalHandlerParam
{
public:
	SignalHandlerParam( InterfaceProxy* p, const SignalMessage& m ):
		proxy(p), smsg(m) {}


	InterfaceProxy* proxy;
	SignalMessage   smsg;
};

void* ObjectProxy::sighandler_launcher( void* p )
{
	SignalHandlerParam* param = reinterpret_cast<SignalHandlerParam*>(p);

	param->proxy->dispatch_signal(param->smsg);

	delete param;

	_pool_pthread_exit(NULL);
}

bool ObjectProxy::handle_message( const Message& msg )
{
	switch( msg.type() )
	{
		case DBUS_MESSAGE_TYPE_SIGNAL:
		{
			const SignalMessage& smsg = reinterpret_cast<const SignalMessage&>(msg);
			const char* interface	= smsg.interface();
			const char* member	= smsg.member();
			const char* objpath	= smsg.path();

			if( objpath != path() ) return false;

			debug_log("filtered signal %s(in %s) from %s to object %s",
				member, interface, msg.sender(), objpath);

			InterfaceProxy* ii = find_interface(interface);
			if( ii )
			{
				pthread_t handler;

				SignalHandlerParam* param = new SignalHandlerParam(ii, smsg);

				if (_pool_pthread_create(&handler, NULL,
						   sighandler_launcher, param)) {

					// TODO: exception?
					debug_log("failed to create signal handler");
					return false;
				}

				if (_pool_pthread_detach(handler)) {
					debug_log("failed to detach signal handler");
					return false;
				}

				// TODO: return code?
				return true;
			}
			else
			{
				return false;
			}
		}
		default:
		{
			return false;
		}
	}
}
