/*
 * os_abstractions.h
 *
 * Copyright (C) STMicroelectronics Limited 2001. All rights reserved.
 *
 * Types and functions to provide an OS independant view of multi-thread
 * issues.
 */

#ifndef _STRPC_OS_ABSTRACTIONS_H_
#define _STRPC_OS_ABSTRACTIONS_H_

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*
 * Initialization
 *
 * Perform any initialization needed for the abstraction layer
 *
 * Provides:
 *   OS_INIT
 *   OS_DEINIT
 */

#define OS_INIT()
#define OS_DEINIT()

/*
 * OS_MUTEX
 *
 * A simple mutex type that cannot (necessarily) be taken more than once. 
 *
 * Provides:
 *   OS_MUTEX_INIT    - Initialize the mutex
 *   OS_MUTEX_DESTROY - Uninitialize the mutex and return any memory used
 *   OS_MUTEX_TAKE    - Obtain/lock the mutex
 *   OS_MUTEX_RELEASE - Release/unlock the mutex
 */

#if defined __OS21__

#include <os21/semaphore.h>

typedef semaphore_t *OS_MUTEX;

#define OS_MUTEX_INIT(pMutex)		*(pMutex) = semaphore_create_fifo(1)
#define OS_MUTEX_DESTROY(pMutex)	semaphore_delete(*(pMutex));
#define OS_MUTEX_TAKE(pMutex)		semaphore_wait(*(pMutex));
#define OS_MUTEX_RELEASE(pMutex)	semaphore_signal(*(pMutex));

#define HZ				time_ticks_per_sec()
#define OS_TASK_DELAY(T)		task_delay((T))

#elif defined __LINUX__ && defined __KERNEL__

/* Linux kernel space OS abstractions */

#include <linux/module.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 23)
#include <linux/config.h>
#endif
#include <linux/stddef.h>
#include <linux/wait.h>
#include <asm/semaphore.h>

typedef struct semaphore OS_MUTEX;

#define OS_MUTEX_INIT(pMutex)		sema_init((pMutex), 1)
#define OS_MUTEX_DESTROY(pMutex)
#define OS_MUTEX_TAKE(pMutex)		down_interruptible((pMutex))
#define OS_MUTEX_RELEASE(pMutex)	up((pMutex))

#define OS_TASK_DELAY(T)		schedule_timeout(T)

#elif defined __LINUX__

/* Linux user space OS abstractions */

#include <semaphore.h>

typedef sem_t OS_MUTEX;

#define OS_MUTEX_INIT(pMutex)		sem_init((pMutex), 0, 1)
#define OS_MUTEX_DESTROY(pMutex)	sem_destroy((pMutex));
#define OS_MUTEX_TAKE(pMutex)		sem_wait((pMutex))
#define OS_MUTEX_RELEASE(pMutex)	sem_post((pMutex))

#include <unistd.h>
#define HZ 				1
#define OS_TASK_DELAY(T)		sleep((T))

#elif defined __WINCE__

#include <windows.h>

typedef HANDLE OS_MUTEX;

#define OS_MUTEX_INIT(pMutex)		*(pMutex) = CreateEvent(NULL, FALSE, TRUE, NULL)
#define OS_MUTEX_DESTROY(pMutex)	CloseHandle(*(pMutex));
#define OS_MUTEX_TAKE(pMutex)		WaitForSingleObject(*(pMutex), INFINITE)
#define OS_MUTEX_RELEASE(pMutex)	SetEvent(*(pMutex))

#else
#error OS_MUTEX not yet ported to this operating system
#endif


/*
 * OS_SIGNAL
 *
 * A simple signal type that provides a means for a task to block whilst
 * waiting for a particular event.
 *
 * Provides:
 *   OS_SIGNAL_INIT    - Initialize the signal (can appear in critical paths)
 *   OS_SIGNAL_DESTROY - Uninitialize the mutex and return any memory used
 *   OS_SIGNAL_WAIT    - Wait to the signal to be posted
 *   OS_SIGNAL_POST    - Wave the flag, signal the signal
 */


/* A default implementation based on OS_MUTEX. This only works if OS_MUTEX
 * cannot be taken twice by the same process.
 */

typedef OS_MUTEX OS_SIGNAL;
#define OS_SIGNAL_INIT(pSignal) \
	{ \
		OS_MUTEX_INIT((pSignal)); \
		OS_MUTEX_TAKE((pSignal)); \
	}
#define OS_SIGNAL_DESTROY(pSignal)	OS_MUTEX_DESTROY((pSignal))
#define OS_SIGNAL_WAIT(pSignal)		OS_MUTEX_TAKE((pSignal))
#define OS_SIGNAL_POST(pSignal)		OS_MUTEX_RELEASE((pSignal))

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* _STRPC_OS_ABSTRACTIONS_H_ */
