/*
 * host.c
 *
 * Copyright (C) STMicroelectronics Limited 2004. All rights reserved.
 *
 * 
 */

#include <mme.h>
#include <os_abstractions.h>

#include "harness.h"
#include "hanger.h"

#define TRANSPORT	"shm"
#define TRANSFORMER	"mme.8bit_mixer"

extern EMBX_ERROR EMBXSHM_RemoveProcessor(EMBX_TRANSPORT htp, EMBX_UINT cpuID);

static OS_SIGNAL callbackSignal;

static char sample1[] = { 34, 38, 33, 19,  5,  8,  8, 23, 36, 38, 31, 16,
			   3, 13, 10, 25, 38, 39, 29, 13, 28, 20, 12, 28 };
static char sample2[] = {  4, 28, 18, 19,  4,  8,  8,  5, 12, 15, 15, 12,
			   8, 29, 13, 28, 20, 12, 28, 23, 26, 28, 21, 16 };

static void mixerCallback(MME_Event_t Event, MME_Command_t *CallbackData, void *UserData)
{
	OS_SIGNAL_POST(&callbackSignal);
}

static MME_ERROR simpleMix(MME_TransformerHandle_t hdl)
{
	MME_ERROR err;
	MME_DataBuffer_t *bufs[3] = { 0, 0, 0 };
	MME_Command_t cmd = { sizeof(MME_Command_t) };

	/* allocate databuffers using MME_ALLOCATION_PHYSICAL to guarantee
	 * a single scatter page per allocation */

	err = MME_AllocDataBuffer(hdl, sizeof(sample1), MME_ALLOCATION_PHYSICAL, bufs+0);
	if (MME_SUCCESS != err) goto error_recovery;
	memcpy(bufs[0]->ScatterPages_p->Page_p, sample1, sizeof(sample1));

	err = MME_AllocDataBuffer(hdl, sizeof(sample2), MME_ALLOCATION_PHYSICAL, bufs+1);
	if (MME_SUCCESS != err) goto error_recovery;
	memcpy(bufs[1]->ScatterPages_p->Page_p, sample2, sizeof(sample2));

	err = MME_AllocDataBuffer(hdl, sizeof(sample1), MME_ALLOCATION_PHYSICAL, bufs+2);
	if (MME_SUCCESS != err) goto error_recovery;

	/* fill in the non-zero parts of the command structure and issue the command.
	 * 
	 * Note that setting constant due times will cause transforms to happen in 
	 * strict FIFO order within a transformer (and round-robin within multiple 
	 * transformers) 
	 */
	cmd.CmdCode = MME_TRANSFORM;
	cmd.CmdEnd = MME_COMMAND_END_RETURN_NOTIFY;
	cmd.DueTime = 0;
	cmd.NumberInputBuffers = 2;
	cmd.NumberOutputBuffers = 1;
	cmd.DataBuffers_p = bufs;

	printf("Simulating a remote failure\n");
	err = MME_SendCommand(hdl, &cmd);

	/* shortly after this command the companion dies for good ... */
	printf("Delay 2s ..... \n");
	OS_TASK_DELAY(2 * HZ);

	/* 3. Use MME_KillCommand() to kill any outstanding commands. */
	err = MME_KillCommand(hdl, cmd.CmdStatus.CmdId);
	if (MME_SUCCESS != err) {
		printf("ERROR: Failed to issue kill command (%d)\n", err);
	}
	OS_SIGNAL_WAIT(&callbackSignal);

	/* in normal use we would send many more commands before returning (at which
	 * point the transformer will be terminated.
	 */

	/*FALLTHRU*/
    error_recovery:

	if (bufs[0]) (void) MME_FreeDataBuffer(bufs[0]);
	if (bufs[1]) (void) MME_FreeDataBuffer(bufs[1]);
	if (bufs[2]) (void) MME_FreeDataBuffer(bufs[2]);

	return err;
}

int cycle(void)
{
	MME_TransformerInitParams_t initParams = { sizeof(MME_TransformerInitParams_t) };
	MME_TransformerCapability_t capability = { sizeof(MME_TransformerCapability_t) };
	MixerParamsInit_t mixerInit;
	MME_TransformerHandle_t remoteHdl;

	MME_ERROR err;

	EMBX_FACTORY hFactory;


	printf("Host: Starting test cycle\n");

	/* (re) initialize the underlying EMBX transport */
	hFactory = embxRuntimeInit();

#ifdef LX_LOAD_CODE
	/* Load the LX code directly in the coprocessor and start */
	if (harness_slaveLoad(LX_LOAD_CPU, LX_LOAD_CODE) != 0)
		return MME_DRIVER_NOT_INITIALIZED;
#endif

	/* other misc. initialization */
	OS_SIGNAL_INIT(&callbackSignal);

	/* pre-construct the data structure to be passed at transformer 
	 * initialization time
	 */
	initParams.Priority = MME_PRIORITY_NORMAL;
	initParams.Callback = mixerCallback;
	initParams.TransformerInitParamsSize = sizeof(MixerParamsInit_t);
	initParams.TransformerInitParams_p = &mixerInit;
	MME_INDEXED_PARAM(mixerInit, MixerParamsInit_volume, 0) = 1.0;
	MME_INDEXED_PARAM(mixerInit, MixerParamsInit_volume, 1) = 4.0/3.0;

#if !(defined(__LINUX__) && !defined(__KERNEL__))
	/* register the transport for the remote transformer (at this point we will
	 * block until the remote processor initializes itself)
	 */
	printf("Registering transport %s\n", TRANSPORT);
	err = MME_RegisterTransport(TRANSPORT);
	if (MME_SUCCESS != err) {
		printf("ERROR: cannot register transport to external processors (%d)\n", err);
		return err;
	}
#endif

	/* repeatly attempt to connect to the remote transformer until it is registered or
	 * a serious error occurs.
	 */
	do {
		err = MME_InitTransformer(TRANSFORMER, &initParams, &remoteHdl);
	} while (err == MME_UNKNOWN_TRANSFORMER);

	if (MME_SUCCESS != err) {
		printf("ERROR: cannot instantiate %s (%d)\n", TRANSFORMER, err);
	}

	/* perform the same mix as above using a remotely registered transformer. the 
	 * results should be identical assuming both processors implement IEEE single
	 * precision floating point maths.
	 */
	err = simpleMix(remoteHdl);
	if (MME_SUCCESS != err) {
		printf("ERROR: cannot perform mixer transformer on %s (%d)\n", TRANSFORMER, err);
	}

	printf("Attempting to get transformer %s capability\n", TRANSFORMER);

	/* Do a blocking GetTransformerCapability call */
	err = MME_GetTransformerCapability(TRANSFORMER, &capability);

	printf("MME_GetTransformerCapability returned %d\n", err);
	
	printf("Attempting to terminate %s\n", TRANSFORMER);

	/* clean up prior to shutting down */
	err = MME_TermTransformer(remoteHdl);
	if (MME_SUCCESS != err) {
		printf("Cannot terminate %s (%d)\n", TRANSFORMER, err);
		
		assert(err == MME_TRANSFORMER_NOT_RESPONDING);

		printf("Attempting recovery\n");
		
		/* 1. place companion into reset (not needed in this example) */

		/* 2. Use EMBXSHM_RemoveProcessor to make EMBX forget about its co-processor */
		{
		  EMBX_TRANSPORT tp;
		  EMBX_ERROR embx_err;
		  
		  EMBX_UINT cpuID = 1;	/* Remove CPU#1 (Audio) */

		  embx_err = EMBX_OpenTransport(TRANSPORT, &tp);
		  if (EMBX_SUCCESS != embx_err) {
		    printf("ERROR: Could not obtain transport handle (%d)\n", embx_err);
		  }
		  
		  embx_err = EMBXSHM_RemoveProcessor(tp, cpuID);
		  if (EMBX_SUCCESS != embx_err) {
		    printf("ERROR: Could not remove remote processor (%d)\n", embx_err);
		  }
		  
		  EMBX_CloseTransport(tp);
		}

		printf("Issuing MME_KillTransformer\n");
		err = MME_KillTransformer(remoteHdl);
		if (MME_SUCCESS != err) {
			printf("ERROR: cannot kill %s (%d)\n", TRANSFORMER, err);
			assert(err != MME_SUCCESS);
		}
	}

	printf("Deregistering transport\n");
	err = MME_DeregisterTransport(TRANSPORT);
	if (MME_SUCCESS != err) {
		printf("ERROR: cannot deregister the transport (%d)\n", err);
	}

	err = EMBX_UnregisterTransport(hFactory);
	if (EMBX_SUCCESS != err) {
	  printf("ERROR: failed to Unregister transport (%d)\n", err);
	}

	return err;
}

int main(int argc, char **argv)
{
  int i, res;
  
  /* initialize the MME API */
  res = MME_Init();
  if (MME_SUCCESS != res) {
    printf("ERROR: cannot initialize the MME API (%d)\n", res);
  }
  
  
  for (i=0; i<5; i++)
  {
    printf("Waiting for other processor (iteration %d)\n", i);
    res = cycle();
    
    if (res)
      break;
  }
  
  res = MME_Term();
  if (MME_SUCCESS != res) {
    printf("ERROR: cannot terminate the MME API (%d)\n", res);
  }
  
  (void) EMBX_Deinit();
  
  return res;
}

/*
 * Local Variables:
 *  tab-width: 8
 *  c-indent-level: 8
 *  c-basic-offset: 8
 * End:
 */
