/*
 * arch/sh/boards/st/dolmenb/setup.c
 *
 * Copyright (C) 2008 STMicroelectronics Limited
 * Author: Stuart Menefy (stuart.menefy@st.com)
 *
 * Copyright (C) 2009,2010 WyPlay SAS.
 * Marek Skuczynski <mskuczynski@wyplay.com>
 *
 * May be copied or modified under the terms of the GNU General Public
 * License.  See linux/COPYING for more information.
 *
 * WyPlay DolmenB support.
 */

#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/leds.h>
#include <linux/stm/pio.h>
#include <linux/stm/soc.h>
#include <linux/stm/emi.h>
#include <linux/delay.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/physmap.h>
#include <linux/mtd/partitions.h>
#include <linux/spi/spi.h>
#include <linux/spi/flash.h>
#include <sound/stm.h>
#include <linux/mtd/nand.h>
#include <linux/stm/nand.h>
#include <linux/stm/soc_init.h>
#include <linux/phy.h>
#include <linux/gpio_keys.h>
#include <linux/input.h>
#include <asm/irq-ilc.h>
#include <asm/irl.h>
#include <asm/io.h>

static int ascs[] __initdata = { 
	2 | (STASC_FLAG_NORTSCTS << 8) 
};


#ifdef CONFIG_BPA2

#include <linux/bpa2.h>

static const char *LMI_SYS_partalias_256[] = {
	"BPA2_Region0",
	"v4l2-coded-video-buffers",
	"BPA2_Region1",
	"v4l2-video-buffers",
	"coredisplay-video",
	NULL
};

static const char *bigphys_partalias_256[] = {
	"gfx-memory-1",
	NULL
};

static struct bpa2_partition_desc bpa2_parts_table[] = {
	{
		.name	= "LMI_SYS",		/* 85MB */
		.start	= 0,
		.size	= 0x05500000,
		.flags	= 0,
		.aka	= LMI_SYS_partalias_256
	},
	{
		.name	= "bigphysarea",	/* 18MB */
		.start	= 0,
		.size	= 0x01200000,
		.flags	= 0,
		.aka	= bigphys_partalias_256
	},
	{
		.name	= "gfx-memory-0",	/* 50MB */
		.start	= 0x4CE00000,
		.size	= 0x03200000,
		.flags	= 0,
		.aka	= NULL
	}
};

#endif	/* CONFIG_BPA2 */

static struct stpio_pin	*power_off_pin;

static void dolmenb_power_off(void)
{
	stpio_set_pin (power_off_pin, 1);
	for(;;);
}

static void __init dolmenb_pio_init(void)
{
#ifndef CONFIG_SPI_STM_SSC
	struct stpio_pin *pio_ptr = NULL;
#endif

#ifdef CONFIG_SPI_STM_SSC
	/* Driving the SPI boot flash with PIO2[4-7] (SSC1).
	 * Put PIO15[0-3] (SPIboot interface) into High-Z state.
	 * Right now we can expect PIO2[4-7] to be in the STPIO_BIDIR_Z1
	 * default state, therefore the lines are not left floating. */
	stpio_request_pin(15, 0, "SPIBoot_SCK", STPIO_IN);
	stpio_request_pin(15, 1, "SPIBoot_MOSI", STPIO_IN);
	stpio_request_pin(15, 2, "SPIBoot_nCS", STPIO_IN);
	stpio_request_pin(15, 3, "SPIBoot_MISO", STPIO_IN);
#else
	/* Driving the SPI boot flash with PIO15[0-3] (GPIO's).
	 * Put PIO2[4-7] (SSC1) into High-Z state.
	 * First enable temporary pull-ups, so the lines are not left
	 * floating. */
	pio_ptr = stpio_request_pin(15, 0, "SPIBoot_SCK", STPIO_BIDIR_Z1);
	stpio_free_pin(pio_ptr);
	pio_ptr = stpio_request_pin(15, 1, "SPIBoot_MOSI", STPIO_BIDIR_Z1);
	stpio_free_pin(pio_ptr);
	pio_ptr = stpio_request_pin(15, 2, "SPIBoot_nCS", STPIO_BIDIR_Z1);
	stpio_free_pin(pio_ptr);
	pio_ptr = stpio_request_pin(15, 3, "SPIBoot_MISO", STPIO_BIDIR_Z1);
	stpio_free_pin(pio_ptr);
	stpio_request_pin(2, 4, "SPI0_nCS", STPIO_IN);
	stpio_request_pin(2, 5, "SPI0_SCK", STPIO_IN);
	stpio_request_pin(2, 6, "SPI0_MOSI", STPIO_IN);
	stpio_request_pin(2, 7, "SPI0_MISO", STPIO_IN);
#endif
}

static void __init dolmenb_setup(char** cmdline_p)
{
	printk("Wyplay DOLMENB board initialisation\n");

	stx7105_early_device_init();
	stx7105_configure_asc(ascs, ARRAY_SIZE(ascs), 0);
	dolmenb_pio_init();

#ifdef CONFIG_BPA2
	bpa2_init(bpa2_parts_table, ARRAY_SIZE(bpa2_parts_table));
#endif
}

#if (defined(CONFIG_SPI_STM_PIO) || defined(CONFIG_SPI_STM_PIO_MODULE)) && \
    (defined(CONFIG_SPI_STM_SSC) || defined(CONFIG_SPI_STM_SSC_MODULE))
#error SPI_STM_PIO plus SPI_STM_SSC results in PIO conflict
#endif

static struct plat_ssc_data ssc_private_info = {
	.capability  =
		ssc0_has(SSC_I2C_CAPABILITY) |
#ifdef CONFIG_SPI_STM_SSC
		ssc1_has(SSC_SPI_CAPABILITY) |
#else
		ssc1_has(SSC_UNCONFIGURED)   |
#endif
		ssc2_has(SSC_UNCONFIGURED)   |
		ssc3_has(SSC_UNCONFIGURED),
	.spi_chipselects = {
		[1] = NULL, /* use spi_stpio_chipselect() */
	},
};

/*
 * SPI busses
 */
#ifndef CONFIG_SPI_STM_SSC
static struct platform_device spi_gpio_bus_boot_flash = {
	.name           = "spi_st_pio",
	.id             = 8,
	.num_resources  = 0,
	.dev            = {
		.platform_data = &(struct ssc_pio_t) {
			.pio = {
				{ 15, 0 },	/* SCK */
				{ 15, 1 },	/* MOSI */
				{ 15, 3 },	/* MISO */
			},
		},
	},
};
#endif

/*
 * SPI devices
 */
static struct mtd_partition spi_boot_flash_parts[] = {
	{
		.name		= "pbl",
		.size		= 16 * 4096,
		.offset		= 0x0,
		.mask_flags	= /*MTD_WRITEABLE | */ 0, /* Read-only */
	},{
		.name		= "k1",
		.size		= 500 * 4096,
		.offset		= MTDPART_OFS_APPEND,
		.mask_flags	= /*MTD_WRITEABLE | */ 0, /* Read-only */
	},{
		.name		= "k2",
		.size		= 500 * 4096,
		.offset		= MTDPART_OFS_APPEND,
		.mask_flags	= 0,		/* Writable */
	},{
		.name		= "rwdata",
		.size		= 1 * 4096,
		.offset		= MTDPART_OFS_APPEND,
		.mask_flags	= 0,		/* Writable */
	},{
		.name		= "rodata",
		.size		= 1 * 4096,
		.offset		= MTDPART_OFS_APPEND,
		.mask_flags	= 0,		/* Writable */
	},{
		.name		= "reserved",
		.size		= 6 * 4096,
		.offset		= MTDPART_OFS_APPEND,
		.mask_flags	= 0,		/* Writable */
	}
};

static struct flash_platform_data spi_boot_flash_data = {
	.name		= "m25p80",
	.parts		= spi_boot_flash_parts,
	.nr_parts	= ARRAY_SIZE(spi_boot_flash_parts),
	.type		= "sst25vf032b",	/* The actual chip model */
};

static struct spi_board_info dolmenb_spi_boot_flash[] __initdata = {
	{
		.modalias       = "m25p80",
		.platform_data  = &spi_boot_flash_data,
#ifdef CONFIG_SPI_STM_SSC
		.max_speed_hz   = 7000000,	/* Max SPI clock frequency */
		.bus_num        = 0,		/* On the first SPI SSC bus */
		.chip_select    = spi_set_cs(2, 4),
#else
		.max_speed_hz   = 1000000,	/* Max SPI clock frequency */
		.bus_num        = 8,		/* On the GPIO-emulated
						 * spi_st_pio bus above */
		.chip_select    = spi_set_cs(15, 2),
#endif
		.mode           = SPI_MODE_3,
	},
};

/*
 * USB
 */
static struct usb_init_data usb_init[2] __initdata = {
	{
		.oc_en  = 0,
		.pwr_en = 0,
	}, {
		.oc_en  = 0,
		.pwr_en = 0,
	}
};

static struct plat_stmmacphy_data phy_private_data = {
	/* Any */
	.bus_id = 0,
	.phy_addr = -1,
	.phy_mask = 0,
	.interface = PHY_INTERFACE_MODE_MII,
};

static struct platform_device dolmenb_phy_device = {
	.name		= "stmmacphy",
	.id		= 0,
	.num_resources	= 1,
	.resource	= (struct resource[]) {
		{
			.name	= "phyirq",
			.start	= -1,/* no IRQ line (?) */
			.end	= -1,
			.flags	= IORESOURCE_IRQ,
		},
	},
	.dev = {
		.platform_data = &phy_private_data,
	}
};

static struct platform_device dolmenb_leds = {
	.name = "leds-gpio",
	.id = 0,
	.dev.platform_data = &(struct gpio_led_platform_data) {
		.num_leds = 2,
		.leds = (struct gpio_led[]) {
			{
				.name = "led:0",
				.gpio = stpio_to_gpio(6, 7),
				.active_low = 1,
				.default_trigger = "timerserv",
			},
			{
				.name = "led:1",
				.gpio = stpio_to_gpio(7, 0),
				.active_low = 1,
				.default_trigger = "timerserv",
			},
		},
	},
};

#if defined(CONFIG_MTD_NAND)
/* Configuration for NAND Flash */
static struct mtd_partition nand_parts[] = {
	{
		.name   = "ubi",
		.offset = 0,
		.size   = 0x7F80000,
	}
};


static struct plat_stmnand_data nand_config = {
	/* STM_NAND_EMI data */
	.emi_withinbankoffset   = 0,
	.rbn_port               = -1,
	.rbn_pin                = -1,

	.timing_data = &(struct nand_timing_data) {
		.sig_setup      = 50,           /* times in ns */
		.sig_hold       = 50,
		.CE_deassert    = 0,
		.WE_to_RBn      = 100,
		.wr_on          = 10,
		.wr_off         = 40,
		.rd_on          = 10,
		.rd_off         = 40,
		.chip_delay     = 50,           /* in us */
	},
	.flex_rbn_connected     = 1,
};

/* Platform data for STM_NAND_EMI/FLEX/AFM. (bank# may be updated later) */
static struct platform_device nand_device =
STM_NAND_DEVICE("stm-nand-afm", 0, &nand_config,
		nand_parts, ARRAY_SIZE(nand_parts), NAND_USE_FLASH_BBT);
#endif /* CONFIG_MTD_NAND */

static struct platform_device *dolmenb_devices[] __initdata = {
	&dolmenb_leds,
	&dolmenb_phy_device,
#ifndef CONFIG_SPI_STM_SSC
	&spi_gpio_bus_boot_flash,
#endif
};

static unsigned lirc_enabled_pios[] = {
	[0] = 3 * 8 + 0
};

static int __init device_init(void)
{
    struct stpio_pin *hub_pin, *reader_pin;
    struct stpio_pin *audio_pin;
    struct stpio_pin *flashwp_pin;
    struct stpio_pin *lanrst_pin;

    /* POWER_OFF line */
    power_off_pin = stpio_request_set_pin (7, 3, "power off", STPIO_OUT, 0);
    /* set proprietary power off function */
    pm_power_off	= dolmenb_power_off;


    /* nLAN_RST line */
    lanrst_pin = stpio_request_set_pin (7, 1, "lanrst", STPIO_OUT, 1);
    stpio_free_pin (lanrst_pin);

    /* SPI */
    spi_register_board_info(dolmenb_spi_boot_flash,
                            ARRAY_SIZE(dolmenb_spi_boot_flash));
    /* I2C */
    stx7105_configure_ssc(&ssc_private_info);

    /* USB card reader (nRST_READER) - set in reset state */
    reader_pin = stpio_request_set_pin(4, 2, "usb_creader_reset", STPIO_OUT, 0);
    /* USB hub (nRST_USBHUB) */
    hub_pin = stpio_request_set_pin(4, 3, "usb_hub_reset", STPIO_OUT, 0);
    udelay(2000);
    stpio_set_pin (hub_pin, 1);
    stpio_free_pin (hub_pin);
    /* USB card reader (nRST_READER) - release reset state */
    stpio_set_pin (reader_pin, 1);
    stpio_free_pin (reader_pin);
    udelay(2000);

    /* usb */
    stx7105_configure_usb(0, &usb_init[0]);
    stx7105_configure_usb(1, &usb_init[1]);

    /* lirc */
    stx7105_configure_lirc(NULL, lirc_enabled_pios,
                   ARRAY_SIZE(lirc_enabled_pios));

    /* ethernet */
    stx7105_configure_ethernet(0, stx7105_ethernet_mii, 1, 0, 0, 0);

    /* audio (AUDIO_EN) - enable analogue output on startup */
    audio_pin = stpio_request_set_pin(6, 5, "audio_en", STPIO_OUT, 1);

    stpio_free_pin (audio_pin);

    /* audio */
    stx7105_configure_audio_pins(1, 1, 1, 0);

    /* flash write protect (nFLASH_WP) */
    /*
     * FLASH_WP is used by NAND. Since MTD NAND has no
     * concept of WP/VPP, we must permanently enable it
     */
    flashwp_pin = stpio_request_set_pin(6, 4, "FLASH_WP", STPIO_OUT, 1);
    stpio_free_pin (flashwp_pin);

#if defined(CONFIG_MTD_NAND)
    stx7105_configure_nand(&nand_device);
#endif

    return platform_add_devices(dolmenb_devices, ARRAY_SIZE(dolmenb_devices));
}
arch_initcall(device_init);

static void __iomem *dolmenb_ioport_map(unsigned long port, unsigned int size)
{
	/* However picking somewhere safe isn't as easy as you might think.
	 * I used to use external ROM, but that can cause problems if you are
	 * in the middle of updating Flash. So I'm now using the processor core
	 * version register, which is guaranted to be available, and non-writable.
	 */
	return (void __iomem *)CCN_PVR;
}

static void __init dolmenb_init_irq(void)
{
}

struct sh_machine_vector mv_dolmenb __initmv = {
	.mv_name		= "dolmenb",
	.mv_setup		= dolmenb_setup,
	.mv_nr_irqs		= NR_IRQS,
	.mv_init_irq		= dolmenb_init_irq,
	.mv_ioport_map		= dolmenb_ioport_map,
};
