/*
 * Copyright (C) 2009 STMicroelectronics Limited
 *
 * May be copied or modified under the terms of the GNU General Public
 * License.  See linux/COPYING for more information.
 *
 * Code to handle the clockgen hardware on the Freeman 510 (FLI7510).
 */

#include <linux/err.h>
#include <linux/io.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/pm.h>
#include <linux/stm/sysconf.h>
#include <linux/stm/clk.h>



/* 30 MHz quartz oscillator clock ----------------------------------------- */

static struct clk fli7510_osc_clk = {
	.name	= "osc_clk",
	.flags	= CLK_ALWAYS_ENABLED | CLK_RATE_PROPAGATES,
	.rate	= 30000000,
};



/* Clockgen A ------------------------------------------------------------- */

#define FLI7510_CLOCKGENA_BASE_ADDR     	0xfde00000
#define FLI7510_CKGA_PLL0_CFG			0x000
#define FLI7510_CKGA_PLL1_CFG			0x004
#define FLI7510_CKGA_CLKOPSRC_SWITCH_CFG(n)	(0x014 + ((n) * 0x10))
#define FLI7510_CKGA_OSC_DIV_CFG(n)		(0x800 + ((n) * 4))
#define FLI7510_CKGA_PLL0HS_DIV_CFG(x)		(0x900 + ((x) * 4))
#define FLI7510_CKGA_PLL0LS_DIV_CFG(x)		(0xa10 + (((x) -4) *4))
#define FLI7510_CKGA_PLL1_DIV_CFG(x)		(0xb00 + ((x) *4))

static void __iomem *fli7510_clkgena_base;

/* PLLs */

static unsigned long pll1600_freq(unsigned long input, unsigned long cfg)
{
	unsigned long freq, ndiv, mdiv;

	mdiv = (cfg >>  0) & 0x7;
	ndiv = (cfg >>  8) & 0xff;
	freq = (((input / 1000) * ndiv) / mdiv) * 1000;

	return freq;
}

static unsigned long pll800_freq(unsigned long input, unsigned long cfg)
{
	unsigned long freq, ndiv, pdiv, mdiv;

	mdiv = (cfg >>  0) & 0xff;
	ndiv = (cfg >>  8) & 0xff;
	pdiv = (cfg >> 16) & 0x7;
	freq = (((2 * (input / 1000) * ndiv) / mdiv) / (1 << pdiv)) * 1000;

	return freq;
}

static int fli7510_clkgena_pll_clk_init(struct clk *clk)
{
	clk->ops->recalc(clk);

	return 0;
}

static void fli7510_clkgena_pll_clk_recalc(struct clk *clk)
{
	int pll_num = (int)clk->private_data;
	unsigned long cfg;

	switch (pll_num) {
	case 0:
		cfg = readl(fli7510_clkgena_base + FLI7510_CKGA_PLL0_CFG);
		clk->rate = pll1600_freq(clk->parent->rate, cfg);
		break;
	case 1:
		cfg = readl(fli7510_clkgena_base + FLI7510_CKGA_PLL1_CFG);
		clk->rate = pll800_freq(clk->parent->rate, cfg);
		break;
	default:
		BUG();
		break;
	}
}

static struct clk_ops fli7510_clkgena_pll_clk_ops = {
	.init	= fli7510_clkgena_pll_clk_init,
	.recalc	= fli7510_clkgena_pll_clk_recalc,
};

static struct clk fli7510_clkgena_pll_clks[] = {
	{
		.name		= "clkgena_pll0_clk",
		.flags		= CLK_ALWAYS_ENABLED | CLK_RATE_PROPAGATES,
		.parent		= &fli7510_osc_clk,
		.ops		= &fli7510_clkgena_pll_clk_ops,
		.private_data	= (void *)0,
	}, {
		.name		= "clkgena_pll1_clk",
		.flags		= CLK_ALWAYS_ENABLED | CLK_RATE_PROPAGATES,
		.parent		= &fli7510_osc_clk,
		.ops		= &fli7510_clkgena_pll_clk_ops,
		.private_data	= (void *)1,
	},
};

/* Generated clocks */

enum fli7510_clkgena_ids {
	CLK_SLIM_FDMA = 0,
	CLK_VCPU,
	CLK_XP70_TANGO,
	CLK_IC_266,
	CLK_ST40_HOST_ICK,
	CLK_ST40_CTL_ICK,
	CLK_ST231_DRA2,
	CLK_ST231_AUD,
	CLK_BLT,
	CLK_XP70_AATV,
	CLK_EMI,
	CLK_PP,
	CLK_ETHERNET,
	CLK_PCI,
	CLK_IC_100,
	CLK_IC_150,
	CLK_SPARE_1,
	CLK_IC_200,
	CLK_SPARE_2,
};

static int fli7510_clkgena_clk_init(struct clk *clk)
{
	int num = (int)clk->private_data;
	unsigned long data;
	unsigned long src_sel;

	data = readl(fli7510_clkgena_base +
			FLI7510_CKGA_CLKOPSRC_SWITCH_CFG(num / 16));
	src_sel = (data >> ((num % 16) * 2)) & 0x3;

	switch (src_sel) {
	case 0:
		clk->parent = &fli7510_osc_clk;
		break;
	case 1:
	case 2:
		clk->parent = &fli7510_clkgena_pll_clks[src_sel - 1];
		break;
	case 3:
		/* clock is stopped */
		clk->parent = NULL;
		break;
	}

	clk->ops->recalc(clk);

	return 0;
}

static void fli7510_clkgena_clk_recalc(struct clk *clk)
{
	int num = (int)clk->private_data;
	unsigned long data;
	unsigned long src_sel;
	unsigned long div_cfg = 0;
	unsigned long ratio;

	data = readl(fli7510_clkgena_base +
			FLI7510_CKGA_CLKOPSRC_SWITCH_CFG(num / 16));
	src_sel = (data >> ((num % 16) * 2)) & 0x3;

	switch (src_sel) {
	case 0:
		div_cfg = readl(fli7510_clkgena_base +
				FLI7510_CKGA_OSC_DIV_CFG(num));
		break;
	case 1:
		div_cfg = readl(fli7510_clkgena_base + ((num <= 3) ?
					FLI7510_CKGA_PLL0HS_DIV_CFG(num) :
					FLI7510_CKGA_PLL0LS_DIV_CFG(num)));
		break;
	case 2:
		div_cfg = readl(fli7510_clkgena_base +
				FLI7510_CKGA_PLL1_DIV_CFG(num));
		break;
	case 3:
		clk->rate = 0;
		return;
	}

	if (div_cfg & 0x10000)
		ratio = 1;
	else
		ratio = (div_cfg & 0x1f) + 1;

	clk->rate = clk->parent->rate / ratio;
}

static struct clk_ops fli7510_clkgena_clk_ops = {
	.init		= fli7510_clkgena_clk_init,
	.recalc		= fli7510_clkgena_clk_recalc,
};

#define FLI7510_CLKGENA_CLK(_num, _name) \
	{ \
		.name		= _name, \
		.flags		= CLK_ALWAYS_ENABLED | CLK_RATE_PROPAGATES, \
		.ops		= &fli7510_clkgena_clk_ops, \
		.private_data	= (void *)_num, \
		.id		= _num, \
	}

struct clk fli7510_clkgena_clks[] = {
	FLI7510_CLKGENA_CLK(CLK_SLIM_FDMA, "slim_fdma"),
	FLI7510_CLKGENA_CLK(CLK_VCPU, "vcpu"),
	FLI7510_CLKGENA_CLK(CLK_XP70_TANGO, "xp70_tango"),
	FLI7510_CLKGENA_CLK(CLK_IC_266, "ic_266"),
	FLI7510_CLKGENA_CLK(CLK_ST40_HOST_ICK, "st40_host_ick"),
	FLI7510_CLKGENA_CLK(CLK_ST40_CTL_ICK, "st40_ctl_ick"),
	FLI7510_CLKGENA_CLK(CLK_ST231_DRA2, "st231_dra2"),
	FLI7510_CLKGENA_CLK(CLK_ST231_AUD, "st241_aud"),
	FLI7510_CLKGENA_CLK(CLK_BLT, "blt"),
	FLI7510_CLKGENA_CLK(CLK_XP70_AATV, "xp70_aatv"),
	FLI7510_CLKGENA_CLK(CLK_EMI, "emi"),
	FLI7510_CLKGENA_CLK(CLK_PP, "pp"),
	FLI7510_CLKGENA_CLK(CLK_ETHERNET, "ethernet"),
	FLI7510_CLKGENA_CLK(CLK_PCI, "pci"),
	FLI7510_CLKGENA_CLK(CLK_IC_100, "ic_100"),
	FLI7510_CLKGENA_CLK(CLK_IC_150, "ic_150"),
	FLI7510_CLKGENA_CLK(CLK_IC_200, "ic_200"),
};

#ifdef CONFIG_PM
void fli7510_clkgena_pm_restore(void)
{
	for (i = 0; i < ARRAY_SIZE(fli7510_clkgena_pll_clks); i++)
		fli7510_clkgena_pll_clk_init(&fli7510_clkgena_pll_clks[i]);

	for (i = 0; i < ARRAY_SIZE(fli7510_clkgena_clks); i++) {
		if (fli7510_clkgena_clk_setrate(&fli7510_clkgena_clks[i],
				fli7510_clkgena_clks[i].rate) < 0)
			fli7510_clkgena_clk_recalc(&fli7510_clkgena_clks[i]);
	}
}

int clk_pm_state(pm_message_t state)
{
	static int prev_state = PM_EVENT_ON;
	int i;

	switch (state.event) {
	case PM_EVENT_ON:
		if (prev_state == PM_EVENT_FREEZE)
			fli7510_clkgena_pm_restore();
	case PM_EVENT_SUSPEND:
	case PM_EVENT_FREEZE:
		prev_state = state.event;
		break;
	}
	return 0;
}
#endif



/* SH4 generic clocks ----------------------------------------------------- */

static int fli7510_generic_clk_init(struct clk *clk)
{
	clk->parent = clk_get(NULL, "ic_100");
	BUG_ON(!clk->parent || IS_ERR(clk->parent));

	clk->ops->recalc(clk);

	return 0;
}

static void fli7510_generic_clk_recalc(struct clk *clk)
{
	clk->rate = clk->parent->rate;
}

static struct clk_ops fli7510_generic_clk_ops = {
	.init		= fli7510_generic_clk_init,
	.recalc		= fli7510_generic_clk_recalc,
};

static struct clk fli7510_generic_module_clk = {
	.name		= "module_clk",
	.flags		= CLK_ALWAYS_ENABLED | CLK_RATE_PROPAGATES,
	.ops		= &fli7510_generic_clk_ops,
};

static struct clk fli7510_generic_comms_clk = {
	.name		= "comms_clk",
	.flags		= CLK_ALWAYS_ENABLED,
	.ops		= &fli7510_generic_clk_ops,
};



/* ------------------------------------------------------------------------ */

int __init clk_init(void)
{
	int err;
	int i;

	/* Clockgen A */

	fli7510_clkgena_base = ioremap(FLI7510_CLOCKGENA_BASE_ADDR, 0x50);
	BUG_ON(!fli7510_clkgena_base);

	err = clk_register(&fli7510_osc_clk);
	if (err != 0)
		goto error;

	for (i = 0; i < ARRAY_SIZE(fli7510_clkgena_pll_clks); i++) {
		err = clk_register(&fli7510_clkgena_pll_clks[i]);
		if (err != 0)
			goto error;
	}

	for (i = 0; i < ARRAY_SIZE(fli7510_clkgena_clks); i++) {
		err = clk_register(&fli7510_clkgena_clks[i]);
		if (err != 0)
			goto error;
	}

	/* Generic SH-4 clocks */

	err = clk_register(&fli7510_generic_module_clk);
	if (err != 0)
		goto error;

	err = clk_register(&fli7510_generic_comms_clk);

error:
	return err;
}
