/*
 * Freeman 510 (FLI7510) Setup
 *
 * Copyright (C) 2009 STMicroelectronics Limited
 * Author: Pawel Moll <pawel.moll@st.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/gpio.h>
#include <linux/i2c.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/pata_platform.h>
#include <linux/phy.h>
#include <linux/platform_device.h>
#include <linux/serial.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <linux/stm/emi.h>
#include <linux/stm/pio.h>
#include <linux/stm/soc.h>
#include <linux/stm/soc_init.h>
#include <linux/stm/sysconf.h>
#include <asm/irq-ilc.h>
#include <asm/processor.h>



/* Returns: 1 if being executed on the "HOST" ST40,
 *          0 if executed on the "RT" ST40 */
#define ST40HOST_CORE ((ctrl_inl(CCN_PRR) & (1 << 7)) != 0)



/* USB resources ---------------------------------------------------------- */

#define AHB2STBUS_WRAPPER_GLUE_OFFSET	0x00000
#define AHB2STBUS_OHCI_OFFSET		0xffc00
#define AHB2STBUS_EHCI_OFFSET		0xffe00
#define AHB2STBUS_PROTOCOL_OFFSET	0xfff00

static u64 st40_dma_mask = DMA_32BIT_MASK;

static struct platform_device fli7510_usb_device =
	USB_DEVICE(0, 0xfda00000 + AHB2STBUS_EHCI_OFFSET, ILC_IRQ(54),
			0xfda00000 + AHB2STBUS_OHCI_OFFSET, ILC_IRQ(55),
			0xfda00000 + AHB2STBUS_WRAPPER_GLUE_OFFSET,
			0xfda00000 + AHB2STBUS_PROTOCOL_OFFSET,
			USB_FLAGS_STRAP_16BIT |
			USB_FLAGS_STRAP_PLL |
			USB_FLAGS_STBUS_CONFIG_THRESHOLD256);

void __init fli7510_configure_usb(enum fli7510_usb_ovrcur_mode ovrcur_mode)
{
	struct sysconf_field *sc;

	sc = sysconf_claim(CFG_COMMS_CONFIG_1, 12, 12,
			"usba_enable_pad_override");
	switch (ovrcur_mode) {
	case fli7510_usb_ovrcur_disabled:
		sysconf_write(sc, 1);
		sc = sysconf_claim(CFG_COMMS_CONFIG_1, 13, 13, "usba_ovrcur");
		sysconf_write(sc, 1);
		break;
	default:
		sysconf_write(sc, 0);
		sc = sysconf_claim(CFG_COMMS_CONFIG_1, 11, 11,
				"usba_ovrcur_polarity");
		switch (ovrcur_mode) {
		case fli7510_usb_ovrcur_active_high:
			sysconf_write(sc, 0);
			break;
		case fli7510_usb_ovrcur_active_low:
			sysconf_write(sc, 1);
			break;
		default:
			BUG();
			break;
		}
		break;
	}

	stpio_request_pin(27, 1, "USB_A_OVRCUR", STPIO_IN);
	stpio_request_pin(27, 2, "USB_A_PWREN", STPIO_ALT_OUT);

	platform_device_register(&fli7510_usb_device);
}



/* FDMA resources --------------------------------------------------------- */

#ifdef CONFIG_STM_DMA

#include <linux/stm/fdma_firmware_7200.h>

static struct stm_plat_fdma_hw fli7510_fdma_hw = {
	.slim_regs = {
		.id       = 0x0000 + (0x000 << 2), /* 0x0000 */
		.ver      = 0x0000 + (0x001 << 2), /* 0x0004 */
		.en       = 0x0000 + (0x002 << 2), /* 0x0008 */
		.clk_gate = 0x0000 + (0x003 << 2), /* 0x000c */
	},
	.periph_regs = {
		.sync_reg = 0x8000 + (0xfe2 << 2), /* 0xbf88 */
		.cmd_sta  = 0x8000 + (0xff0 << 2), /* 0xbfc0 */
		.cmd_set  = 0x8000 + (0xff1 << 2), /* 0xbfc4 */
		.cmd_clr  = 0x8000 + (0xff2 << 2), /* 0xbfc8 */
		.cmd_mask = 0x8000 + (0xff3 << 2), /* 0xbfcc */
		.int_sta  = 0x8000 + (0xff4 << 2), /* 0xbfd0 */
		.int_set  = 0x8000 + (0xff5 << 2), /* 0xbfd4 */
		.int_clr  = 0x8000 + (0xff6 << 2), /* 0xbfd8 */
		.int_mask = 0x8000 + (0xff7 << 2), /* 0xbfdc */
	},
	.dmem_offset = 0x8000,
	.dmem_size   = 0x800 << 2, /* 2048 * 4 = 8192 */
	.imem_offset = 0xc000,
	.imem_size   = 0x1000 << 2, /* 4096 * 4 = 16384 */
};

static struct stm_plat_fdma_data fli7510_fdma_platform_data = {
	.hw = &fli7510_fdma_hw,
	.fw = &stm_fdma_firmware_7200,
	.min_ch_num = CONFIG_MIN_STM_DMA_CHANNEL_NR,
	.max_ch_num = CONFIG_MAX_STM_DMA_CHANNEL_NR,
};

#define fli7510_fdma_platform_data_addr &fli7510_fdma_platform_data

#else

#define fli7510_fdma_platform_data_addr NULL

#endif /* CONFIG_STM_DMA */

static struct platform_device fli7510_fdma_devices[] = {
	{
		.name		= "stm-fdma",
		.id		= 0,
		.num_resources	= 2,
		.resource = (struct resource[]) {
			{
				.start = 0xfd660000,
				.end   = 0xfd66ffff,
				.flags = IORESOURCE_MEM,
			}, {
				.start = ILC_IRQ(36),
				.end   = ILC_IRQ(36),
				.flags = IORESOURCE_IRQ,
			},
		},
		.dev.platform_data = fli7510_fdma_platform_data_addr,
	}, {
		.name		= "stm-fdma",
		.id		= 1,
		.num_resources	= 2,
		.resource = (struct resource[]) {
			{
				.start = 0xfd910000,
				.end   = 0xfd91ffff,
				.flags = IORESOURCE_MEM,
			}, {
				.start = ILC_IRQ(38),
				.end   = ILC_IRQ(38),
				.flags = IORESOURCE_IRQ,
			},
		},
		.dev.platform_data = fli7510_fdma_platform_data_addr,
	}
};

static struct platform_device fli7510_fdma_xbar_device = {
	.name		= "fdma-xbar",
	.id		= -1,
	.num_resources	= 1,
	.resource	= (struct resource[]) {
		{
			.start	= 0xfd980000,
			.end	= 0xfd980fff,
			.flags	= IORESOURCE_MEM,
		},
	},
};



/* SSC resources ---------------------------------------------------------- */

static char fli7510_i2c_dev_name[] = "i2c_st";
static char fli7510_spi_dev_name[] = "spi_st_ssc";

static struct platform_device fli7510_ssc_devices[] = {
	STSSC_DEVICE(0xfdb40000, ILC_IRQ(19), 10, 2, 3, 0xff),
	STSSC_DEVICE(0xfdb41000, ILC_IRQ(20), 9, 4, 5, 0xff),
	STSSC_DEVICE(0xfdb42000, ILC_IRQ(21), 9, 6, 7, 0xff),
	STSSC_DEVICE(0xfdb43000, ILC_IRQ(22), 10, 0, 1, 0xff),
	STSSC_DEVICE(0xfdb44000, ILC_IRQ(23), 17, 2, 3, 5),
};

void __init fli7510_configure_ssc(struct plat_ssc_data *data)
{
	int num_i2c = 0;
	int num_spi = 0;
	int i;
	int capability = data->capability;

	for (i = 0; i < ARRAY_SIZE(fli7510_ssc_devices); i++,
			capability >>= SSC_BITS_SIZE) {
		struct ssc_pio_t *ssc =
				fli7510_ssc_devices[i].dev.platform_data;

		if (capability & SSC_UNCONFIGURED)
			continue;

		if (capability & SSC_SPI_CAPABILITY) {
			fli7510_ssc_devices[i].name = fli7510_spi_dev_name;
			fli7510_ssc_devices[i].id = num_spi++;
			ssc->chipselect = data->spi_chipselects[i];
		} else {
			/* I2C bus number reservation (to prevent any hot-plug
			 * device from using it) */
#ifdef CONFIG_I2C_BOARDINFO
			i2c_register_board_info(num_i2c, NULL, 0);
#endif
			fli7510_ssc_devices[i].name = fli7510_i2c_dev_name;
			fli7510_ssc_devices[i].id = num_i2c++;
			if (capability & SSC_I2C_CLK_UNIDIR)
				ssc->clk_unidir = 1;
		}

		platform_device_register(&fli7510_ssc_devices[i]);
	}
}



/* PATA resources --------------------------------------------------------- */

/*
 * EMI A20 = CS1 (active low)
 * EMI A21 = CS0 (active low)
 * EMI A19 = DA2
 * EMI A18 = DA1
 * EMI A17 = DA0
 */

static struct resource pata_resources[] = {
	[0] = {	/* I/O base: CS1=N, CS0=A */
		.start	= (1<<20),
		.end	= (1<<20) + (8<<17)-1,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {	/* CTL base: CS1=A, CS0=N, DA2=A, DA1=A, DA0=N */
		.start	= (1<<21) + (6<<17),
		.end	= (1<<21) + (6<<17) + 3,
		.flags	= IORESOURCE_MEM,
	},
	[2] = {	/* IRQ */
		.flags	= IORESOURCE_IRQ,
	}
};

static struct pata_platform_info pata_info = {
	.ioport_shift	= 17,
};

static struct platform_device pata_device = {
	.name		= "pata_platform",
	.id		= -1,
	.num_resources	= ARRAY_SIZE(pata_resources),
	.resource	= pata_resources,
	.dev = {
		.platform_data = &pata_info,
	}
};

void __init fli7510_configure_pata(int bank, int pc_mode, int irq)
{
	unsigned long bank_base;

	bank_base = emi_bank_base(bank);
	pata_resources[0].start += bank_base;
	pata_resources[0].end   += bank_base;
	pata_resources[1].start += bank_base;
	pata_resources[1].end   += bank_base;
	pata_resources[2].start = irq;
	pata_resources[2].end   = irq;

	emi_config_pata(bank, pc_mode);

	platform_device_register(&pata_device);
}



/* Ethernet MAC resources ------------------------------------------------- */

static void fli7510_gmac_fix_speed(void *priv, unsigned int speed)
{
	struct sysconf_field *sc = priv;

	sysconf_write(sc, (speed == SPEED_100) ? 1 : 0);
}

static struct plat_stmmacenet_data fli7510_gmac_private_data = {
	.bus_id = 0,
	.pbl = 32,
	.has_gmac = 1,
	.fix_mac_speed = fli7510_gmac_fix_speed,
};

static struct platform_device fli7510_gmac_device = {
	.name 		= "stmmaceth",
	.id		= 1,
	.num_resources	= 2,
	.resource	= (struct resource[]) {
		{
			.start	= 0xfd920000,
			.end	= 0xfd927fff,
			.flags	= IORESOURCE_MEM,
		}, {
			.name	= "macirq",
			.start	= ILC_IRQ(40),
			.end	= ILC_IRQ(40),
			.flags	= IORESOURCE_IRQ,
		},
	},
	.dev = {
		.power.can_wakeup = 1,
		.platform_data = &fli7510_gmac_private_data,
	}
};

struct fli7510_gmac_pin {
	unsigned char port, pin, dir;
};

static struct fli7510_gmac_pin fli7510_gmac_mii_pins[] __initdata = {
	{ 18, 5, },			/* PHYCLK */
	{ 18, 0, STPIO_ALT_OUT },	/* MDC */
	{ 18, 1, STPIO_IN },		/* COL */
	{ 18, 2, STPIO_IN },		/* CRS */
	{ 18, 3, STPIO_IN },		/* MDINT */
	{ 18, 4, STPIO_ALT_BIDIR },	/* MDIO */
	{ 20, 0, STPIO_ALT_OUT },	/* TXD[0] */
	{ 20, 1, STPIO_ALT_OUT },	/* TXD[1] */
	{ 20, 2, STPIO_ALT_OUT },	/* TXD[2] */
	{ 20, 3, STPIO_ALT_OUT },	/* TXD[3] */
	{ 20, 4, STPIO_ALT_OUT },	/* TXEN */
	{ 20, 6, STPIO_IN },		/* TXCLK */
	{ 21, 0, STPIO_IN },		/* RXD[0] */
	{ 21, 1, STPIO_IN },		/* RXD[1] */
	{ 21, 2, STPIO_IN },		/* RXD[2] */
	{ 21, 3, STPIO_IN },		/* RXD[3] */
	{ 21, 4, STPIO_IN },		/* RXDV */
	{ 21, 5, STPIO_IN },		/* RX_ER */
	{ 21, 6, STPIO_IN },		/* RXCLK */
};

static struct fli7510_gmac_pin fli7510_gmac_gmii_pins[] __initdata = {
	{ 18, 5, },			/* PHYCLK */
	{ 18, 0, STPIO_ALT_OUT },	/* MDC */
	{ 18, 1, STPIO_IN },		/* COL */
	{ 18, 2, STPIO_IN },		/* CRS */
	{ 18, 3, STPIO_IN },		/* MDINT */
	{ 18, 4, STPIO_ALT_BIDIR },	/* MDIO */
	{ 20, 0, STPIO_ALT_OUT },	/* TXD[0] */
	{ 20, 1, STPIO_ALT_OUT },	/* TXD[1] */
	{ 20, 2, STPIO_ALT_OUT },	/* TXD[2] */
	{ 20, 3, STPIO_ALT_OUT },	/* TXD[3] */
	{ 24, 4, STPIO_ALT_OUT },	/* TXD[4] */
	{ 24, 5, STPIO_ALT_OUT },	/* TXD[5] */
	{ 24, 6, STPIO_ALT_OUT },	/* TXD[6] */
	{ 24, 7, STPIO_ALT_OUT },	/* TXD[7] */
	{ 20, 4, STPIO_ALT_OUT },	/* TXEN */
	{ 20, 6, STPIO_IN },		/* TXCLK */
	{ 21, 0, STPIO_IN },		/* RXD[0] */
	{ 21, 1, STPIO_IN },		/* RXD[1] */
	{ 21, 2, STPIO_IN },		/* RXD[2] */
	{ 21, 3, STPIO_IN },		/* RXD[3] */
	{ 24, 0, STPIO_IN },		/* RXD[4] */
	{ 24, 1, STPIO_IN },		/* RXD[5] */
	{ 24, 2, STPIO_IN },		/* RXD[6] */
	{ 24, 3, STPIO_IN },		/* RXD[7] */
	{ 21, 4, STPIO_IN },		/* RXDV */
	{ 21, 5, STPIO_IN },		/* RX_ER */
	{ 21, 6, STPIO_IN },		/* RXCLK */
};

static struct fli7510_gmac_pin fli7510_gmac_rmii_pins[] __initdata = {
	{ 18, 5, },			/* PHYCLK */
	{ 18, 0, STPIO_ALT_OUT },	/* MDC */
	{ 18, 3, STPIO_IN },		/* MDINT */
	{ 18, 4, STPIO_ALT_BIDIR },	/* MDIO */
	{ 20, 0, STPIO_ALT_OUT },	/* TXD[0] */
	{ 20, 1, STPIO_ALT_OUT },	/* TXD[1] */
	{ 20, 4, STPIO_ALT_OUT },	/* TXEN */
	{ 21, 0, STPIO_IN },		/* RXD[0] */
	{ 21, 1, STPIO_IN },		/* RXD[1] */
	{ 21, 4, STPIO_IN },		/* RXDV */
	{ 21, 5, STPIO_IN },		/* RX_ER */
};

static struct fli7510_gmac_pin fli7510_gmac_reverse_mii_pins[] __initdata = {
	{ 18, 5, },			/* PHYCLK */
	{ 18, 0, STPIO_IN },		/* MDC */
	{ 18, 1, STPIO_ALT_OUT },	/* COL */
	{ 18, 2, STPIO_ALT_OUT },	/* CRS */
	{ 18, 3, STPIO_IN },		/* MDINT */
	{ 18, 4, STPIO_ALT_BIDIR },	/* MDIO */
	{ 20, 0, STPIO_ALT_OUT },	/* TXD[0] */
	{ 20, 1, STPIO_ALT_OUT },	/* TXD[1] */
	{ 20, 2, STPIO_ALT_OUT },	/* TXD[2] */
	{ 20, 3, STPIO_ALT_OUT },	/* TXD[3] */
	{ 20, 4, STPIO_ALT_OUT },	/* TXEN */
	{ 20, 6, STPIO_IN },		/* TXCLK */
	{ 21, 0, STPIO_IN },		/* RXD[0] */
	{ 21, 1, STPIO_IN },		/* RXD[1] */
	{ 21, 2, STPIO_IN },		/* RXD[2] */
	{ 21, 3, STPIO_IN },		/* RXD[3] */
	{ 21, 4, STPIO_IN },		/* RXDV */
	{ 21, 5, STPIO_IN },		/* RX_ER */
	{ 21, 6, STPIO_IN },		/* RXCLK */
};

void __init fli7510_configure_ethernet(enum fli7510_ethernet_mode mode,
		int ext_clk, int phy_bus)
{
	struct sysconf_field *sc;
	struct fli7510_gmac_pin *pins;
	int pins_num;
	unsigned char phy_sel, enmii;
	int i;

	sc = sysconf_claim(CFG_COMMS_CONFIG_2, 24, 24, "gmac_enable");
	sysconf_write(sc, 1);

	switch (mode) {
	case fli7510_ethernet_mii:
		phy_sel = 0;
		enmii = 1;
		pins = fli7510_gmac_mii_pins;
		pins_num = ARRAY_SIZE(fli7510_gmac_mii_pins);
		break;
	case fli7510_ethernet_rmii:
		phy_sel = 4;
		enmii = 1;
		pins = fli7510_gmac_rmii_pins;
		pins_num = ARRAY_SIZE(fli7510_gmac_rmii_pins);
		break;
	case fli7510_ethernet_gmii:
		phy_sel = 0;
		enmii = 1;
		pins = fli7510_gmac_gmii_pins;
		pins_num = ARRAY_SIZE(fli7510_gmac_gmii_pins);
		sc = sysconf_claim(CFG_COMMS_CONFIG_1, 17, 18,
				"conf_pio24_alternate");
		sysconf_write(sc, 2);
		break;
	case fli7510_ethernet_reverse_mii:
		phy_sel = 0;
		enmii = 0;
		pins = fli7510_gmac_reverse_mii_pins;
		pins_num = ARRAY_SIZE(fli7510_gmac_reverse_mii_pins);
		break;
	default:
		BUG();
		return;
	}

	fli7510_gmac_private_data.bsp_priv = sysconf_claim(CFG_COMMS_CONFIG_2,
			25, 25, "gmac_mac_speed");

	sc = sysconf_claim(CFG_COMMS_CONFIG_2, 26, 28, "phy_intf_sel");
	sysconf_write(sc, phy_sel);

	sc = sysconf_claim(CFG_COMMS_CONFIG_2, 8, 8, "gmac_mii_enable");
	sysconf_write(sc, enmii);

	sc = sysconf_claim(CFG_COMMS_CONFIG_2, 9, 9, "gmac_phy_clock_sel");
	if (ext_clk) {
		pins[0].dir = STPIO_IN;
		sysconf_write(sc, 1);
	} else {
		pins[0].dir = STPIO_ALT_OUT;
		sysconf_write(sc, 0);
	}

	for (i = 0; i < pins_num; i++)
		stpio_request_pin(pins[i].port, pins[i].pin, "stmmac",
				pins[i].dir);

	fli7510_gmac_private_data.bus_id = phy_bus;

	platform_device_register(&fli7510_gmac_device);
}



/* PWM resources ---------------------------------------------------------- */

static struct platform_device fli7510_pwm_device = {
	.name		= "stm-pwm",
	.id		= -1,
	.num_resources	= 2,
	.resource	= (struct resource[]) {
		{
			.start	= 0xfdb10000,
			.end	= 0xfdb10067,
			.flags	= IORESOURCE_MEM
		}, {
			.start	= ILC_IRQ(27),
			.end	= ILC_IRQ(27),
			.flags	= IORESOURCE_IRQ
		}
	},
};

void fli7510_configure_pwm(struct plat_stm_pwm_data *data)
{
	int pwm;

	fli7510_pwm_device.dev.platform_data = data;

	for (pwm = 0; pwm < 4; pwm++)
		if (data->flags & (1 << pwm))
			stpio_request_pin(8, 4 + pwm, "pwm", STPIO_ALT_OUT);

	platform_device_register(&fli7510_pwm_device);
}



/* Hardware RNG resources ------------------------------------------------- */

static struct platform_device fli7510_devhwrandom_device = {
	.name		= "stm_hwrandom",
	.id		= -1,
	.num_resources	= 1,
	.resource	= (struct resource []) {
		{
			.start  = 0xfd3e0000,
			.end    = 0xfd3e0027,
			.flags  = IORESOURCE_MEM
		},
	}
};

static struct platform_device fli7510_devrandom_device = {
	.name		= "stm_rng",
	.id		= 0,
	.num_resources	= 1,
	.resource	= (struct resource []) {
		{
			.start  = 0xfd3e0000,
			.end    = 0xfd3e0027,
			.flags  = IORESOURCE_MEM
		},
	}
};



/* ASC resources ---------------------------------------------------------- */

static struct platform_device fli7510_asc_devices[] = {
	STASC_DEVICE(0xfdb30000, ILC_IRQ(24), 5, 6,
			9, 3, 2, 1, 0,
			STPIO_ALT_OUT, STPIO_IN, STPIO_IN, STPIO_ALT_OUT),
	STASC_DEVICE(0xfdb31000, ILC_IRQ(25), 7, 8,
			25, 5, 4, 3, 2,
			STPIO_ALT_OUT, STPIO_IN, STPIO_IN, STPIO_ALT_OUT),
	STASC_DEVICE(0xfdb32000, ILC_IRQ(26), 9, 10,
			25, 7, 6, 0xff, 0xff,
			STPIO_ALT_OUT, STPIO_IN, STPIO_IN, STPIO_ALT_OUT),
};

/*
 * Note these three variables are global, and shared with the stasc driver
 * for console bring up prior to platform initialisation.
 */

/* the serial console device */
int stasc_console_device __initdata;

/* Platform devices to register */
struct platform_device *stasc_configured_devices[
		ARRAY_SIZE(fli7510_asc_devices)] __initdata;
unsigned int stasc_configured_devices_count __initdata = 0;

/* Configure the ASC's for this board.
 * This has to be called before console_init().
 */
void __init fli7510_configure_asc(const int *ascs, int num_ascs, int console)
{
	int i;

	for (i = 0; i < num_ascs; i++) {
		int port;
		unsigned char flags;
		struct platform_device *pdev;
		struct stasc_uart_data *uart_data;

		port = ascs[i] & 0xff;
		flags = ascs[i] >> 8;
		pdev = &fli7510_asc_devices[port];
		uart_data = pdev->dev.platform_data;

		pdev->id = i;
		((struct stasc_uart_data *)(pdev->dev.platform_data))->flags =
				flags | STASC_FLAG_TXFIFO_BUG;
		stasc_configured_devices[stasc_configured_devices_count] = pdev;
		stasc_configured_devices_count++;
	}

	stasc_console_device = console;
	/* the console will be always a wakeup-able device */
	stasc_configured_devices[console]->dev.power.can_wakeup = 1;
	device_set_wakeup_enable(&stasc_configured_devices[console]->dev, 0x1);
}

/* Add platform device as configured by board specific code */
static int __init fli7510_add_asc(void)
{
	return platform_add_devices(stasc_configured_devices,
				    stasc_configured_devices_count);
}
arch_initcall(fli7510_add_asc);



/* LiRC resources --------------------------------------------------------- */

static struct lirc_pio fli7510_lirc_pios[] = {
	{
		.bank  = 26,
		.pin   = 2,
		.pinof = 0x00 | LIRC_IR_RX | LIRC_PIO_ON,
		.dir   = STPIO_IN,
	},
};

static struct plat_lirc_data fli7510_lirc_private_info = {
	/* The clock settings will be calculated by the driver
	 * from the system clock */
	.irbclock	= 0, /* use current_cpu data */
	.irbclkdiv	= 0, /* automatically calculate */
	.irbperiodmult	= 0,
	.irbperioddiv	= 0,
	.irbontimemult	= 0,
	.irbontimediv	= 0,
	.irbrxmaxperiod = 0x5000,
	.sysclkdiv	= 1,
	.rxpolarity	= 1,
	.pio_pin_arr  = fli7510_lirc_pios,
	.num_pio_pins = ARRAY_SIZE(fli7510_lirc_pios),
#ifdef CONFIG_PM
	.clk_on_low_power = 1000000,
#endif
};

static struct platform_device fli7510_lirc_device = {
	.name		= "lirc",
	.id		= -1,
	.num_resources	= 2,
	.resource	= (struct resource []) {
		{
			.start = 0xfdb18000,
			.end   = 0xfdb180df,
			.flags = IORESOURCE_MEM
		}, {
			.start = ILC_IRQ(28),
			.end   = ILC_IRQ(28),
			.flags = IORESOURCE_IRQ
		},
	},
	.dev = {
		   .power.can_wakeup = 1,
		   .platform_data = &fli7510_lirc_private_info
	}
};

void __init fli7510_configure_lirc(lirc_scd_t *scd)
{
	fli7510_lirc_private_info.scd_info = scd;

	platform_device_register(&fli7510_lirc_device);
}

void __init fli7510_configure_nand(struct platform_device *pdev)
{
	/* EMI Bank base address */
	/*  - setup done in stm_nand_emi probe */

	/* NAND Controller base address */
	pdev->resource[0].start	= 0xFD101000;
	pdev->resource[0].end	= 0xFD101FFF;

	/* NAND Controller IRQ */
	pdev->resource[1].start = ILC_IRQ(35);
	pdev->resource[1].end	= ILC_IRQ(35);

	platform_device_register(pdev);
}


/* PCI Resources ---------------------------------------------------------- */

/* You may pass one of the PCI_PIN_* constants to use dedicated pin or
 * just pass interrupt number generated with gpio_to_irq() when PIO pads
 * are used as interrupts or IRLx_IRQ when using external interrupts inputs */
int fli7510_pcibios_map_platform_irq(struct pci_config_data *pci_config,
		u8 pin)
{
	int result;
	int type;

	if (pin < 1 || pin > 4)
		return -1;

	type = pci_config->pci_irq[pin - 1];

	switch (type) {
	case PCI_PIN_ALTERNATIVE:
		/* Actually there are no alternative pins... */
		BUG();
		result = -1;
		break;
	case PCI_PIN_DEFAULT:
		/* Only INTA/INTB are described as "dedicated" PCI
		 * interrupts and even if these two are described as that,
		 * they are actually just "normal" external interrupt
		 * inputs (INT2 & INT3)... Additionally, depending
		 * on the spec version, the number below may seem wrong,
		 * but believe me - they are correct :-) */
		switch (pin) {
		case 1 ... 2:
			result = ILC_IRQ(119 - (pin - 1));
			/* Configure the ILC input to be active low,
			 * which is the PCI way... */
			set_irq_type(result, IRQ_TYPE_LEVEL_LOW);
			break;
		default:
			/* Other should be passed just as interrupt number
			 * (eg. result of the ILC_IRQ() macro) */
			BUG();
			result = -1;
			break;
		}
		break;
	case PCI_PIN_UNUSED:
		result = -1; /* Not used */
		break;
	default:
		/* Take whatever interrupt you are told */
		result = type;
		break;
	}

	return result;
}

static struct platform_device fli7510_pci_device = {
	.name = "pci_stm",
	.id = -1,
	.num_resources = 7,
	.resource = (struct resource[]) {
		{
			.name = "Memory",
			.start = 0xc0000000,
			.end = 0xdfffffff, /* 512 MB */
			.flags = IORESOURCE_MEM,
		}, {
			.name = "IO",
			.start = 0x0400,
			.end = 0xffff,
			.flags = IORESOURCE_IO,
		}, {
			.name = "EMISS",
			.start = 0xfd200000,
			.end = 0xfd2017fc,
			.flags = IORESOURCE_MEM,
		}, {
			.name = "PCI-AHB",
			.start = 0xfd080000,
			.end = 0xfd0800ff,
			.flags = IORESOURCE_MEM,
		}, {
			.name = "DMA",
			.start = ILC_IRQ(47),
			.end = ILC_IRQ(47),
			.flags = IORESOURCE_IRQ,
		}, {
			.name = "Error",
			.start = ILC_IRQ(48),
			.end = ILC_IRQ(48),
			.flags = IORESOURCE_IRQ,
		}, { /* Keep this one last */
			.name = "SERR",
			/* .start & .end set in stx7105_configure_pci() */
			.flags = IORESOURCE_IRQ,
		}
	},
};

void __init fli7510_configure_pci(struct pci_config_data *pci_conf)
{
	struct sysconf_field *sc;
	int i;

	/* Fill in the default values */
	if (!pci_conf->ad_override_default) {
		pci_conf->ad_threshold = 5;
		pci_conf->ad_read_ahead = 1;
		pci_conf->ad_chunks_in_msg = 0;
		pci_conf->ad_pcks_in_chunk = 0;
		pci_conf->ad_trigger_mode = 1;
		pci_conf->ad_max_opcode = 5;
		pci_conf->ad_posted = 1;
	}

	/* The EMI_BUS_REQ[0] pin (also know just as EmiBusReq) is
	 * internally wired to the arbiter's PCI request 3 line.
	 * And the answer to the obvious question is: That's right,
	 * the EMI_BUSREQ[3] is not wired at all... */
	pci_conf->req0_to_req3 = 1;
	BUG_ON(pci_conf->req_gnt[3] != PCI_PIN_UNUSED);

	/* Copy over platform specific data to driver */
	fli7510_pci_device.dev.platform_data = pci_conf;

	/* REQ/GNT[0] are dedicated EMI pins */
	BUG_ON(pci_conf->req_gnt[0] != PCI_PIN_DEFAULT);

	/* REQ/GNT[1..2] PIOs setup */
	for (i = 1; i <= 2; i++) {
		switch (pci_conf->req_gnt[i]) {
		case PCI_PIN_DEFAULT:
			/* emiss_bus_req_enable */
			sc = sysconf_claim(CFG_COMMS_CONFIG_1,
					24 + (i - 1), 24 + (i - 1), "PCI");
			sysconf_write(sc, 1);
			stpio_request_pin(15, (i - 1) * 2, "PCI REQ",
					STPIO_IN);
			stpio_request_pin(15, ((i - 1) * 2) + 1, "PCI GNT",
					STPIO_ALT_OUT);
			break;
		case PCI_PIN_UNUSED:
			/* Unused is unused - nothing to do */
			break;
		default:
			/* No alternative here... */
			BUG();
			break;
		}
	}

	/* REG/GNT[3] are... unavailable... */
	BUG_ON(pci_conf->req_gnt[3] != PCI_PIN_UNUSED);

	/* Claim "dedicated" interrupt pins... */
	for (i = 0; i < 4; i++) {
		static const char *int_name[] = {
			"PCI INTA",
			"PCI INTB",
		};

		switch (pci_conf->pci_irq[i]) {
		case PCI_PIN_DEFAULT:
			if (i > 1) {
				BUG();
				break;
			}
			stpio_request_pin(25, i, int_name[i], STPIO_IN);
			break;
		case PCI_PIN_ALTERNATIVE:
			/* No alternative here... */
			BUG();
			break;
		default:
			/* Unused or interrupt number passed, nothing to do */
			break;
		}
	}

	/* Configure the SERR interrupt (if wired up) */
	switch (pci_conf->serr_irq) {
	case PCI_PIN_DEFAULT:
		if (stpio_request_pin(16, 6, "PCI SERR#", STPIO_IN)) {
			pci_conf->serr_irq = gpio_to_irq(stpio_to_gpio(16, 6));
			set_irq_type(pci_conf->serr_irq, IRQ_TYPE_LEVEL_LOW);
		} else {
			printk(KERN_WARNING "%s(): Failed to claim PCI SERR# "
					"PIO!\n", __func__);
			pci_conf->serr_irq = PCI_PIN_UNUSED;
		}
		break;
	case PCI_PIN_ALTERNATIVE:
		/* No alternative here */
		BUG();
		pci_conf->serr_irq = PCI_PIN_UNUSED;
		break;
	}
	if (pci_conf->serr_irq == PCI_PIN_UNUSED) {
		/* "Disable" the SERR IRQ resource (it's last on the list) */
		fli7510_pci_device.num_resources--;
	} else {
		/* The SERR IRQ resource is last */
		int res_num = fli7510_pci_device.num_resources - 1;
		struct resource *res = &fli7510_pci_device.resource[res_num];

		res->start = pci_conf->serr_irq;
		res->end = pci_conf->serr_irq;
	}


#if defined(CONFIG_PM)
#warning TODO: PCI Power Management
#endif
	/* pci_pwr_dwn_req */
	sc = sysconf_claim(CFG_PWR_DWN_CTL, 2, 2, "PCI");
	sysconf_write(sc, 0);

	/* status_pci_pwr_dwn_grant */
	sc = sysconf_claim(CFG_PCI_ROPC_STATUS, 18, 18, "PCI");
	while (sysconf_read(sc))
		cpu_relax();

	platform_device_register(&fli7510_pci_device);
}



/* Other resources (ILC, sysconf and PIO) --------------------------------- */

static struct platform_device fli7510_st40host_ilc3_device = {
	.name		= "ilc3",
	.id		= -1,
	.num_resources	= 1,
	.resource	= (struct resource[]) {
		{
			.start	= 0xfd218000,
			.end	= 0xfd218000 + 0x900,
			.flags	= IORESOURCE_MEM
		}
	},
};

static struct platform_device fli7510_st40rt_ilc3_device = {
	.name		= "ilc3",
	.id		= -1,
	.num_resources	= 1,
	.resource	= (struct resource[]) {
		{
			.start	= 0xfd210000,
			.end	= 0xfd210000 + 0x900,
			.flags	= IORESOURCE_MEM
		}
	},
};

static struct platform_device fli7510_sysconf_devices[] = {
	{
		.name		= "sysconf",
		.id		= 0,
		.num_resources	= 1,
		.resource	= (struct resource[]) {
			{
				.start	= 0xfd220000,
				.end	= 0xfd22001f,
				.flags	= IORESOURCE_MEM
			}
		},
		.dev.platform_data = &(struct plat_sysconf_data) {
			.groups_num = 1,
			.groups = (struct plat_sysconf_group []) {
				PLAT_SYSCONF_GROUP(PRB_PU_CFG_1, 0),
			},
		}
	}, {
		.name		= "sysconf",
		.id		= 1,
		.num_resources	= 1,
		.resource	= (struct resource[]) {
			{
				.start	= 0xfd228000,
				.end	= 0xfd22801f,
				.flags	= IORESOURCE_MEM
			}
		},
		.dev.platform_data = &(struct plat_sysconf_data) {
			.groups_num = 1,
			.groups = (struct plat_sysconf_group []) {
				PLAT_SYSCONF_GROUP(PRB_PU_CFG_2, 0),
			},
		}
	}, {
		.name		= "sysconf",
		.id		= 2,
		.num_resources	= 1,
		.resource	= (struct resource[]) {
			{
				.start	= 0xfd9ec000,
				.end	= 0xfd9ec01f,
				.flags	= IORESOURCE_MEM
			}
		},
		.dev.platform_data = &(struct plat_sysconf_data) {
			.groups_num = 1,
			.groups = (struct plat_sysconf_group []) {
				PLAT_SYSCONF_GROUP(TRS_PU_CFG_0, 0),
			},
		}
	}, {
		.name		= "sysconf",
		.id		= 3,
		.num_resources	= 1,
		.resource	= (struct resource[]) {
			{
				.start	= 0xfd9f4000,
				.end	= 0xfd9f401f,
				.flags	= IORESOURCE_MEM
			}
		},
		.dev.platform_data = &(struct plat_sysconf_data) {
			.groups_num = 1,
			.groups = (struct plat_sysconf_group []) {
				PLAT_SYSCONF_GROUP(TRS_PU_CFG_1, 0),
			},
		}
	}, {
		.name		= "sysconf",
		.id		= 4,
		.num_resources	= 1,
		.resource	= (struct resource[]) {
			{
				.start	= 0xfd7a0000,
				.end	= 0xfd7a001f,
				.flags	= IORESOURCE_MEM
			}
		},
		.dev.platform_data = &(struct plat_sysconf_data) {
			.groups_num = 1,
			.groups = (struct plat_sysconf_group []) {
				PLAT_SYSCONF_GROUP(VDEC_PU_CFG_0, 0),
			},
		}
	}, {
		.name		= "sysconf",
		.id		= 5,
		.num_resources	= 1,
		.resource	= (struct resource[]) {
			{
				.start	= 0xfd7c0000,
				.end	= 0xfd7c001f,
				.flags	= IORESOURCE_MEM
			}
		},
		.dev.platform_data = &(struct plat_sysconf_data) {
			.groups_num = 1,
			.groups = (struct plat_sysconf_group []) {
				PLAT_SYSCONF_GROUP(VDEC_PU_CFG_1, 0),
			},
		}
	}, {
		.name		= "sysconf",
		.id		= 6,
		.num_resources	= 1,
		.resource	= (struct resource[]) {
			{
				.start	= 0xfd5e8000,
				.end	= 0xfd5e801f,
				.flags	= IORESOURCE_MEM
			}
		},
		.dev.platform_data = &(struct plat_sysconf_data) {
			.groups_num = 1,
			.groups = (struct plat_sysconf_group []) {
				PLAT_SYSCONF_GROUP(VOUT_PU_CFG_1, 0),
			},
		}
	}
};

static struct platform_device fli7510_pio_devices[] = {
	STPIO_DEVICE(0, 0xfd5c0000, ILC_IRQ(75)),
	STPIO_DEVICE(1, 0xfd5c4000, ILC_IRQ(76)),
	STPIO_DEVICE(2, 0xfd5c8000, ILC_IRQ(77)),
	STPIO_DEVICE(3, 0xfd5cc000, ILC_IRQ(78)),
	STPIO_DEVICE(4, 0xfd5d0000, ILC_IRQ(79)),
	STPIO_DEVICE(5, 0xfd5d4000, ILC_IRQ(80)),
	STPIO_DEVICE(6, 0xfd5d8000, -1),
	STPIO_DEVICE(7, 0xfd5dc000, -1),
	STPIO_DEVICE(8, 0xfd5e0000, -1),
	STPIO_DEVICE(9, 0xfd5e4000, -1),
	STPIO_DEVICE(10, 0xfd984000, ILC_IRQ(125)),
	STPIO_DEVICE(11, 0xfd988000, -1),
	STPIO_DEVICE(12, 0xfd98c000, -1),
	STPIO_DEVICE(13, 0xfd990000, ILC_IRQ(2)),
	STPIO_DEVICE(14, 0xfd994000, ILC_IRQ(3)),
	STPIO_DEVICE(15, 0xfd998000, ILC_IRQ(81)),
	STPIO_DEVICE(16, 0xfd99c000, ILC_IRQ(82)),
	STPIO_DEVICE(17, 0xfd9a0000, -1),
	STPIO_DEVICE(18, 0xfd9a4000, -1),
	STPIO_DEVICE(19, 0xfd9a8000, -1),
	STPIO_DEVICE(20, 0xfd9ac000, -1),
	STPIO_DEVICE(21, 0xfd9b0000, -1),
	STPIO_DEVICE(22, 0xfd9b4000, ILC_IRQ(83)),
	STPIO_DEVICE(23, 0xfd9b8000, ILC_IRQ(84)),
	STPIO_DEVICE(24, 0xfd9bc000, ILC_IRQ(85)),
	STPIO_DEVICE(25, 0xfd9c0000, -1),
	STPIO_DEVICE(26, 0xfd9c4000, -1),
	STPIO_DEVICE(27, 0xfd9c8000, -1),
};



/* Early devices initialization ------------------------------------------- */

/* Initialise devices which are required early in the boot process
 * (called from the board setup file) */
void __init fli7510_early_device_init(void)
{
	struct sysconf_field *sc;
	unsigned long devid;
	unsigned long chip_revision;

	/* Initialise PIO and sysconf drivers */

	sysconf_early_init(fli7510_sysconf_devices,
			ARRAY_SIZE(fli7510_sysconf_devices));

	stpio_early_init(fli7510_pio_devices,
			ARRAY_SIZE(fli7510_pio_devices),
			ILC_FIRST_IRQ + ILC_NR_IRQS);

	sc = sysconf_claim(CFG_DEVICE_ID, 0, 31, "devid");
	devid = sysconf_read(sc);
	chip_revision = (devid >> 28) + 1;
	boot_cpu_data.cut_major = chip_revision;

	printk(KERN_INFO "Freeman 510 version %ld.x, ST40%s core\n",
			chip_revision, ST40HOST_CORE ? "HOST" : "RT");

	/* We haven't configured the LPC, so the sleep instruction may
	 * do bad things. Thus we disable it here. */
	disable_hlt();
}

/* Pre-arch initialisation ------------------------------------------------ */

/* EMI access is required very early */
static int __init fli7510_postcore_setup(void)
{
	int i;

	emi_init(0, 0xfd100000);

	for (i = 0; i < ARRAY_SIZE(fli7510_pio_devices); i++)
		platform_device_register(&fli7510_pio_devices[i]);

	return 0;
}
postcore_initcall(fli7510_postcore_setup);



/* Late resources --------------------------------------------------------- */

static struct platform_device *fli7510_late_devices[] __initdata = {
	&fli7510_fdma_devices[0],
	&fli7510_fdma_devices[1],
	&fli7510_fdma_xbar_device,
	&fli7510_sysconf_devices[0],
	&fli7510_sysconf_devices[1],
	&fli7510_sysconf_devices[2],
	&fli7510_sysconf_devices[3],
	&fli7510_sysconf_devices[4],
	&fli7510_sysconf_devices[5],
	&fli7510_sysconf_devices[6],
	&fli7510_devhwrandom_device,
	&fli7510_devrandom_device,
};

#include "./platform-pm-fli7510.c"

static int __init fli7510_devices_setup(void)
{
	int result;

	platform_add_pm_devices(fli7510_pm_devices,
			ARRAY_SIZE(fli7510_pm_devices));

	if (ST40HOST_CORE)
		result = platform_device_register(
				&fli7510_st40host_ilc3_device);
	else
		result = platform_device_register(&fli7510_st40rt_ilc3_device);
	if (result != 0)
		return result;

	return platform_add_devices(fli7510_late_devices,
				    ARRAY_SIZE(fli7510_late_devices));
}
device_initcall(fli7510_devices_setup);



/* Interrupt initialisation ----------------------------------------------- */

enum {
	UNUSED = 0,

	/* interrupt sources */
	IRL0, IRL1, IRL2, IRL3, /* only IRLM mode described here */
	TMU0, TMU1, TMU2, WDT, HUDI,
};

static struct intc_vect fli7510_intc_vectors[] = {
	INTC_VECT(TMU0, 0x400), INTC_VECT(TMU1, 0x420),
	INTC_VECT(TMU2, 0x440), INTC_VECT(TMU2, 0x460),
	INTC_VECT(WDT, 0x560),
	INTC_VECT(HUDI, 0x600),
};

static struct intc_prio_reg fli7510_intc_prio_registers[] = {
					   /*   15-12, 11-8,  7-4,   3-0 */
	{ 0xffd00004, 0, 16, 4, /* IPRA */     { TMU0, TMU1, TMU2,     0 } },
	{ 0xffd00008, 0, 16, 4, /* IPRB */     {  WDT,    0,    0,     0 } },
	{ 0xffd0000c, 0, 16, 4, /* IPRC */     {    0,    0,    0,  HUDI } },
	{ 0xffd00010, 0, 16, 4, /* IPRD */     { IRL0, IRL1,  IRL2, IRL3 } },
};

static DECLARE_INTC_DESC(fli7510_intc_desc, "fli7510",
		fli7510_intc_vectors, NULL,
		NULL, fli7510_intc_prio_registers, NULL);

void __init plat_irq_setup(void)
{
	int irq;

	register_intc_controller(&fli7510_intc_desc);

	/* The ILC outputs (16 lines) are connected to the IRL0..3
	 * INTC inputs via priority level encoder */
	for (irq = 0; irq < 16; irq++) {
		set_irq_chip(irq, &dummy_irq_chip);
		set_irq_chained_handler(irq, ilc_irq_demux);
	}

	if (ST40HOST_CORE)
		ilc_early_init(&fli7510_st40host_ilc3_device);
	else
		ilc_early_init(&fli7510_st40rt_ilc3_device);

	ilc_demux_init();
}
