/*
 * (C) Copyright 2009 WyPlay SAS.
 * Frederic Mazuel <fmazuel@wyplay.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/sysdev.h>
#include <linux/ctype.h>
#include <linux/leds.h>
#include "leds.h"

struct flash_ctx {
	unsigned long delay_on;		/* milliseconds on */
	unsigned long delay_off;	/* milliseconds off */
	unsigned char state[10];
	int (*set_flash)(struct led_classdev *ldev, char *state);
	int (*read_flash)(struct led_classdev *ldev, char *state);
};

static ssize_t led_delay_on_show(struct device *dev,
				 struct device_attribute *attr,
				 char *buf)
{
	struct led_classdev *led_cdev = dev_get_drvdata(dev);
	struct flash_ctx *flash_data = led_cdev->trigger_data;

	sprintf(buf, "%lu\n", flash_data->delay_on);

	return strlen(buf) + 1;
}

static ssize_t led_delay_on_store(struct device *dev,
				  struct device_attribute *attr,
				  const char *buf,
				  size_t size)
{
	struct led_classdev *led_cdev = dev_get_drvdata(dev);
	struct flash_ctx *flash_data = led_cdev->trigger_data;
	int ret = -EINVAL;
	char *after;
	unsigned long value = simple_strtoul(buf, &after, 10);
	size_t count = after - buf;

	if (*after && isspace(*after))
		count++;

	if (count == size) {
		flash_data->delay_on = value;
		ret = count;
	}

	return ret;
}

static ssize_t led_delay_off_show(struct device *dev,
				  struct device_attribute *attr,
				  char *buf)
{
	struct led_classdev *led_cdev = dev_get_drvdata(dev);
	struct flash_ctx *flash_data = led_cdev->trigger_data;

	sprintf(buf, "%lu\n", flash_data->delay_off);

	return strlen(buf) + 1;
}

static ssize_t led_delay_off_store(struct device *dev,
				   struct device_attribute *attr,
				   const char *buf,
				   size_t size)
{
	struct led_classdev *led_cdev = dev_get_drvdata(dev);
	struct flash_ctx *flash_data = led_cdev->trigger_data;
	int ret = -EINVAL;
	char *after;
	unsigned long value = simple_strtoul(buf, &after, 10);
	size_t count = after - buf;

	if (*after && isspace(*after))
		count++;

	if (count == size) {
		flash_data->delay_off = value;
		ret = count;
	}

	return ret;
}

static ssize_t flash_show(struct device *dev,
			  struct device_attribute *attr,
			  char *buf)
{
	struct led_classdev *led_cdev = dev_get_drvdata(dev);
	struct flash_ctx *flash_data = led_cdev->trigger_data;

	if (flash_data->read_flash)
		flash_data->read_flash(led_cdev, buf);

	return strlen(buf) + 1;
}

static ssize_t flash_store(struct device *dev,
			   struct device_attribute *attr,
			   const char *buf,
			   size_t size)
{
	struct led_classdev *led_cdev = dev_get_drvdata(dev);
	struct flash_ctx *flash_data = led_cdev->trigger_data;
	int ret = -EINVAL;

	if (strlen(buf) >= sizeof(flash_data->state)) {
		ret = -EINVAL;
		goto exit;
	}

	ret = sscanf(buf, "%s", flash_data->state);
	if (ret != 1) {
		ret = -EINVAL;
		goto exit;
	}

	if (flash_data->set_flash)
		flash_data->set_flash(led_cdev, flash_data->state);

	ret = strnlen(buf, size);
exit:
	return ret;
}

static DEVICE_ATTR(delay_on, 0644, led_delay_on_show, led_delay_on_store);
static DEVICE_ATTR(delay_off, 0644, led_delay_off_show, led_delay_off_store);
static DEVICE_ATTR(flash_mode, 0644, flash_show, flash_store);

static void ledserv_trig_activate(struct led_classdev *led_cdev)
{
	int rc;
	struct flash_ctx *flash_data = led_cdev->trigger_data;

	printk("%s\n",__FUNCTION__);

	if (flash_data == NULL)
		goto exit;

	rc = device_create_file(led_cdev->dev, &dev_attr_delay_on);
	if (rc)
		goto exit;
	rc = device_create_file(led_cdev->dev, &dev_attr_delay_off);
	if (rc)
		goto err_out_delayon;
	rc = device_create_file(led_cdev->dev, &dev_attr_flash_mode);
	if (rc)
		goto err_out_delayoff;

exit:
	return;

err_out_flasmode:
	device_remove_file(led_cdev->dev, &dev_attr_flash_mode);
err_out_delayoff:
	device_remove_file(led_cdev->dev, &dev_attr_delay_off);
err_out_delayon:
	device_remove_file(led_cdev->dev, &dev_attr_delay_on);
}

static void ledserv_trig_deactivate(struct led_classdev *led_cdev)
{
	struct flash_ctx *flash_data = led_cdev->trigger_data;

	if (flash_data) {
		device_remove_file(led_cdev->dev, &dev_attr_delay_on);
		device_remove_file(led_cdev->dev, &dev_attr_delay_off);
		device_remove_file(led_cdev->dev, &dev_attr_flash_mode);
	}
}

static struct led_trigger ledserv_led_trigger = {
	.name     = "ledservtrig",
	.activate = ledserv_trig_activate,
	.deactivate = ledserv_trig_deactivate,
};

static int __init ledserv_trig_init(void)
{
	printk("%s\n",__FUNCTION__);
	return led_trigger_register(&ledserv_led_trigger);
}

static void __exit ledserv_trig_exit(void)
{
	led_trigger_unregister(&ledserv_led_trigger);
}

module_init(ledserv_trig_init);
module_exit(ledserv_trig_exit);

MODULE_AUTHOR("Frederic Mazuel <fmazuel@wyplay.com>");
MODULE_DESCRIPTION("ledserv LED trigger");
MODULE_LICENSE("GPL");
