/*
 * (C) Copyright 2008 WyPlay SAS.
 * Jean-Christophe PLAGNIOL-VILLARD <jcplagniol@wyplay.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/autoconf.h>
#include <linux/platform_device.h>
#include <linux/wymcu/mcu.h>
#include <linux/i2c.h>

struct mcu_dev_bus {
	unsigned int bus_type;
	struct mcu_dev *dev;
};

static struct mcu_bus_ctx {
	unsigned int dev_nb;
	unsigned int demo_nb;
	unsigned int i2c_nb;
	unsigned int serial_nb;
	struct mcu_dev_bus devs[MAX_MCU_NB];
} bus_ctx;


static int mcu_i2c_drv_probe(struct i2c_client *);
static int mcu_i2c_drv_remove(struct i2c_client *);

#define BUS_NAME  "mcu_bus_i2c"

static struct i2c_driver mcu_i2c_driver =
{
	.driver = {
		.name	= BUS_NAME,
	},
	.probe	= mcu_i2c_drv_probe,
	.remove	= mcu_i2c_drv_remove,
};

#ifdef CONFIG_PM
static int mcu_suspend(struct platform_device *dev, pm_message_t state)
{
	return 0;
}

static int mcu_resume(struct platform_device *dev)
{
	return 0;
}
#endif

static void mcu_dev_release(struct device *_dev)
{
	struct mcu_dev *dev = to_mcu_dev(_dev);

	kfree(dev);
}

static int mcu_init_one_child(struct device *me,
			       struct mcu_dev *dev,
			       int index)
{
	int ret;
	char name[20];

	if (!dev) {
		ret = -ENOMEM;
		goto exit;
	}

	sprintf(name, "%s%d", dev->name, index);
	strncpy(dev->dev.bus_id,name,sizeof(dev->dev.bus_id));

	dev->dev.parent = me;
	dev->dev.bus = &mcu_bus_type;
	dev->dev.release = mcu_dev_release;

	printk(KERN_DEBUG "%s: register bus_id %s\n",__FUNCTION__,name);

	ret = device_register(&dev->dev);
exit:
	return ret;

}

static int mcu_remove_child(struct device *dev, void *data)
{
	device_unregister(dev);
	return 0;
}

static int mcu_i2c_drv_probe(struct i2c_client *client)
{
	int ret = 0;
	int i;
	struct mcu_dev_bus *mcu_i2c_dev = &bus_ctx.devs[0];
	unsigned int mcu_i2c_addr;

	printk(KERN_DEBUG "%s: client %s used driver %s at addr 0x%x\n",
	       __FUNCTION__, client->name, client->driver_name, client->addr);

	if (i2c_smbus_xfer(client->adapter, client->addr, 0, 0, 0,
			   I2C_SMBUS_QUICK, NULL) < 0) {
		ret = -1;
		goto exit;
	}

	for (i = 0; i < bus_ctx.dev_nb; i++, mcu_i2c_dev++) {
		printk(KERN_DEBUG "%s: device %d ",__FUNCTION__, i);
		mcu_i2c_addr = (unsigned int)(mcu_i2c_dev->dev->bus_data);
		if ((mcu_i2c_dev->bus_type == I2C) &&
		    (mcu_i2c_addr == client->addr)) {
			printk("match i2c device \n");
			mcu_i2c_dev->dev->bus_client = (void *)client;
		}
		printk("\n");
	}
exit:
	return ret;
}

static int mcu_i2c_drv_remove(struct i2c_client *client)
{
	int ret = 0;

	printk(KERN_DEBUG "%s\n",__FUNCTION__);

	return ret;
}

struct i2c_board_info info;
static int mcu_probe(struct platform_device *dev)
{
	struct device *me = &dev->dev;
	struct mcu_devs *mcu_devs = (struct mcu_devs *)(me->platform_data);
	int i;
	int ret = 0;
	struct i2c_adapter *adapter;

	if (bus_ctx.dev_nb) {
		printk(KERN_WARNING "mcu bus already used\n");
		ret = -1;
		goto exit;
	}

	if(mcu_devs->nb > MAX_MCU_NB) {
		printk(KERN_WARNING "Too many mcu devices ressources %d > %d\n",
		       mcu_devs->nb, MAX_MCU_NB);
		ret = -1;
		goto exit;
	}

	printk(KERN_DEBUG "%d mcu device have been detected\n",mcu_devs->nb);
	for (i = 0; i < mcu_devs->nb; i++) {
		switch (mcu_devs->devs[i].bus_type) {
		case I2C:
			bus_ctx.devs[i].bus_type = I2C;
			bus_ctx.i2c_nb++;

			strcpy(info.driver_name, BUS_NAME);
			strcpy(info.type, "mcu");
			info.platform_data = NULL;
			info.flags	= 0;
			info.irq	= 0;
			info.addr	= (unsigned short)(unsigned long)
					  (mcu_devs->devs[i].bus_data);

			adapter = i2c_get_adapter(mcu_devs->devs[i].bus_id);
			i2c_new_device(adapter, &info);
			break;
		case SERIAL:
			bus_ctx.devs[i].bus_type = SERIAL;
			bus_ctx.serial_nb++;
			/* Maybe we should add something about uart driver here */
			break;
		case DEMO:
			bus_ctx.devs[i].bus_type = DEMO;
			bus_ctx.demo_nb++;
			break;
		default:
			bus_ctx.devs[i].bus_type = UNKNOWN;
			printk("Unsupported mcu device %d type: %d",
			       i, mcu_devs->devs[i].bus_type);
			continue;
		}
		bus_ctx.devs[i].dev = &(mcu_devs->devs[i]);
		bus_ctx.dev_nb++;
	}

	if(bus_ctx.i2c_nb == 1) {
		if (i2c_register_driver(THIS_MODULE,
					&mcu_i2c_driver)) {
			printk("Registering I2C drv failed");
		}
	}
	/* serial support */
	if(bus_ctx.serial_nb == 1) {
		/* Maybe we should add something about uart driver here */


	}

	for (i = 0; i < bus_ctx.dev_nb; i++) {
		mcu_init_one_child(me, bus_ctx.devs[i].dev, i);
	}

exit:
	return ret;
}

static int mcu_remove(struct platform_device *dev)
{
	struct device *me = &dev->dev;

	if (me)
		device_for_each_child(me, NULL, mcu_remove_child);

	memset(&bus_ctx, 0, sizeof(mcu_bus_type));

	return 0;
}

/*
 *	Not sure if this should be on the system bus or not yet.
 *	We really want some way to register a system device at
 *	the per-machine level, and then have this driver pick
 *	up the registered devices.
 */
static struct platform_driver mcu_device_driver = {
	.driver		= {
		.owner	= THIS_MODULE,
		.name	= "mcu",
	},
	.probe		= mcu_probe,
	.remove		= mcu_remove,
#ifdef CONFIG_PM
	.suspend	= mcu_suspend,
	.resume		= mcu_resume,
#endif
};


/*
 *	MCU "Register Access Bus."
 *
 *	We model this as a regular bus type, and hang devices directly
 *	off this.
 */
static int mcu_bus_match(struct device *_dev, struct device_driver *_drv)
{
	struct mcu_dev *dev = to_mcu_dev(_dev);
	struct mcu_driver *drv = to_mcu_drv(_drv);

	if(dev->bus_type != drv->bus_type)
		return 0;

	switch(dev->bus_type) {
	case I2C:
	case SERIAL:
		return dev->bus_data == drv->bus_data;
	case DEMO:
		return dev->bus_data == drv->bus_data;
	default:
		return 0;
	}
}

static int mcu_bus_suspend(struct device *dev, pm_message_t state)
{
	struct mcu_dev *ldev = to_mcu_dev(dev);
	struct mcu_driver *drv = to_mcu_drv(dev->driver);
	int ret = 0;

	if (drv && drv->suspend)
		ret = drv->suspend(ldev, state);
	return ret;
}

static int mcu_bus_resume(struct device *dev)
{
	struct mcu_dev *ldev = to_mcu_dev(dev);
	struct mcu_driver *drv = to_mcu_drv(dev->driver);
	int ret = 0;

	if (drv && drv->resume)
		ret = drv->resume(ldev);
	return ret;
}

static int mcu_bus_probe(struct device *dev)
{
	struct mcu_dev *ldev = to_mcu_dev(dev);
	struct mcu_driver *drv = to_mcu_drv(dev->driver);
	int ret = -ENODEV;

	if (drv && drv->probe)
		ret = drv->probe(ldev);
	return ret;
}

static int mcu_bus_remove(struct device *dev)
{
	struct mcu_dev *ldev = to_mcu_dev(dev);
	struct mcu_driver *drv = to_mcu_drv(dev->driver);
	int ret = 0;

	if (drv && drv->remove)
		ret = drv->remove(ldev);
	return ret;
}

struct bus_type mcu_bus_type = {
	.name		= "mcu-bus",
	.match		= mcu_bus_match,
	.probe		= mcu_bus_probe,
	.remove		= mcu_bus_remove,
	.suspend	= mcu_bus_suspend,
	.resume		= mcu_bus_resume,
};

int mcu_driver_register(struct mcu_driver *driver)
{
	driver->driver.bus = &mcu_bus_type;
	return driver_register(&driver->driver);
}

void mcu_driver_unregister(struct mcu_driver *driver)
{
	driver_unregister(&driver->driver);
}

static int __init mcu_init(void)
{
	int ret = 0;


	printk(KERN_DEBUG "MCU bus init\n");

	memset(&bus_ctx, 0, sizeof(bus_ctx));

	ret = bus_register(&mcu_bus_type);

	if (ret == 0)
		platform_driver_register(&mcu_device_driver);

	return ret;
}

static void __exit mcu_exit(void)
{
	printk(KERN_DEBUG "Mcu bus finishing\n");

	platform_driver_unregister(&mcu_device_driver);
	bus_unregister(&mcu_bus_type);
}

module_init(mcu_init);
module_exit(mcu_exit);

MODULE_DESCRIPTION("MCU bus");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jean-Christophe PLAGNIOL-VILLARD <jcplagniol@wyplay.com>");

EXPORT_SYMBOL(mcu_driver_register);
EXPORT_SYMBOL(mcu_driver_unregister);
EXPORT_SYMBOL(mcu_bus_type);
