/*
 * (C) Copyright 2009 WyPlay SAS.
 * Jean-Christophe PLAGNIOL-VILLARD <jcplagniol@wyplay.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/autoconf.h>
#include <linux/device.h>
#include <linux/wymcu/mcu.h>
#include <linux/wymcu/mcu_service.h>

static int index = 0;

static void mcu_service_dev_release(struct device *_dev)
{
	struct mcu_service_dev *dev = to_mcu_service_dev(_dev);

	kfree(dev);
}

int mcu_service_device_register(struct mcu_service_dev* dev)
{
	int ret;
	char name[20];

	if (!dev || !dev->mcu_dev || !dev->service) {
		ret = -ENOMEM;
		goto out;
	}

	sprintf(name, "%s%d", dev->name, index);
	index++;
	strncpy(dev->dev.bus_id,name,sizeof(dev->dev.bus_id));

	dev->dev.parent = &(dev->mcu_dev->dev);
	dev->dev.bus = &mcu_service_bus_type;
	dev->dev.release = mcu_service_dev_release;

	ret = device_register(&dev->dev);
out:
	return ret;

}

int mcu_service_device_unregister(struct mcu_service_dev* dev)
{
	device_unregister(&(dev->dev));
	return 0;
}

/*
 *	VMII "Register Access Bus."
 *
 *	We model this as a regular bus type, and hang devices directly
 *	off this.
 */
static int mcu_service_match(struct device *_dev, struct device_driver *_drv)
{
	struct mcu_service_dev *dev = to_mcu_service_dev(_dev);
	struct mcu_service_driver *drv = to_mcu_service_drv(_drv);

	return dev->service_type == drv->service_type;
}

static int mcu_service_bus_suspend(struct device *dev, pm_message_t state)
{
	struct mcu_service_dev *ldev = to_mcu_service_dev(dev);
	struct mcu_service_driver *drv = to_mcu_service_drv(dev->driver);
	int ret = 0;

	if (drv && drv->suspend)
		ret = drv->suspend(ldev, state);
	return ret;
}

static int mcu_service_bus_resume(struct device *dev)
{
	struct mcu_service_dev *ldev = to_mcu_service_dev(dev);
	struct mcu_service_driver *drv = to_mcu_service_drv(dev->driver);
	int ret = 0;

	if (drv && drv->resume)
		ret = drv->resume(ldev);
	return ret;
}

static int mcu_service_bus_probe(struct device *dev)
{
	struct mcu_service_dev *ldev = to_mcu_service_dev(dev);
	struct mcu_service_driver *drv = to_mcu_service_drv(dev->driver);
	int ret = -ENODEV;

	if (drv->probe)
		ret = drv->probe(ldev);
	return ret;
}

static int mcu_service_bus_remove(struct device *dev)
{
	struct mcu_service_dev *ldev = to_mcu_service_dev(dev);
	struct mcu_service_driver *drv = to_mcu_service_drv(dev->driver);
	int ret = 0;

	if (drv->remove)
		ret = drv->remove(ldev);
	return ret;
}

struct bus_type mcu_service_bus_type = {
	.name		= "mcu-service-bus",
	.match		= mcu_service_match,
	.probe		= mcu_service_bus_probe,
	.remove		= mcu_service_bus_remove,
	.suspend	= mcu_service_bus_suspend,
	.resume		= mcu_service_bus_resume,
};

int mcu_service_driver_register(struct mcu_service_driver *driver)
{
	driver->driver.bus = &mcu_service_bus_type;
	return driver_register(&driver->driver);
}

void mcu_service_driver_unregister(struct mcu_service_driver *driver)
{
	driver_unregister(&driver->driver);
}

static int __init mcu_service_init(void)
{
	return bus_register(&mcu_service_bus_type);
}

static void __exit mcu_service_exit(void)
{
	bus_unregister(&mcu_service_bus_type);
}

module_init(mcu_service_init);
module_exit(mcu_service_exit);

MODULE_DESCRIPTION("MCU Service bus");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jean-Christophe PLAGNIOL-VILLARD <jcplagniol@wyplay.com>");

EXPORT_SYMBOL(mcu_service_driver_register);
EXPORT_SYMBOL(mcu_service_driver_unregister);
EXPORT_SYMBOL(mcu_service_device_register);
EXPORT_SYMBOL(mcu_service_device_unregister);
EXPORT_SYMBOL(mcu_service_bus_type);
