/*
 * (C) Copyright 2009 WyPlay SAS.
 * Frederic Mazuel <fmazuel@wyplay.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/errno.h>
#include <linux/device.h>
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/stm/pio.h>
#include <linux/wymcu/mcu.h>
#include <linux/wymcu/mcu_service.h>
#include "mcu_1.h"

#if 0
#define MCU_DEBUG_LED
#endif

struct stpio_pin *pio_led = NULL;

static inline void wait_for_mcu(void) {
	int i;
	/* It should have at least 60ms between each i2c messages send 
	   to the MCU. We wait 100ms */
	for (i = 0; i < 100; i++)
		udelay(1000);
}

static int get_cpu_state(void);
static int set_cpu_state(int);
static int get_mcu_state(void);
static int get_snd_mode(struct mcu_service_dev *);
static int set_snd_mode(struct mcu_service_dev *, int);
static int get_rtc_status(void);
#ifdef CONFIG_MCU_LED
static int set_led_flash(struct mcu_service_dev *, int);
static int get_led_flash(struct mcu_service_dev *);
#endif
static int get_mcu_version(void);
static int get_firm_version(void);

static struct i2c_client *mcu_i2c_client;

#ifdef CONFIG_MCU_LED
static struct mcu_service_dev led_dev;
#endif
static struct mcu_service_dev snd_dev;
static struct mcu_service_dev mcu_dev;
static struct mcu_service_dev tmr_dev;

#ifdef CONFIG_MCU_LED
static struct led_service led_dev_callback = {
	.set_brightness		= NULL,
	.set_flash		= set_led_flash,
	.get_flash		= get_led_flash,
	.set_frequency		= NULL,
	.get_frequency		= NULL,
	.set_duty_cycle		= NULL,
	.get_duty_cycle		= NULL,
};
#endif

static struct sound_service sound_dev_callback = {
	.get_mode = get_snd_mode,
	.set_mode = set_snd_mode,
};

static struct timer_service timer_dev_callback = {
	.get_rtc_state	= get_rtc_status,
};

static struct mcu_service mcu_dev_callback = {
	.get_cpu_state	= get_cpu_state,
	.set_cpu_state	= set_cpu_state,
	.get_mcu_state	= get_mcu_state,
	.get_mcu_vers	= get_mcu_version,
	.get_firm_vers	= get_firm_version,
};

#ifdef CONFIG_MCU_LED
static int led_dev_registration(struct mcu_dev *dev){
	led_dev.name = NULL;
#if defined(CONFIG_MCU_P2)
	led_dev.name = "led:white";
#else
	led_dev.name = "wyled:blue";
#endif
	led_dev.mcu_dev = dev;
	led_dev.service_type = SERVICE_LED;
	led_dev.service = &led_dev_callback;
	led_dev.priv = NULL;
	return mcu_service_device_register(&led_dev);
}
#endif

static int sound_dev_registration(struct mcu_dev *dev){
	snd_dev.name = NULL;
	snd_dev.name = "service_snd";
	snd_dev.mcu_dev = dev;
	snd_dev.service_type = SERVICE_SND;
	snd_dev.service = &sound_dev_callback;
	snd_dev.priv = NULL;
	return mcu_service_device_register(&snd_dev);
}

static int timer_dev_registration(struct mcu_dev *dev){
	tmr_dev.name = NULL;
	tmr_dev.name = "service_tmr";
	tmr_dev.mcu_dev = dev;
	tmr_dev.service_type = SERVICE_TIMER;
	tmr_dev.service = &timer_dev_callback;
	tmr_dev.priv = NULL;
	return mcu_service_device_register(&tmr_dev);
}

static int mcu_dev_registration(struct mcu_dev *dev){
	mcu_dev.name = NULL;
	mcu_dev.name = "mcu";
	mcu_dev.mcu_dev = dev;
	mcu_dev.service_type = SERVICE_MCU;
	mcu_dev.service = &mcu_dev_callback;
	mcu_dev.priv = NULL;
	return mcu_service_device_register(&mcu_dev);
}

static int get_cpu_state(void)
{
	int ret;

	wait_for_mcu();
	ret = i2c_smbus_read_byte_data(mcu_i2c_client, I2C_MCU_HOST_STATE);
	return ret;
}

static int set_cpu_state(int new_val)
{
	int ret = 0;

	switch (new_val) {
	case C_PLAYER:
	case C_POWER_DOWN:
	case C_IDLE:
	case C_BOOT_UP:
	case C_SHUT_DOWN_READY:
	case C_RECORDING:
		wait_for_mcu();
		i2c_smbus_write_byte_data(mcu_i2c_client, I2C_MCU_HOST_STATE, new_val);
		break;
	default:
		printk(KERN_WARNING "%s UNKNOWN CPU STATE: %d\n",
		       __FUNCTION__, new_val);
		ret = -EINVAL;
		break;
	}

	return ret;
}

static int get_mcu_state(void)
{
	wait_for_mcu();
	return i2c_smbus_read_byte_data(mcu_i2c_client, I2C_MCU_STATE);
}

static int get_snd_mode(struct mcu_service_dev *dev)
{
	int ret = 0;
	int val;

	wait_for_mcu();
	val = i2c_smbus_read_byte_data(mcu_i2c_client, I2C_MCU_CNTL0);
	val &= MUTE;
	val = val >> MUTE_POS;

	switch (val) {
	case 0:
		ret = MUTE_OFF;
		break;
	case 1:
		ret = MUTE_ON;
		break;
	default:
		ret = -EINVAL;
		break;
	}

	return ret;
}

static int set_snd_mode(struct mcu_service_dev *dev, int new_val)
{
	int ret = 0;
	int val;

	switch (new_val) {
	case MUTE_OFF:
		new_val = 0;
		break;
	case MUTE_ON:
		new_val = 1 << MUTE_POS;
		break;
	default:
		printk(KERN_WARNING "%s COMMAND NOT SUPPORTED: %d\n",
		       __FUNCTION__, new_val);
		ret = -EINVAL;
		goto exit;
	}

	wait_for_mcu();
	val = i2c_smbus_read_byte_data(mcu_i2c_client, I2C_MCU_CNTL0);
	val &= ~MUTE;
	val |= new_val;
	wait_for_mcu();
	i2c_smbus_write_byte_data(mcu_i2c_client, I2C_MCU_CNTL0, val);

exit:
	return ret;
}

static int get_rtc_status(void)
{
	int val;

	wait_for_mcu();
	val = i2c_smbus_read_byte_data(mcu_i2c_client, I2C_MCU_CNTL1);
	val &= CNTL1_MSK_RTC_INT;

	return val;
}

#ifdef CONFIG_MCU_LED
static int get_led_flash(struct mcu_service_dev *dev)
{
	int val;
	int ret;

	wait_for_mcu();
	val = i2c_smbus_read_byte_data(mcu_i2c_client, I2C_MCU_CNTL0);
	val &= LED;
	val = val >> LED_POS;

	if (val)
		ret = LED_FLASH_ON;
	else
		ret = LED_FLASH_OFF;

	return ret;
}

static int set_led_flash(struct mcu_service_dev *dev, int new_val)
{
	int val;
	int ret = 0;

	switch (new_val) {
	case LED_FLASH_OFF:
		#ifdef MCU_DEBUG_LED
		printk(KERN_DEBUG "mcu drv: DISABLE LED FLASH\n");
		#endif
		if (pio_led)
			stpio_set_pin(pio_led, 0);
		new_val = 0;
		break;

	case LED_FLASH_RCU:
		#ifdef MCU_DEBUG_LED
		printk(KERN_DEBUG "mcu drv: ENABLE LED FLASH\n");
		#endif
		if (pio_led)
			stpio_set_pin(pio_led, 0);
		new_val = 1 << LED_POS;
		break;

	case LED_FLASH_ON:
		#ifdef MCU_DEBUG_LED
		printk(KERN_DEBUG "mcu drv: ENABLE LED ON\n");
		#endif
		if (pio_led)
			stpio_set_pin(pio_led, 1);
		new_val = 0;
		break;

	default:
		printk(KERN_WARNING "mcu drv: COMMAND NOT SUPPORTED: %d\n",
		       new_val);
		ret = -EINVAL;
		goto exit;
	}

	wait_for_mcu();
	val = i2c_smbus_read_byte_data(mcu_i2c_client, I2C_MCU_CNTL0);
	val &= ~LED;
	val |= new_val;
	wait_for_mcu();
	i2c_smbus_write_byte_data(mcu_i2c_client, I2C_MCU_CNTL0, val);

exit:
	return ret;
}
#endif /* CONFIG_MCU_LED */

static int get_mcu_version(void)
{
	wait_for_mcu();
	return i2c_smbus_read_byte_data(mcu_i2c_client, I2C_MCU_VERSION);
}

static int get_firm_version(void)
{
	wait_for_mcu();
	return i2c_smbus_read_byte_data(mcu_i2c_client, I2C_MCU_APP_CODE);
}

static int mcu_1_probe(struct mcu_dev *dev)
{
	int ret = 0;

	printk(KERN_DEBUG "driver MCU_1 instance is starting\n");
#if defined(CONFIG_MCU_LED) && defined(CONFIG_MCU_P2)
	pio_led = stpio_request_pin(7, 0, "LED_IR_ACT", STPIO_OUT);
#endif

	mcu_i2c_client = dev->bus_client;
	if (mcu_i2c_client == NULL) {
		printk(KERN_WARNING "%s: No i2c client\n",__FUNCTION__);
		ret = 0;
		goto exit;
	}

	/* Device registrations */
#ifdef CONFIG_MCU_LED
	led_dev_registration(dev);
#endif
	sound_dev_registration(dev);
	timer_dev_registration(dev);
	mcu_dev_registration(dev);

exit:
	return ret;
}

static int mcu_1_remove(struct mcu_dev *dev)
{
	printk(KERN_DEBUG "driver MCU_1 instance is stopping\n");

#ifdef CONFIG_MCU_LED
	mcu_service_device_unregister(&led_dev);
#endif
	mcu_service_device_unregister(&snd_dev);
	mcu_service_device_unregister(&tmr_dev);
	mcu_service_device_unregister(&mcu_dev);
	if (pio_led) {
		stpio_free_pin(pio_led);
		pio_led = NULL;
	}
	return 0;
}

static struct mcu_driver mcu_1_drv = {
	.driver = {
		.owner	= THIS_MODULE,
		.name = "mcu_1",
	},
	.bus_type	= I2C,
	.bus_data	= (void *)I2C_MCU_ADDR,
	.probe		= mcu_1_probe,
	.remove		= mcu_1_remove,
};

static int __init mcu_1_init(void)
{
	printk(KERN_DEBUG "Driver MCU_1 is registring\n");
	return mcu_driver_register(&mcu_1_drv);
}

static void __exit mcu_1_exit(void)
{
	printk(KERN_DEBUG "Driver MCU_1 is unregistring\n");
	mcu_driver_unregister(&mcu_1_drv);
}

/*
 * This function aims to shutdown the SoC through the MCU.
 */
void mcu_soc_power_off(void)
{
	set_cpu_state(C_SHUT_DOWN_READY);
}
EXPORT_SYMBOL(mcu_soc_power_off);

module_init(mcu_1_init);
module_exit(mcu_1_exit);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Frederic MAZUEL <fmazuel@wyplay.com>");
MODULE_DESCRIPTION("MCU argosy X1-T2 driver");
