/*
 * (C) Copyright 2009 WyPlay SAS.
 * Frederic Mazuel <fmazuel@wyplay.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/sysfs.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/wymcu/mcu_service.h>
#include "mcu_1.h"


static struct mcu_service *services;
static dev_t mcuserv_dev_number;
static struct class *mcuserv_class;
static struct class_device *mcuserv_class_dev;

static struct state cstate[] = {
	{"idle",C_IDLE},
	{"boot_up",C_BOOT_UP},
	{"player",C_PLAYER},
	{"shut_down_ready", C_SHUT_DOWN_READY},
	{"power_down", C_POWER_DOWN},
	{"recording_only", C_RECORDING},
	{NULL, 0}
};

static struct state mcu_state[] = {
	{"idle",M_IDLE},
	{"self_test",M_SELF_TEST},
	{"system_standby",M_SYS_STANDBY},
	{"shut_down", M_SHUT_DOWN},
	{"USB_standby",M_USB_STANDBY},
	{"player_standby",M_PLAYER_STANDBY},
	{"player",M_PLAYER},
	{"USB",M_USB},
#if defined(CONFIG_MCU_P2)
	{"Auto_Play",M_AUTO_PLAY},
#endif
	{NULL, 0}
};

static struct state firm_vers[] = {
	{"MCU_ESS6461",MCU_ESS6461},
	{"MCU_RT107x", MCU_RT107x},
	{"MCU_ST7109_X1", MCU_ST7109_X1},
	{"MCU_ST7109_T2", MCU_ST7109_T2},
#if defined(CONFIG_MCU_P2)
	{"MCU_ST7105_P2", MCU_ST7105_P2},
#endif
	{NULL, 0}
};

static int get_cpu_state_val(char *state)
{
	struct state *pt = cstate;
	int val = -1;

	while (pt->name != NULL) {
		if(!strcmp(state, pt->name)) {
			val = pt->val;
			break;
		}
		pt++;
	}

	return val;
}

static char *get_cpu_state_name(int val)
{
	if (val < sizeof(cstate) /sizeof(struct state))
		return cstate[val].name;
	else
		return NULL;
}

static char *get_firm_vers_name(int val)
{
	struct state *pt = firm_vers;

	while (pt->name != NULL) {
		if(pt->val == val) {
			break;
		}
		pt++;
	}

	return pt->name;
}

static char *get_mcu_state_name(int val)
{
	if (val < sizeof(mcu_state) /sizeof(struct state))
		return mcu_state[val].name;
	else
		return NULL;
}

static ssize_t show_mcu_state(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int state;
	char *state_name;

	state = services->get_mcu_state();
	state_name = get_mcu_state_name(state);
	if (state_name != NULL) {
		pos += sprintf(pos, "%s\n",state_name);
	}

	return pos - buf;
}

static CLASS_DEVICE_ATTR(mcu_state, S_IRUGO,
			 show_mcu_state, NULL);

static ssize_t store_cpu_state(struct class_device *cdev, const char *buf, size_t count)
{
	char state_name[20];
	int res = -EINVAL;
	int state;

	res = sscanf(buf, "%s", state_name);
	if (res != 1) {
		res = -EINVAL;
		goto exit;
	}

	state = get_cpu_state_val(state_name);
	if (state < 0) {
		printk(KERN_WARNING "%s, unknown CPU state %s\n",
		       __FUNCTION__, state_name);
		res = -EINVAL;
		goto exit;
	}
	res = services->set_cpu_state(state);
	if (res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
 exit:
	return res;
}

static ssize_t show_cpu_state(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int state;
	char *state_name = NULL;

	state = services->get_cpu_state();
	state_name = get_cpu_state_name(state);
	if (state_name != NULL) {
		printk(KERN_DEBUG "%s: cpu state : %s\n",__FUNCTION__,state_name);
		pos += sprintf(pos, "%s\n",state_name);
	}

	return pos - buf;
}

static ssize_t show_version(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int version;

	version = services->get_mcu_vers();
	pos += sprintf(pos, "0x%x\n",version);

	return pos - buf;
}

static ssize_t show_firmware_version(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int version;
	char *version_name;

	version = services->get_firm_vers();
	if ((version_name = get_firm_vers_name(version)) != NULL) {
		pos += sprintf(pos, "%s\n",version_name);
	}

	return pos - buf;
}

static CLASS_DEVICE_ATTR(cpu_state, S_IRUGO | S_IWUSR,
			 show_cpu_state, store_cpu_state);

static CLASS_DEVICE_ATTR(version, S_IRUGO,
			 show_version, NULL);

static CLASS_DEVICE_ATTR(firmware_version, S_IRUGO,
			 show_firmware_version, NULL);


static int mcu_service_mcu_probe(struct mcu_service_dev *dev)
{
	int ret = 0;

	printk(KERN_DEBUG "MCU_1 service drv instance is starting\n");

	services = (struct mcu_service *)(dev->service);

	ret = alloc_chrdev_region (&mcuserv_dev_number, 0, 1, "mcuserv");
	if (ret < 0) {
		printk (KERN_ERR
			"%s: MCU service can't register device err: 0x%x\n",
			__FUNCTION__, ret);
		goto exit;
	}

	mcuserv_class = class_create (THIS_MODULE, "mcuserv");
	mcuserv_class_dev = class_device_create (mcuserv_class, NULL,
						 mcuserv_dev_number,
						 NULL, "mcuserv1");
	ret = class_device_create_file(mcuserv_class_dev,
				       &class_device_attr_cpu_state);
	if (ret) {
		printk( KERN_ERR
			"%s: ERROR: Creating CPU service state file: 0x%x)\n",
			__FUNCTION__, ret);
		goto exit;
	}

	ret = class_device_create_file(mcuserv_class_dev,
				       &class_device_attr_mcu_state);
	if (ret) {
		printk( KERN_ERR
			"%s: ERROR: Creating MCU service state file: 0x%x)\n",
			__FUNCTION__, ret);
		goto exit;
	}

	ret = class_device_create_file(mcuserv_class_dev,
				       &class_device_attr_version);
	if (ret) {
		printk( KERN_ERR
			"%s: ERROR: Creating MCU version file: 0x%x)\n",
			__FUNCTION__, ret);
		goto exit;
	}

	ret = class_device_create_file(mcuserv_class_dev,
				       &class_device_attr_firmware_version);
	if (ret) {
		printk( KERN_ERR
			"%s: ERROR: Creating MCU firmware version file: 0x%x)\n",
			__FUNCTION__, ret);
		goto exit;
	}

 exit:
	return ret;
}

static int mcu_service_mcu_remove(struct mcu_service_dev *dev)
{
	printk(KERN_DEBUG "MCU_1 service drv instance is stopping\n");
	return 0;
}

struct mcu_service_driver mcu_service_mcu_driver = {
	.driver = {
		.owner	= THIS_MODULE,
		.name	= "service_mcu",
	},
	.service_type = SERVICE_MCU,
	.probe = mcu_service_mcu_probe,
	.remove = mcu_service_mcu_remove,
};

static int __init mcu_service_init(void)
{
	printk(KERN_DEBUG "MCU_1 service drv is registering\n");
	return mcu_service_driver_register(&mcu_service_mcu_driver);
}

static void __exit mcu_service_exit(void)
{
	printk(KERN_DEBUG "MCU_1 service drv is unregistering\n");
	mcu_service_driver_unregister(&mcu_service_mcu_driver);
}

module_init(mcu_service_init);
module_exit(mcu_service_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Mcu service device support");
MODULE_AUTHOR("Frederic Mazuel <fmazuel@wyplay.com>");
