/*
 * (C) Copyright 2009 WyPlay SAS.
 * Frederic Mazuel <fmazuel@wyplay.com>
 * Jerry Zhang <zhangb@wyplayasia.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/errno.h>
#include <linux/device.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/tty_ldisc.h>
#include <linux/spinlock.h>
#include <linux/wait.h>
#include <linux/wymcu/mcu.h>
#include <linux/wymcu/mcu_service.h>
#include <linux/wymcu/mcu_ldisc.h>

/*
 * A new line discipline N_MCU.
 * N_MCU is defined in <linux/tty.h>
 */

/* some local function */
static void smcu_write_wakeup(struct tty_struct *tty);
static int mcu_write(struct smcu *smcu_info, pkg_general_t *pkg);
static int mcu_flush(struct smcu * smcu_info);
static int mcu_rcu_handler_register(struct smcu * smcu_info,
				    int (*handler_register)(struct smcu *));

/* pointer to the smcu main block */
struct smcu *smcu_local_ptr = NULL;

struct smcu** smcu_attach(void)
{
	return &smcu_local_ptr;
}

/*
 * @smcu_info: smcu main block
 * Return Value: for error check
 * We call mcu_write() one time only for one packet.
 * 1. For pkg of tirial command or activity command, it will return
 *    after the last byte sent out.
 * 2. For pkg of query-type command, it will return after the reiceiving
 *    of expected data.
 */
static int mcu_write(struct smcu * smcu_info, pkg_general_t *pkg)
{
	/* Note: Should add init.
	 *
	 */
	struct tty_struct *tty = smcu_info->tty;
	int ct, to_be_write;
	unsigned char *tx_ptr;
	int ret = 0;

	if(smcu_info->pkg_general)
		printk(KERN_ERR "%s -- Another package is still on the smcu\n",
			__FUNCTION__);
	/* attach the pkg to the smcu */
	smcu_info->pkg_general = pkg;
	tx_ptr = smcu_info->pkg_general->tx_buf;
	to_be_write = smcu_info->pkg_general->tx_len;

	if(smcu_info->mode == MODE_NORMAL) {
		if(smcu_info->pkg_general->rx_len > 0) {
			smcu_info->flag |= (1 << F_HOLD_RX);
		}else {
			smcu_info->flag &= ~(1 << F_HOLD_RX);
		}
		/* For activity command, we should attach the pkg to another
		 * dedicated place. So when the current pkg dettacked, it could
		 * be hanged until its answer reaches.
		 * There is only one activity command currently.
		 */
		switch(smcu_info->pkg_general->type) {
			case LED_SET_BI_BLK:
				/* The new blink command will overwrite the former one, even the
				 * former blinking is not finish yet.
				 */
				spin_lock_bh(smcu_info->lock);
				smcu_info->flag &= ~(1 << F_LED_IN_BLINK);
				spin_unlock_bh(smcu_info->lock);
				break;
			default:
				break;
		}

		/* Start transfer */
		while(to_be_write > 0) {
			if(to_be_write >= MCU_FIFO_SIZE)
				ct = MCU_FIFO_SIZE;
			else
				ct = to_be_write;
			to_be_write -= ct;

			spin_lock_bh(&smcu_info->lock);
			while(ct--) {
				smcu_info->tx_buff[smcu_info->tx_left++] = *tx_ptr++;
			}
			smcu_info->tx_head = smcu_info->tx_buff;
			/* new packet in buffer, hold the TX flag */
			smcu_info->flag |= (1 << F_HOLD_TX);
			smcu_info->flag &= ~(1 << F_ACK);
			smcu_info->flag &= ~(1 << F_PKG_RX);
			spin_unlock_bh(&smcu_info->lock);
			/* trigger the TX */
			tty->flags |= (1 << TTY_DO_WRITE_WAKEUP);
			smcu_write_wakeup(tty);

			/* sleep until ack received */
		}
		/* all data has been transfered, then we should check the Rx
		 * data if any*/
		spin_lock_bh(&smcu_info->lock);
		if(smcu_info->flag & (1 << F_HOLD_RX)) {
			if(!(smcu_info->flag & (1 << F_PKG_RX))) {
				printk(KERN_DEBUG "I am going to sleep for RX_PKG\n");
				spin_unlock_bh(&smcu_info->lock);
				wait_event_interruptible(smcu_info->read_queue_t, \
						   smcu_info->flag & (1 << F_PKG_RX));
			}else {
				spin_unlock_bh(&smcu_info->lock);
			}
		}else {
			spin_unlock_bh(&smcu_info->lock);
		}
		/* All expected data has been received, return */


	}else {			/* submode */
		smcu_info->flag &= ~(1 << F_PKG_RX);
		while(to_be_write > 0) {
			ct = to_be_write >TX_BUF_SIZE ? TX_BUF_SIZE : to_be_write;
			to_be_write -= ct;
			spin_lock_bh(&smcu_info->lock);
			while(ct--) {
				smcu_info->tx_buff[smcu_info->tx_left++] = *tx_ptr++;
			}
			smcu_info->tx_head = smcu_info->tx_buff;
			smcu_info->flag |= (1 << F_HOLD_TX);
			smcu_info->flag |= (1 << F_HOLD_RX);
			spin_unlock_bh(&smcu_info->lock);
			/* trigger the TX */
			tty->flags |= (1 << TTY_DO_WRITE_WAKEUP);
			smcu_write_wakeup(tty);
			/* sleep until all expected data sent */
			wait_event_interruptible(smcu_info->write_queue_t, \
						    !(smcu_info->flag &= (1 << F_HOLD_TX)));
		}
			/* all data has been sent successfully, then we should
			 * check the Rx */
			spin_lock_bh(&smcu_info->lock);
			if(!(smcu_info->flag & (1 << F_PKG_RX))) {
				spin_unlock_bh(&smcu_info->lock);
				wait_event_interruptible(smcu_info->read_queue_t, \
						    !(smcu_info->flag &= (1 << F_HOLD_RX)));
			}else {
				spin_unlock_bh(&smcu_info->lock);
			}

			/* all expected data has been received */
	}
#ifdef DUMP_PKG
			printk(KERN_DEBUG "%s --After package processing-------------\n",
			       __FUNCTION__);
			dump_pkg(pkg);
#endif

	/* dettach the package and return */
	smcu_info->pkg_general = NULL;

	return ret;

}

/*
 * flush all data in the buffer.
 * Please think it again before calling it.
 */
int mcu_flush(struct smcu * smcu_info)
{
	struct tty_struct *tty = smcu_info->tty;

	spin_lock_bh(&smcu_info->lock);
	/* something about the smcu buffer */
	smcu_info->flag = 0;
	smcu_info->rx_in_count = 0;
	smcu_info->rx_out_count = 0;
	smcu_info->tx_left = 0;
	smcu_info->rx_errors = 0;
	smcu_info->tx_head = smcu_info->tx_buff;

	spin_unlock_bh(&smcu_info->lock);

	if (tty->driver->flush_buffer)
		tty->driver->flush_buffer(tty);

	return 0;
}

/*
 * Free a smcu.
 */

static void smcu_free(struct smcu *smcu_info)
{
	kfree(smcu_info);
}

/* allocate one smcu */
static struct smcu *smcu_alloc(void)
{
	struct smcu *smcu_info;

	if(!(smcu_info = kmalloc(sizeof(struct smcu), GFP_KERNEL))) {
		return NULL;
	}
	smcu_info->magic = SMCU_MAGIC;
	spin_lock_init(&smcu_info->lock);
	spin_lock_init(&smcu_info->rcu_lock);
	sema_init(&smcu_info->sem, 1);
	init_waitqueue_head(&smcu_info->read_queue_t);
	init_waitqueue_head(&smcu_info->write_queue_t);
	smcu_info->mode = MODE_NORMAL;
	smcu_info->current_state = STA_MONITOR;
	smcu_info->pkg_general = NULL;
	smcu_info->pkg_led_bi_blk = NULL;
	smcu_info->rcu_handler = NULL;
	smcu_info->flag = 0;
	smcu_info->rx_in_count = 0;
	smcu_info->rx_out_count = 0;
	smcu_info->tx_left = 0;
	smcu_info->rx_errors = 0;
	smcu_info->tx_head = smcu_info->tx_buff;
	smcu_info->smcu_write = mcu_write;
	smcu_info->smcu_flush = mcu_flush;
	smcu_info->smcu_rcu_handler_register = mcu_rcu_handler_register;
	smcu_info->tty = NULL;

	return smcu_info;
}
/* called by tty */
static int smcu_open(struct tty_struct *tty)
{
	struct smcu *smcu_info = (struct smcu *) tty->disc_data;
	struct ktermios old_termios;
	/*
	 * First make sure we're not already connected.
	 */

	if (smcu_info && smcu_info->magic == SMCU_MAGIC)
		return -EEXIST;

	/*
	 * allocate one smcu object
	 */

	if ((smcu_info = smcu_alloc()) == NULL)
		return -ENFILE;

	/* store it to local global pointer */
	smcu_local_ptr = smcu_info;


	/*
	 * Maybe some led_dev will be registered here
	 *
	 */

	smcu_info->tty = tty;
	tty->disc_data = smcu_info;
	tty->receive_room = 65536;

	if (tty->driver->flush_buffer)
		tty->driver->flush_buffer(tty);

	/*
	 * Set tty options, Baudrate = 115200, 8 data + no parity + 1 stop
	 */
	old_termios = *(tty->termios);
	tty->termios->c_iflag &= ~PARENB;
	tty->termios->c_cflag &= ~CSTOPB;
	tty->termios->c_cflag &= ~CBAUD;	/* Clear the old baud setting */
	tty->termios->c_cflag |= B115200;	/* Set the new baud setting */
	if (tty->driver->set_termios)
		tty->driver->set_termios(tty, &old_termios);

	printk(KERN_INFO "SMCU: device activated\n");

	return 0;

}

/* called by tty module */
static void smcu_close(struct tty_struct *tty)
{
	struct smcu *smcu_info = (struct smcu *) tty->disc_data;

	/*
	 * First make sure we're connected.
	 */

	if (!smcu_info || smcu_info->magic != SMCU_MAGIC)
		return;

	tty->disc_data = NULL;
	smcu_info->tty = NULL;
	smcu_local_ptr = NULL;
	printk(KERN_INFO "SMCU: device closed down\n");
	smcu_free(smcu_info);
}

/* called by down low level driver */
static void smcu_receive_buf(struct tty_struct *tty,
			     const unsigned char *cp, char *fp, int count)
{
	struct smcu *smcu_info = (struct smcu *) tty->disc_data;
	int ct;
	int i;

	printk(KERN_DEBUG "%s\n", __FUNCTION__);

	if (!smcu_info || smcu_info->magic != SMCU_MAGIC)
		return;
	ct = count;
	/* Debug code */
#if 1
	printk(KERN_DEBUG "%s -- %d bytes received\n", __FUNCTION__, count);
	for(i = 0; i < count; i++) {
		printk(KERN_DEBUG "DUMP --byte[%d] = 0x%x\n", i, *(cp+i));
	}
#endif
	spin_lock_bh(&smcu_info->lock);
	/* Read the characters out of the buffer */
	while (ct--) {
#if 0
		/* discard the error data received */
		if (fp && *fp++) {
			printk(KERN_ERR "Error byte from UART(flag = %d)\n", *(fp-1));
			smcu_info->rx_errors++;
			cp++;
			continue;
		}
#endif
		/* protocol specific */
		if(smcu_info->mode == MODE_NORMAL) {
			switch (smcu_info->current_state) {
				/* Monitor any start code and ack */
				case STA_MONITOR:
					switch(*cp & 0xF0) {
						/* reply start code */
						case 0x10:
							if(smcu_info->pkg_general->rx_len > 1) {
								smcu_info->pkg_general->rx_buf[smcu_info-> \
									       pkg_general->rx_ct++] = *cp;
								smcu_info->current_state = STA_PARSING_GEN;
							}else if(smcu_info->pkg_general->rx_len > 0){
								/* event get, reply is one byte data */
								smcu_info->pkg_general->rx_buf[smcu_info-> \
									       pkg_general->rx_ct++] = *cp;
							    smcu_info->flag |= (1 << F_PKG_RX);
								wake_up_interruptible(&smcu_info->read_queue_t);
							}else {
								printk(KERN_ERR "%s Error-- unexpected data", __FUNCTION__);
							}

							break;
						/* synchronous event */
						case 0x20:
							/* led blinking finish */
							smcu_info->flag |= (1 << F_LED_IN_BLINK);
							break;
						/* asynchronous event */
						case 0x30:
							printk(KERN_DEBUG "Got RCU event\n");
							smcu_info->pkg_evt_rcu.rx_ct = 0;
							smcu_info->pkg_evt_rcu.rx_buf[smcu_info-> \
										 pkg_evt_rcu.rx_ct++] = *cp;
							smcu_info->current_state = STA_PARSING_EVT_RCU;
							smcu_info->pkg_evt_rcu.rx_len = 3;
							break;	/* end of 0x30 */

						/* Flow control */
						case 0x00:
							if(*cp ==0x00) {
								smcu_info->flag |= (1 << F_ACK);
								wake_up_interruptible(&smcu_info->\
												 write_queue_t);
							}
							break;
						default:
							break;
					}
					break;
				case STA_PARSING_GEN:
					smcu_info->pkg_general->rx_buf[smcu_info->\
								pkg_general->rx_ct++] = *cp;
					if(smcu_info->pkg_general->rx_ct == smcu_info->\
									     pkg_general->rx_len) {
						/* one pakcage received */
						printk(KERN_DEBUG "%s -- One gen_pkg received\n", __FUNCTION__);
						smcu_info->current_state = STA_MONITOR;
						smcu_info->flag |= (1 << F_PKG_RX);
						wake_up_interruptible(&smcu_info->read_queue_t);
					}
					break;
				case STA_PARSING_EVT_RCU:
					printk(KERN_DEBUG "Enter PARSING_EVT\n");
					smcu_info->pkg_evt_rcu.rx_buf[smcu_info->\
									pkg_evt_rcu.rx_ct++] = *cp;
					if(smcu_info->pkg_evt_rcu.rx_ct == smcu_info->
									      pkg_evt_rcu.rx_len) {
						/* one package received */
						smcu_info->current_state = STA_MONITOR;
						/* Trigger the event process */
						if(smcu_info->rcu_handler) {
							smcu_info->rcu_handler(smcu_info);
						}
					}
					break;

				default:
					break;
			}
		}else {	/* for submode */
			if(smcu_info->pkg_general->rx_ct < smcu_info->pkg_general->rx_len) {
				smcu_info->pkg_general->rx_buf[smcu_info->pkg_general->rx_ct++] = *cp;
			}else {
				smcu_info->flag |= (1 << F_PKG_RX);
				wake_up_interruptible(&smcu_info->read_queue_t);
			}
		} /* end of submode */
		cp++;
	}
	spin_unlock_bh(&smcu_info->lock);
}

/* register a rcu handler a smcu object */
static int mcu_rcu_handler_register(struct smcu * smcu_info,
				    int (*handler)(struct smcu *))
{
	int ret = 0;
	smcu_info->rcu_handler = handler;

	return ret;
}

/*
 * called by down low level driver
 */
static void smcu_write_wakeup(struct tty_struct *tty)
{
	struct smcu *smcu_info = (struct smcu *) tty->disc_data;
	int	num_written;

	/* First make sure we're connected. */
	if (!smcu_info || smcu_info->magic != SMCU_MAGIC)
		return;
	spin_lock_bh(&smcu_info->lock);
	if (smcu_info->tx_left > 0) {
		num_written =
		    tty->driver->write(tty, smcu_info->tx_head,
				      smcu_info->tx_left);
		smcu_info->tx_left -= num_written;
		smcu_info->tx_head += num_written;
	} else {	/* Else start transmission of another packet */
		tty->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);
		smcu_info->flag &= ~(1 << F_HOLD_TX);
		if(smcu_info->current_state == MODE_SUB) {
			wake_up_interruptible(&smcu_info->write_queue_t);
		}
	}
	spin_unlock_bh(&smcu_info->lock);
}

struct tty_ldisc smcu_ldisc = {
	.owner		= THIS_MODULE,
	.magic		= TTY_LDISC_MAGIC,
	.name		= "s_mcu",
	.open		= smcu_open,
	.close		= smcu_close,
	.receive_buf	= smcu_receive_buf,
	.write_wakeup	= smcu_write_wakeup,
};

#ifdef DUMP_PKG
static void dump_pkg(pkg_general_t *pkg)
{
	int i;
	printk(KERN_DEBUG "pkg.tx_len == %d\n", pkg->tx_len);
	for(i = 0; i < pkg->tx_len; i++) {
		printk(KERN_DEBUG "pkg.tx_buf[%d] == 0x%x\n", i, pkg->tx_buf[i]);
	}
	printk(KERN_DEBUG "\n");
	printk(KERN_DEBUG "pkg.rx_len == %d\n", pkg->rx_len);
	printk(KERN_DEBUG "pkg.rx_ct == %d\n", pkg->rx_ct);
	if(pkg->rx_ct > 0) {
		for(i = 0; i < pkg->rx_ct; i++) {
			printk(KERN_DEBUG "pkg.rx_buf[%d] == 0x%x\n", i, pkg->rx_buf[i] );
		}
		printk(KERN_DEBUG "\n");
	}
}
#endif
