/*
 * (C) Copyright 2009 WyPlay SAS.
 * Frederic Mazuel <fmazuel@wyplay.com>
 * Jerry Zhang <zhangb@wyplayasia.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/ctype.h>
#include <linux/sysfs.h>
#include <linux/firmware.h>
#include <linux/wymcu/mcu_service.h>
#include <linux/wymcu/mcu_ldisc.h>
#include "./mcu_serial.h"

struct device *tmp_dev;
static struct mcu_service *services;
static dev_t mcuserv_dev_number, mcu_power_dev_number, mcu_reset_dev_number;
static struct class *mcuserv_class;
static struct class_device *mcuserv_class_dev, *mcu_power_class_dev, *mcu_reset_class_dev;

/* for wakeup */
unsigned char local_wake_up_keycodes;

/* for mcu firmware upgrade */

static struct state boot_state[] = {
	{"standby_key", REG_WUP_MSK_MSK_KEY},
	{"WOL", REG_WUP_MSK_MSK_WOL},
	{"RF", REG_WUP_MSK_MSK_RF},
	{"timer", REG_WUP_MSK_MSK_TMR},
	{"IR", REG_WUP_MSK_MSK_IR},
	{"CEC", REG_WUP_MSK_MSK_CEC},
	{NULL, 0}
};
static struct state event_state[] = {
	{"enable", REG_STA_EVT_ENA},
	{"disable", REG_STA_EVT_DSA},
	{NULL, 0}
};
static struct state power_state[] = {
	{"on", REG_PWR_ON},
	{"off", REG_PWR_OFF},
	{NULL, 0}
};
static int get_st_value(struct state *st, char *st_name)
{
	struct state *pt = st;
	int val = -1;

	while (pt->name != NULL) {
		if(!strcmp(st_name, pt->name)) {
			val = pt->val;
			break;
		}
		pt++;
	}

	return val;
}

static char *get_st_name(struct state *st, int val)
{
	struct state *pt = st;

	while (pt->name != NULL) {
		if(pt->val == val) {
			break;
		}
		pt++;
	}

	return pt->name;
}


/* firmware version */
ssize_t show_mcu_firmware_version(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int ret;
	char tx_buffer[5], rx_buffer[10];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	/* Read the firmware version */
	pkg.type = VER_GET;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	ret = services->smcu_transfer(&pkg, mode);
	if(ret) {
		pos += sprintf(pos, "%s:%d\n", "Error", ret);
		return pos - buf ;
	}
	/* Update the status buffer */
	pos += sprintf(pos, "%d:%d:%d:%d\n", rx_buffer[1], \
		       rx_buffer[2],					\
		       rx_buffer[3],					\
		       rx_buffer[4]);

	return pos - buf;
}

static CLASS_DEVICE_ATTR(firmware_version, S_IRUGO,
			 show_mcu_firmware_version, NULL);
/* Boot origin */
ssize_t show_mcu_boot_origin(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int ret, value;
	char *name;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	/* Read boot status */
	pkg.type = WUP_GET_LST_SRC;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	ret = services->smcu_transfer(&pkg, mode);
	if(ret) {
		pos += sprintf(pos, "%s:%d\n", "Error", ret);
		return pos - buf;
	}
	/* Update the status buffer */
	value = rx_buffer[1] & REG_WUP_SRC_MSK_ALL;
	name = get_st_name(boot_state, value);
	if(name) {
		pos += sprintf(pos, "%s\n", name);
	}else {
		/* shoule be normal ? */
		pos += sprintf(pos, "%s\n", "NORMAL");
	}
	return pos - buf;
}
static CLASS_DEVICE_ATTR(wakeup_source, S_IRUGO,
			show_mcu_boot_origin, NULL);
/* event state */
ssize_t show_mcu_event_state(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int ret, value;
	char *name;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	/* Read event status */
	pkg.type = EVT_GET_STA;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	ret = services->smcu_transfer(&pkg, mode);
	if(ret) {
		pos += sprintf(pos, "%s:%d\n", "Error", ret);
		return pos - buf;
	}
	/* Update the status buffer */
	value = rx_buffer[0] & 0x03;
	name = get_st_name(event_state, value);
	pos += sprintf(pos, "%s\n", name);

	return pos - buf;
}

ssize_t store_mcu_event_state(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res, value;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = get_st_value(event_state, name);
	if(value < 0) {
		res = -EINVAL;
		goto exit;
	}
	if(!value) {
		pkg.type = EVT_SET_DIS;
	}else {
		pkg.type = EVT_SET_ENA;
	}
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(event_state, S_IRUGO | S_IWUSR,
			show_mcu_event_state, store_mcu_event_state);
/* wakeup mask */
ssize_t show_mcu_wakeup_mask(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int ret, value;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	/* Read event status */
	pkg.type = WUP_GET_IPT_SRC_MSK;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	ret = services->smcu_transfer(&pkg, mode);
	if(ret) {
		pos += sprintf(pos, "%s:%d\n", "Error", ret);
		return pos - buf;
	}
	/* Update the status buffer */
	value = rx_buffer[1];
	pos += sprintf(pos, "%d\n", value);

	return pos - buf;
}

ssize_t store_mcu_wakeup_mask(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res, value;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = simple_strtoul(name, NULL, 10);
	pkg.type = WUP_SET_IPT_SRC_MSK;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	tx_buffer[1] = 0x3F;
	tx_buffer[2] = value;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(wakeup_mask, S_IRUGO | S_IWUSR,
			show_mcu_wakeup_mask, store_mcu_wakeup_mask);
/* wakeup keycodes */
ssize_t show_mcu_wakeup_keycodes(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int value;

	/* Read boot status */
	value = local_wake_up_keycodes;
	pos += sprintf(pos, "%d\n", value);

	return pos - buf;
}
ssize_t store_mcu_wakeup_keycodes(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res, value;
	char reg;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = simple_strtoul(name, NULL, 10);
	reg = value;
	/* store it, to be uset by read */
	local_wake_up_keycodes = value;

	tx_buffer[1] = reg;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	pkg.type = WUP_SET_ADD_AUT_KEY_COD;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(wakeup_keycodes, S_IRUGO | S_IWUSR,
			show_mcu_wakeup_keycodes, store_mcu_wakeup_keycodes);

/* Power, board */
ssize_t store_mcu_board_power(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res, value;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = get_st_value(power_state, name);
	if(value < 0) {
		res = -EINVAL;
		goto exit;
	}
	if(value != REG_PWR_OFF) {
		res = -EINVAL;
		goto exit;
	}
	pkg.type = PWR_SET_SOC_OFF;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	/* actually, maybe it has been dead, maybe it will not come to here */
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);

exit:
	return res;
}
static CLASS_DEVICE_ATTR(board, S_IWUSR,
			NULL, store_mcu_board_power);
/* Power, display(VFD) */
ssize_t show_mcu_display_power(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int ret, value;
	unsigned char reg;
	char *name;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	/* Read status */
	pkg.type = PWR_GET_GRP_STA;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	ret = services->smcu_transfer(&pkg, mode);
	if(ret) {
		pos += sprintf(pos, "%s:%d\n", "Error", ret);
		return pos - buf;
	}
	/* Update the status buffer */
	reg = rx_buffer[1];
	if(reg & REG_PWR_MSK_VFD) {
		value = REG_PWR_ON;
	}else {
		value = REG_PWR_OFF;
	}
	name = get_st_name(power_state, value);
	pos += sprintf(pos, "%s\n", name);

	return pos - buf;
}
ssize_t store_mcu_display_power(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res, value;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = get_st_value(power_state, name);
	if(value < 0) {
		res = -EINVAL;
		goto exit;
	}
	if(value == REG_PWR_ON) {
		pkg.type = PWR_SET_VFD_ON;
	}else {
		pkg.type = PWR_SET_VFD_OFF;
	}
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(display, S_IRUGO | S_IWUSR,
			show_mcu_display_power, store_mcu_display_power);
/* Power, DVB-T */
ssize_t show_mcu_dvbt_power(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int ret, value;
	unsigned char reg;
	char *name;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	/* Read status */
	pkg.type = PWR_GET_GRP_STA;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	ret = services->smcu_transfer(&pkg, mode);
	if(ret) {
		pos += sprintf(pos, "%s:%d\n", "Error", ret);
		return pos - buf;
	}
	/* Update the status buffer */
	reg = rx_buffer[1];
	if(reg & REG_PWR_MSK_DVT) {
		value = REG_PWR_ON;
	}else {
		value = REG_PWR_OFF;
	}
	name = get_st_name(power_state, value);
	pos += sprintf(pos, "%s\n", name);

	return pos - buf;
}
ssize_t store_mcu_dvbt_power(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res, value;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = get_st_value(power_state, name);
	if(value < 0) {
		res = -EINVAL;
		goto exit;
	}
	if(value == REG_PWR_ON) {
		pkg.type = PWR_SET_DVT_TUN_ON;
	}else {
		pkg.type = PWR_SET_DVT_TUN_OFF;
	}
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(DVB_T, S_IRUGO | S_IWUSR,
			show_mcu_dvbt_power, store_mcu_dvbt_power);
/* Power, DVB-S2 */
ssize_t show_mcu_dvbs2_power(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int ret, value;
	unsigned char reg;
	char *name;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	/* Read status */
	pkg.type = PWR_GET_GRP_STA;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	ret = services->smcu_transfer(&pkg, mode);
	if(ret) {
		pos += sprintf(pos, "%s:%d\n", "Error", ret);
		return pos - buf;
	}
	/* Update the status buffer */
	reg = rx_buffer[1];
	if(reg & REG_PWR_MSK_DS2) {
		value = REG_PWR_ON;
	}else {
		value = REG_PWR_OFF;
	}
	name = get_st_name(power_state, value);
	pos += sprintf(pos, "%s\n", name);

	return pos - buf;
}
ssize_t store_mcu_dvbs2_power(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res, value;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = get_st_value(power_state, name);
	if(value < 0) {
		res = -EINVAL;
		goto exit;
	}
	if(value == REG_PWR_ON) {
		pkg.type = PWR_SET_DS2_NIM_ON;
	}else {
		pkg.type = PWR_SET_DS2_NIM_OFF;
	}
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(DVB_S2, S_IRUGO | S_IWUSR,
			show_mcu_dvbs2_power, store_mcu_dvbs2_power);
/* Power, HDD */
ssize_t show_mcu_hdd_power(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int ret, value;
	unsigned char reg;
	char *name;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	/* Read status */
	pkg.type = PWR_GET_GRP_STA;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	ret = services->smcu_transfer(&pkg, mode);
	if(ret) {
		pos += sprintf(pos, "%s:%d\n", "Error", ret);
		return pos - buf;
	}
	/* Update the status buffer */
	reg = rx_buffer[1];
	if(reg & REG_PWR_MSK_HDD) {
		value = REG_PWR_ON;
	}else {
		value = REG_PWR_OFF;
	}
	name = get_st_name(power_state, value);
	pos += sprintf(pos, "%s\n", name);

	return pos - buf;
}
ssize_t store_mcu_hdd_power(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res, value;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = get_st_value(power_state, name);
	if(value < 0) {
		res = -EINVAL;
		goto exit;
	}
	if(value == REG_PWR_ON) {
		pkg.type = PWR_SET_HDD_ON;
	}else {
		pkg.type = PWR_SET_HDD_OFF;
	}
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(HDD, S_IRUGO | S_IWUSR,
			show_mcu_hdd_power, store_mcu_hdd_power);
/* Power, Wifi */
ssize_t show_mcu_wifi_power(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int ret, value;
	unsigned char reg;
	char *name;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	/* Read status */
	pkg.type = PWR_GET_GRP_STA;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	ret = services->smcu_transfer(&pkg, mode);
	if(ret) {
		pos += sprintf(pos, "%s:%d\n", "Error", ret);
		return pos - buf;
	}
	/* Update the status buffer */
	reg = rx_buffer[1];
	if(reg & REG_PWR_MSK_WIF) {
		value = REG_PWR_ON;
	}else {
		value = REG_PWR_OFF;
	}
	name = get_st_name(power_state, value);
	pos += sprintf(pos, "%s\n", name);

	return pos - buf;
}
ssize_t store_mcu_wifi_power(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res, value;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = get_st_value(power_state, name);
	if(value < 0) {
		res = -EINVAL;
		goto exit;
	}
	if(value == REG_PWR_ON) {
		pkg.type = PWR_SET_WFI_CRD_ON;
	}else {
		pkg.type = PWR_SET_WFI_CRD_OFF;
	}
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(wifi, S_IRUGO | S_IWUSR,
			show_mcu_wifi_power, store_mcu_wifi_power);
/* Power ethernet */
ssize_t show_mcu_eth_power(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int ret, value;
	unsigned char reg;
	char *name;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	/* Read status */
	pkg.type = PWR_GET_GRP_STA;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	ret = services->smcu_transfer(&pkg, mode);
	if(ret) {
		pos += sprintf(pos, "%s:%d\n", "Error", ret);
		return pos - buf;
	}
	/* Update the status buffer */
	reg = rx_buffer[1];
	if(reg & REG_PWR_MSK_ETH) {
		value = REG_PWR_ON;
	}else {
		value = REG_PWR_OFF;
	}
	name = get_st_name(power_state, value);
	pos += sprintf(pos, "%s\n", name);

	return pos - buf;
}
ssize_t store_mcu_eth_power(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res, value;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = get_st_value(power_state, name);
	if(value < 0) {
		res = -EINVAL;
		goto exit;
	}
	if(value == REG_PWR_ON) {
		pkg.type = PWR_SET_ETH_ON;
	}else {
		pkg.type = PWR_SET_ETH_OFF;
	}
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(Ethernet, S_IRUGO | S_IWUSR,
			show_mcu_eth_power, store_mcu_eth_power);
/* Power, VIXS */
ssize_t show_mcu_vixs_power(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int ret, value;
	unsigned char reg;
	char *name;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	/* Read status */
	pkg.type = PWR_GET_GRP_STA;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	ret = services->smcu_transfer(&pkg, mode);
	if(ret) {
		pos += sprintf(pos, "%s:%d\n", "Error", ret);
		return pos - buf;
	}
	/* Update the status buffer */
	reg = rx_buffer[1];
	if(reg & REG_PWR_MSK_TRC) {
		value = REG_PWR_ON;
	}else {
		value = REG_PWR_OFF;
	}
	name = get_st_name(power_state, value);
	pos += sprintf(pos, "%s\n", name);

	return pos - buf;
}
ssize_t store_mcu_vixs_power(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res, value;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = get_st_value(power_state, name);
	if(value < 0) {
		res = -EINVAL;
		goto exit;
	}
	if(value == REG_PWR_ON) {
		pkg.type = PWR_SET_TRC_FUN_ON;
	}else {
		pkg.type = PWR_SET_TRC_FUN_OFF;
	}
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(VIXS, S_IRUGO | S_IWUSR,
			show_mcu_vixs_power, store_mcu_vixs_power);
/* Power, PLC */
ssize_t show_mcu_plc_power(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int ret, value;
	unsigned char reg;
	char *name;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	/* Read status */
	pkg.type = PWR_GET_GRP_STA;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	ret = services->smcu_transfer(&pkg, mode);
	if(ret) {
		pos += sprintf(pos, "%s:%d\n", "Error", ret);
		return pos - buf;
	}
	/* Update the status buffer */
	reg = rx_buffer[1];
	if(reg & REG_PWR_MSK_PLC) {
		value = REG_PWR_ON;
	}else {
		value = REG_PWR_OFF;
	}
	name = get_st_name(power_state, value);
	pos += sprintf(pos, "%s\n", name);

	return pos - buf;
}
ssize_t store_mcu_plc_power(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res, value;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = get_st_value(power_state, name);
	if(value < 0) {
		res = -EINVAL;
		goto exit;
	}
	if(value == REG_PWR_ON) {
		pkg.type = PWR_SET_PLC_ON;
	}else {
		pkg.type = PWR_SET_PLC_OFF;
	}
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(PLC, S_IRUGO | S_IWUSR,
			show_mcu_plc_power, store_mcu_plc_power);
/* Reset, board */
ssize_t store_mcu_board_reset(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	if(strcmp("reset", name)) {
		res = -EINVAL;
		goto exit;
	}
	pkg.type = PWR_SET_RST_BOD;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(Board, S_IWUSR,
			NULL, store_mcu_board_reset);
/* Reset, display */
ssize_t store_mcu_display_reset(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	if(strcmp("reset", name)) {
		res = -EINVAL;
		goto exit;
	}
	pkg.type = PWR_SET_RST_VFD;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(Display, S_IWUSR,
			NULL, store_mcu_display_reset);
/* Reset, RF_RCU */
ssize_t store_mcu_rcu_reset(struct class_device *cdev, const char *buf, size_t count)
{
	char name[10];
	int	res;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	res = sscanf(buf, "%s", name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	if(strcmp("reset", name)) {
		res = -EINVAL;
		goto exit;
	}
	pkg.type = PWR_SET_RST_RF;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(RF_RCU, S_IWUSR,
			NULL, store_mcu_rcu_reset);
/* Upgrade firmware */
ssize_t store_upgrade_firmware(struct class_device *cdev, const char *buf, size_t count)
{
	const struct firmware *fw_entry;
	char firm_name[20];
	int res, ct, tx_left;
	char *tx_head;
	char tx_buffer[5], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode;

	res = sscanf(buf, "%s", firm_name);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	/* Step 0, aquire the data first */
	if (request_firmware(&fw_entry, firm_name, tmp_dev)!=0) {
		printk(KERN_ERR "%s -- Request_firmware failed\n", __FUNCTION__);
		res = -1;
		goto exit;
	}
	/* The smcu should be put to submode firstly, and then
	 * step 1: Upgrade Enter.
	 */
	pkg.type = SUB_UPG_FRM_ENT;
	mode = MODE_SUB;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		res = -EIO;
		goto exit;
	}
	if(pkg.rx_buf[0] == CODE_NACK) {
		printk(KERN_ERR "Upgrade: Enter the upgrade firmware failed\n");
		res = -EIO;
		goto exit;
	}
	/* step 2: Upgrade Send Payload */
	tx_left = fw_entry->size;
	tx_head = fw_entry->data;
	while(tx_left > 0) {
		ct = tx_left > MCU_FLASH_PAGE_SIZE ? MCU_FLASH_PAGE_SIZE : tx_left;
		pkg.type = INITIALIZED_PKG;
		mode = MODE_SUB;
		pkg.tx_buf = tx_head;
		pkg.tx_len = ct;
		pkg.rx_len = 1;
		pkg.rx_ct = 0;
		pkg.rx_buf = rx_buffer;
		res = services->smcu_transfer(&pkg, mode);
		if(res < 0) {
			res = -EIO;
			goto exit;
		}
		if(pkg.rx_buf[0] == CODE_NACK) {
			printk(KERN_ERR "Upgrade: Send Payload failed\n");
			res = -EIO;
			goto exit;
		}
		tx_left -= ct;
		tx_head += ct;
	}
	release_firmware(fw_entry);
	/* step 3: Upgrade Apply */
	pkg.type = SUB_UPG_FRM_APL;
	mode = MODE_SUB;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		res = -EIO;
		goto exit;
	}
	printk(KERN_DEBUG "SMCU: Upgrade Succed\n");
	/* Actually, maybe we can not reach here, because the Soc will be reset */
	res = strnlen(buf, count);
exit:
	return res;

}
static CLASS_DEVICE_ATTR(upgrade, S_IWUGO,
			NULL, store_upgrade_firmware);

/* Death letter message */
ssize_t show_death_msg(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int ret;
	char tx_buffer[5], rx_buffer[10];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	memset(rx_buffer, 0, sizeof(rx_buffer));
	/* Read status */
	pkg.type = DTH_LET_RCV;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	ret = services->smcu_transfer(&pkg, mode);
	if(ret) {
		pos += sprintf(pos, "%d\n", ret);
		return pos - buf;
	}
	/* Update the status buffer */
	pos += sprintf(pos, "%s\n", &rx_buffer[1]);

	return pos - buf;
}

ssize_t store_death_msg(struct class_device *cdev, const char *buf,
			size_t count)
{
	int	res;
	char tx_buffer[15], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	if(count > 9) {
		res = -EINVAL;
		goto exit;
	}
	memset(tx_buffer, 0, sizeof(tx_buffer));
	strncpy(&tx_buffer[1], buf, count - 1);
	pkg.type = DTH_LET_PST;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);

 exit:
	return res;
 }
 static CLASS_DEVICE_ATTR(deathmsg, S_IRUGO | S_IWUSR,
			show_death_msg, store_death_msg);

static int mcu_service_probe(struct mcu_service_dev *dev)
{
	int ret = 0;

	printk(KERN_INFO "MCU service drv instance is starting\n");
	services = (struct mcu_service *)(dev->service);
	if(services == NULL) {
		printk(KERN_WARNING "no mcu services\n");
		goto exit;
	}
	/* Create class device: mcuserv0 */
	mcuserv_class_dev = class_device_create (mcuserv_class, NULL,
						 mcuserv_dev_number, &dev->dev,
						 "mcuserv.0");
	tmp_dev = &dev->dev;
	if (IS_ERR(mcuserv_class_dev)) {
		printk( KERN_ERR
			"%s: ERROR: Creating mcuserv0 class device: %ld)\n",
			__FUNCTION__, PTR_ERR(mcuserv_class_dev));
		ret = PTR_ERR(mcuserv_class_dev);
		goto exit;
	}
	/* create class device: power */
	mcu_power_class_dev = class_device_create (	mcuserv_class,
							mcuserv_class_dev,
							mcuserv_dev_number,
							NULL,
							"power");
	if (IS_ERR(mcu_power_class_dev)) {
		printk( KERN_ERR
			"%s: ERROR: Creating power class device: %ld)\n",
			__FUNCTION__, PTR_ERR(mcu_power_class_dev));
		ret = PTR_ERR(mcu_power_class_dev);
		goto err1;
	}
	/* create class device: reset */
	mcu_reset_class_dev = class_device_create (mcuserv_class, mcuserv_class_dev,
						 mcuserv_dev_number, NULL,
						 "reset");
	if (IS_ERR(mcu_reset_class_dev)) {
		printk( KERN_ERR
			"%s: ERROR: Creating mcureset class device: %ld)\n",
			__FUNCTION__, PTR_ERR(mcu_reset_class_dev));
		ret = PTR_ERR(mcu_reset_class_dev);
		goto err2;
	}
	/* firmware version */
	ret = class_device_create_file(	mcuserv_class_dev,
					&class_device_attr_firmware_version);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu service firmware version file (%d)\n",
			__FUNCTION__, ret);
		goto err3;
	}
	/* mcu boot origin */
	ret = class_device_create_file(mcuserv_class_dev,
						&class_device_attr_wakeup_source);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu service wakeup_source file (%d)\n",
			__FUNCTION__, ret);
		goto err4;
	}
	/* mcu event state */
	ret = class_device_create_file(mcuserv_class_dev,
						&class_device_attr_event_state);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu service event_state file (%d)\n",
			__FUNCTION__, ret);
		goto err5;
	}
	/* mcu wakeup keycodes */
	ret = class_device_create_file(mcuserv_class_dev,
						&class_device_attr_wakeup_keycodes);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu service wakeup_keycodes file (%d)\n",
			__FUNCTION__, ret);
		goto err6;
	}
	/* mcu board power */
	ret = class_device_create_file(mcu_power_class_dev,
						&class_device_attr_board);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu power board file (%d)\n",
			__FUNCTION__, ret);
		goto err7;
	}
	/* mcu display power */
	ret = class_device_create_file(mcu_power_class_dev,
						&class_device_attr_display);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu power display file (%d)\n",
			__FUNCTION__, ret);
		goto err8;
	}
	/* mcu dvb-t power */
	ret = class_device_create_file(mcu_power_class_dev,
						&class_device_attr_DVB_T);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu power DVB-T file (%d)\n",
			__FUNCTION__, ret);
		goto err9;
	}
	/* mcu dvb-S2 power */
	ret = class_device_create_file(mcu_power_class_dev,
						&class_device_attr_DVB_S2);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu power DVB-S2 file (%d)\n",
			__FUNCTION__, ret);
		goto err10;
	}
	/* mcu HDD power */
	ret = class_device_create_file(mcu_power_class_dev,
						&class_device_attr_HDD);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu power HDD file (%d)\n",
			__FUNCTION__, ret);
		goto err11;
	}
	/* mcu WiFi power */
	ret = class_device_create_file(mcu_power_class_dev,
						&class_device_attr_wifi);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu power wifi file (%d)\n",
			__FUNCTION__, ret);
		goto err12;
	}
	/* mcu ethernet power */
	ret = class_device_create_file(mcu_power_class_dev,
						&class_device_attr_Ethernet);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu power Ethernet file (%d)\n",
			__FUNCTION__, ret);
		goto err13;
	}
	/* mcu VIXS power */
	ret = class_device_create_file(mcu_power_class_dev,
						&class_device_attr_VIXS);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu power VIXS file (%d)\n",
			__FUNCTION__, ret);
		goto err14;
	}
	/* mcu PLC power */
	ret = class_device_create_file(mcu_power_class_dev,
						&class_device_attr_PLC);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu power PLC file (%d)\n",
			__FUNCTION__, ret);
		goto err15;
	}
	/* mcu board reset */
	ret = class_device_create_file(mcu_reset_class_dev,
						&class_device_attr_Board);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu reset Board file (%d)\n",
			__FUNCTION__, ret);
		goto err16;
	}
	/* mcu display(VFD) reset */
	ret = class_device_create_file(mcu_reset_class_dev,
						&class_device_attr_Display);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu reset Board file (%d)\n",
			__FUNCTION__, ret);
		goto err17;
	}
	/* mcu RF_RCU reset */
	ret = class_device_create_file(mcu_reset_class_dev,
						&class_device_attr_RF_RCU);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu power Board file (%d)\n",
			__FUNCTION__, ret);
		goto err18;
	}
	/* mcu upgrade */
	ret = class_device_create_file(mcuserv_class_dev,
						&class_device_attr_upgrade);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu power Board file (%d)\n",
			__FUNCTION__, ret);
		goto err19;
	}
	/* wakeup_mask */
	ret = class_device_create_file(mcuserv_class_dev,
						&class_device_attr_wakeup_mask);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat mcu power Board file (%d)\n",
			__FUNCTION__, ret);
		goto err20;
	}
	/* deathmsg */
	ret = class_device_create_file(mcuserv_class_dev,
						&class_device_attr_deathmsg);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creat deathmsg file (%d)\n",
			__FUNCTION__, ret);
		goto err21;
	}
	goto exit;
err21:
	class_device_remove_file(mcuserv_class_dev,
				 &class_device_attr_wakeup_mask);
err20:
	class_device_remove_file(mcu_reset_class_dev, \
				 &class_device_attr_upgrade);
err19:
	class_device_remove_file(mcu_reset_class_dev, \
				 &class_device_attr_RF_RCU);
err18:
	class_device_remove_file(mcu_reset_class_dev, \
				 &class_device_attr_Display);
err17:
	class_device_remove_file(mcu_reset_class_dev, \
				 &class_device_attr_Board);
err16:
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_PLC);
err15:
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_VIXS);
err14:
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_Ethernet);
err13:
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_wifi);
err12:
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_HDD);
err11:
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_DVB_S2);
err10:
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_DVB_T);
err9:
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_display);
err8:
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_board);
err7:
	class_device_remove_file(mcuserv_class_dev, \
				 &class_device_attr_wakeup_keycodes);
err6:
	class_device_remove_file(mcuserv_class_dev, \
				 &class_device_attr_event_state);
err5:
	class_device_remove_file(mcuserv_class_dev, \
				 &class_device_attr_wakeup_source);
err4:
	class_device_remove_file(mcuserv_class_dev, \
				 &class_device_attr_firmware_version);
err3:
	class_device_del(mcu_reset_class_dev);
err2:
	class_device_del(mcu_power_class_dev);
err1:
	class_device_del(mcuserv_class_dev);
exit:
	return ret;
}

static int mcu_service_remove(struct mcu_service_dev *dev)
{
	printk(KERN_DEBUG "Mcu service drv instance is stopping\n");
	class_device_remove_file(mcuserv_class_dev, \
				 &class_device_attr_deathmsg);
	class_device_remove_file(mcuserv_class_dev, \
				 &class_device_attr_wakeup_mask);
	class_device_remove_file(mcuserv_class_dev, \
				 &class_device_attr_upgrade);
	class_device_remove_file(mcu_reset_class_dev, \
				 &class_device_attr_RF_RCU);
	class_device_remove_file(mcu_reset_class_dev, \
				 &class_device_attr_Display);
	class_device_remove_file(mcu_reset_class_dev, \
				 &class_device_attr_Board);
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_PLC);
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_VIXS);
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_Ethernet);
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_wifi);
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_HDD);
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_DVB_S2);
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_DVB_T);
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_display);
	class_device_remove_file(mcu_power_class_dev, \
				 &class_device_attr_board);
	class_device_remove_file(mcuserv_class_dev, \
				 &class_device_attr_wakeup_keycodes);
	class_device_remove_file(mcuserv_class_dev, \
				 &class_device_attr_event_state);
	class_device_remove_file(mcuserv_class_dev, \
				 &class_device_attr_wakeup_source);
	class_device_remove_file(mcuserv_class_dev, \
				 &class_device_attr_firmware_version);
	class_device_del(mcu_reset_class_dev);
	class_device_del(mcu_power_class_dev);
	class_device_del(mcuserv_class_dev);
	return 0;
}

struct mcu_service_driver mcu_serv_driver = {
	.driver = {
		.owner	= THIS_MODULE,
		.name	= "service_mcu",
	},
	.service_type = SERVICE_MCU,
	.probe = mcu_service_probe,
	.remove = mcu_service_remove,
};

static int __init mcu_service_init(void)
{
	int ret;

	printk(KERN_INFO "MCU service is registering\n");
	/* allocate class device number */
	ret = alloc_chrdev_region (&mcuserv_dev_number, 0, 1, "mcuservice");
	if (ret < 0) {
		printk (KERN_ERR
			"%s: mcu service can't register device err: 0x%x\n",
			__FUNCTION__, ret);
		goto exit;
	}
	ret = alloc_chrdev_region (&mcuserv_dev_number, 0, 1, "mcupower");
	if (ret < 0) {
		printk (KERN_ERR
			"%s: mcu power can't register device err: 0x%x\n",
			__FUNCTION__, ret);
		goto err1;
	}
	ret = alloc_chrdev_region (&mcuserv_dev_number, 0, 1, "mcureset");
	if (ret < 0) {
		printk (KERN_ERR
			"%s: mcu reset can't register device err: 0x%x\n",
			__FUNCTION__, ret);
		goto err2;
	}
	mcuserv_class = class_create (THIS_MODULE, "mcuserv");
	if (IS_ERR(mcuserv_class)) {
		printk( KERN_ERR
			"%s: ERROR: Creating mcuserv class: %ld)\n",
			__FUNCTION__, PTR_ERR(mcuserv_class));
		ret = PTR_ERR(mcuserv_class);
		goto err3;
	}
	ret = mcu_service_driver_register(&mcu_serv_driver);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't register the MCU service sound driver: 0x%x\n",
			__FUNCTION__, ret);
		goto err4;
	}

	goto exit;

err4:
	class_destroy(mcuserv_class);
err3:
	unregister_chrdev_region(mcu_reset_dev_number, 1);
err2:
	unregister_chrdev_region(mcu_power_dev_number, 1);
err1:
	unregister_chrdev_region(mcuserv_dev_number, 1);
exit:
	return ret;
}

static void __exit mcu_service_exit(void)
{
	printk(KERN_DEBUG "mcu service is unregistering\n");
	mcu_service_driver_unregister(&mcu_serv_driver);
	class_destroy(mcuserv_class);
	unregister_chrdev_region(mcu_reset_dev_number, 1);
	unregister_chrdev_region(mcu_power_dev_number, 1);
	unregister_chrdev_region(mcuserv_dev_number, 1);
}

module_init(mcu_service_init);
module_exit(mcu_service_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("mcu service device support");
MODULE_AUTHOR("Frederic Mazuel <fmazuel@wyplay.com> Jerry Zhang <zhangb@wyplayasia.com>");
