/*
 * (C) Copyright 2009 WyPlay SAS.
 * Frederic Mazuel <fmazuel@wyplay.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/ctype.h>
#include <linux/fs.h>
#include <linux/leds.h>
#include <linux/wymcu/mcu_service.h>

struct led_dev_ctx {
	unsigned int id;
	char name[20];
	struct led_classdev *cdev;
	struct mcu_service_dev *dev;
};

static struct led_dev_ctx ledc[MAX_MCU_LED];

static void set_brightness(struct led_classdev *led_cdev,
			   enum led_brightness brightness);

static void ledc_init(void)
{
	int i = MAX_MCU_LED;
	while(i--) {
		ledc[i].id = 0xFF;
	}
}

static struct led_dev_ctx *get_ledc_from_cdev(struct led_classdev *cdev)
{
	int i;
	for(i = 0; i < MAX_MCU_LED; i++) {
		if(ledc[i].cdev == cdev) {
			return &ledc[i];
		}
	}
	return NULL;
}

static struct led_dev_ctx *get_ledc_from_dev(struct mcu_service_dev *dev)
{
	int i;
	for(i = 0; i < MAX_MCU_LED; i++) {
		if(ledc[i].dev == dev) {
			return &ledc[i];
		}
	}
	return NULL;
}

static struct led_dev_ctx *get_dev_ledc(struct mcu_service_dev *dev)
{
	int i;
	struct led_classdev *ledserv;

	for(i = 0; i < MAX_MCU_LED; i++) {
		if(ledc[i].id == 0xFF) {
			ledc[i].id = i;
			if(dev->name == NULL)
				sprintf(ledc[i].name, "led.%d");
			else
				sprintf(ledc[i].name, "%s.%d",dev->name, i);
			ledserv = kzalloc (sizeof(struct led_classdev),
					   GFP_KERNEL);
			if (ledserv != NULL) {
				ledserv->name = ledc[i].name;
				ledserv->brightness_set = set_brightness;
				ledserv->default_trigger = "ledservtrig";
				ledc[i].cdev = ledserv;
			}
			else {
				printk(KERN_INFO "%s OUT KO\n",__FUNCTION__);
				return NULL;
			}

			ledc[i].dev = dev;
			return &ledc[i];
		}
	}

	printk(KERN_INFO "%s OUT KO\n",__FUNCTION__);
	return NULL;
}

static void release_dev_ledc(unsigned int id)
{
	if(id < MAX_MCU_LED){
		kfree (ledc[id].cdev);
		ledc[id].id = 0xFF;
	}
}
#if 0
static ssize_t read_flash(struct class_device *cdev, char *buf)
{
	ssize_t ret = 0;
	enum led_flash state;
	struct led_dev_ctx *pledc = get_ledc_from_cdev(cdev);

	if (pledc->sdev->get_flash == NULL) {
		ret = -EINVAL;
		goto exit;
	}

	state = (pledc->sdev->get_flash)();

	/* no lock needed for this */
	sprintf(buf, "%u\n", state);
	ret = strlen(buf) + 1;
exit:
	return ret;
}

static ssize_t set_flash(struct class_device *cdev, const char *buf, size_t size)
{
	ssize_t ret = -EINVAL;
	char *after;
	unsigned long state = simple_strtoul(buf, &after, 16);
	size_t count = after - buf;
	struct led_dev_ctx *pledc = get_ledc_from_cdev(cdev);

	if (pledc->sdev->set_flash == NULL) {
		ret = -EINVAL;
		goto exit;
	}

	if (*after && isspace(*after))
		count++;

	if (count == size) {
		ret = count;
		(pledc->sdev->set_flash)(state);
	}
exit:
	return ret;
}
#endif

static void set_brightness(struct led_classdev *led_cdev,
			   enum led_brightness brightness)
{
	struct led_dev_ctx *pledc;
	struct led_service *servfct;

	printk("%s: brightness = %d/n",__FUNCTION__, brightness);

	if ((pledc = get_ledc_from_cdev(led_cdev)) != NULL) {
		servfct = pledc->dev->service;
		servfct->set_brightness(brightness);
	}
}

int mcu_service_led_probe(struct mcu_service_dev *dev)
{
	int ret = 0;
	struct led_dev_ctx *pledc = get_dev_ledc(dev);

	printk(KERN_DEBUG "led service device is found\n");

	if(pledc == NULL) {
		ret = -1;
		goto exit;
	}

	ret = led_classdev_register(NULL, pledc->cdev);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't register the ledserv led class : 0x%x\n",
			__FUNCTION__, ret);
		goto exit;
	}

	goto exit;

exit:
	return ret;
}

static int mcu_service_led_remove(struct mcu_service_dev *dev)
{
	struct led_dev_ctx *pledc = get_ledc_from_dev(dev);

	led_classdev_unregister(pledc->cdev);
	release_dev_ledc(pledc->id);
	return 0;
}

struct mcu_service_driver mcu_service_led_driver = {
	.driver = {
		.owner	= THIS_MODULE,
		.name	= "service_led",
	},
	.service_type = SERVICE_LED,
	.probe = mcu_service_led_probe,
	.remove = mcu_service_led_remove,
};

static int __init led_service_init(void)
{
	int ret;

	printk(KERN_DEBUG "led service is ready\n");

	ledc_init();
	ret = mcu_service_driver_register(&mcu_service_led_driver);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't register the MCU service led driver: 0x%x\n",
			__FUNCTION__, ret);
	}

	return ret;
}

static void __exit led_service_exit(void)
{
	printk(KERN_DEBUG "led service is finishing\n");
	mcu_service_driver_unregister(&mcu_service_led_driver);
}

module_init(led_service_init);
module_exit(led_service_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("LED service device support");
MODULE_AUTHOR("Frederic Mazuel <fmazuel@wyplay.com>");
