/*
 * (C) Copyright 2009 WyPlay SAS.
 * Frederic Mazuel <fmazuel@wyplay.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/ctype.h>
#include <linux/fs.h>
#include <linux/leds.h>
#include <linux/wymcu/mcu_service.h>

struct led_dev_ctx {
	unsigned int id;
	char name[20];
	struct led_classdev *cdev;
	struct mcu_service_dev *dev;
};

struct flash_ctx {
	unsigned long delay_on;		/* milliseconds on */
	unsigned long delay_off;	/* milliseconds off */
	unsigned char state[10];
	int (*set_flash)(struct led_classdev *ldev, char *state);
	int (*read_flash)(struct led_classdev *ldev, char *state);
};

static struct led_dev_ctx ledc[MAX_MCU_LED];

static struct serv_tag ledflashstate[] = {
	{"off",LED_FLASH_OFF},
	{"on",LED_FLASH_ON},
	{"rcu",LED_FLASH_RCU},
	{NULL, 0}
};

static void set_brightness(struct led_classdev *led_cdev,
			   enum led_brightness brightness);
static int read_flash(struct led_classdev *ldev, char *state);
static int set_flash(struct led_classdev *ldev, char *state);


static void ledc_init(void)
{
	int i = MAX_MCU_LED;
	while(i--) {
		ledc[i].id = 0xFF;
	}
}

static struct led_dev_ctx *get_ledc_from_cdev(struct led_classdev *cdev)
{
	int i;
	for(i = 0; i < MAX_MCU_LED; i++) {
		if(ledc[i].cdev == cdev) {
			return &ledc[i];
		}
	}
	return NULL;
}

static struct led_dev_ctx *get_ledc_from_dev(struct mcu_service_dev *dev)
{
	int i;
	for(i = 0; i < MAX_MCU_LED; i++) {
		if(ledc[i].dev == dev) {
			return &ledc[i];
		}
	}
	return NULL;
}

static struct led_dev_ctx *get_dev_ledc(struct mcu_service_dev *dev)
{
	int i;
	struct led_classdev *ledserv;
	struct flash_ctx *flash_data;

	for(i = 0; i < MAX_MCU_LED; i++) {
		if(ledc[i].id == 0xFF) {
			ledc[i].id = i;
			if(dev->name == NULL)
				sprintf(ledc[i].name, "led:%d", i);
			else
				sprintf(ledc[i].name, "%s",dev->name);
			ledserv = kzalloc (sizeof(struct led_classdev),
					   GFP_KERNEL);
			if (ledserv != NULL) {
				struct led_service *led_fct = dev->service;
				ledserv->name = ledc[i].name;
				ledserv->brightness_set = set_brightness;
				if(led_fct->set_flash != NULL) {
					ledserv->default_trigger = "ledservtrig";
					flash_data = kzalloc(sizeof(struct flash_ctx), GFP_KERNEL);
					if (!flash_data) {
						kfree(ledserv);
						return NULL;
					}
					sprintf(flash_data->state, "off");
					flash_data->read_flash = read_flash;
					flash_data->set_flash = set_flash;
					ledserv->trigger_data = flash_data;
				}
				ledc[i].cdev = ledserv;
			}
			else {
				printk(KERN_INFO "%s OUT KO\n",__FUNCTION__);
				return NULL;
			}

			ledc[i].dev = dev;
			return &ledc[i];
		}
	}

	printk(KERN_INFO "%s OUT KO\n",__FUNCTION__);
	return NULL;
}

static void release_dev_ledc(unsigned int id)
{
	if(id < MAX_MCU_LED){
		kfree (ledc[id].cdev);
		ledc[id].id = 0xFF;
	}
}

static int serv_str2int(char *state, struct serv_tag *array)
{
	struct serv_tag *pt = array;
	int val = -1;

	if (!array)
		goto exit;

	while (pt->name != NULL) {
		if(!strcmp(state, pt->name)) {
			val = pt->val;
			break;
		}
		pt++;
	}

exit:
	return val;
}

static char *serv_int2str(int val, struct serv_tag *array, int array_size)
{
	if (val < array_size)
		return ledflashstate[val].name;
	else
		return NULL;
}

static int read_flash(struct led_classdev *ldev, char *buf)
{
	int ret = 0;
	struct led_dev_ctx *pledc;
	struct led_service *servfct;
	int status;
	char * pstr;

	printk(KERN_DEBUG "%s led %s\n",__FUNCTION__,ldev->name);

	if ((pledc = get_ledc_from_cdev(ldev)) != NULL) {
		servfct = pledc->dev->service;
		if (servfct->get_flash != NULL) {
			status = servfct->get_flash(pledc->dev);
			pstr = serv_int2str(status,
					    ledflashstate,
					    sizeof (ledflashstate) /
					    sizeof(struct serv_tag));
			sprintf(buf, "%s\n", pstr);
		}
		else
			ret = -ENOTSUPP;
	}

	return ret;
}

static int set_flash(struct led_classdev *ldev, char *state)
{
	int ret = 0;
	struct led_dev_ctx *pledc;
	struct led_service *servfct;
	int status;

	printk(KERN_DEBUG "%s led %s\n",__FUNCTION__,ldev->name);

	if ((pledc = get_ledc_from_cdev(ldev)) != NULL) {
		servfct = pledc->dev->service;
		if (servfct->set_flash != NULL) {
			status = serv_str2int(state, ledflashstate);
			servfct->set_flash(pledc->dev, status);
		}
		else
			ret = -ENOTSUPP;
	}

	return ret;
}

static void set_brightness(struct led_classdev *led_cdev,
			   enum led_brightness brightness)
{
	struct led_dev_ctx *pledc;
	struct led_service *servfct;

	printk(KERN_DEBUG "%s: brightness = %d\n",__FUNCTION__, brightness);

	if ((pledc = get_ledc_from_cdev(led_cdev)) != NULL) {
		servfct = pledc->dev->service;
		if (servfct->set_brightness)
			servfct->set_brightness(pledc->dev, brightness);
	}
}

int mcu_service_led_probe(struct mcu_service_dev *dev)
{
	int ret = 0;
	struct led_dev_ctx *pledc = get_dev_ledc(dev);

	printk(KERN_DEBUG "led service device is found\n");

	if(pledc == NULL) {
		ret = -1;
		goto exit;
	}

	ret = led_classdev_register(NULL, pledc->cdev);
	if (ret < 0) {
		printk(KERN_ERR
		       "%s: Can't register the ledserv led class : 0x%x\n",
		       __FUNCTION__, ret);
		goto exit;
	}
exit:
	return ret;
}

static int mcu_service_led_remove(struct mcu_service_dev *dev)
{
	struct led_dev_ctx *pledc = get_ledc_from_dev(dev);

	led_classdev_unregister(pledc->cdev);
	release_dev_ledc(pledc->id);
	return 0;
}

struct mcu_service_driver mcu_service_led_driver = {
	.driver = {
		.owner	= THIS_MODULE,
		.name	= "service_led",
	},
	.service_type = SERVICE_LED,
	.probe = mcu_service_led_probe,
	.remove = mcu_service_led_remove,
};

static int __init led_service_init(void)
{
	int ret;

	printk(KERN_DEBUG "led service is ready\n");

	ledc_init();
	ret = mcu_service_driver_register(&mcu_service_led_driver);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't register the MCU service led driver: 0x%x\n",
			__FUNCTION__, ret);
	}

	return ret;
}

static void __exit led_service_exit(void)
{
	printk(KERN_DEBUG "led service is finishing\n");
	mcu_service_driver_unregister(&mcu_service_led_driver);
}

module_init(led_service_init);
module_exit(led_service_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("LED service device support");
MODULE_AUTHOR("Frederic Mazuel <fmazuel@wyplay.com>");
