/*
 * (C) Copyright 2009 WyPlay SAS.
 * Frederic Mazuel <fmazuel@wyplay.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/ctype.h>
#include <linux/sysfs.h>
#include <linux/wymcu/mcu_service.h>

static struct sound_service *services;
static struct mcu_service_dev *serv_dev;
static dev_t sndserv_dev_number;
static struct class *sndserv_class;
static struct class_device *sndserv_class_dev;

ssize_t store_snd_mode(struct class_device *cdev, const char *buf, size_t count)
{
	char status[10];
	int res = 0;
	int val;

	res = sscanf(buf, "%s", status);
	if (res != 1)
		return -EINVAL;

	val = simple_strtoul(status, NULL, 10);

	/* Call the MCU driver API to set the LED status */
	services->set_mode(serv_dev, val);

	return strnlen(buf, count);
}

ssize_t show_snd_mode(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int mode;

	/* Read LED status */
	mode = services->get_mode(serv_dev);

	/* Update the status buffer */
	pos += sprintf(pos, "%d\n",mode);

	return pos - buf;
}

static CLASS_DEVICE_ATTR(mode, S_IRUGO | S_IWUSR,
			 show_snd_mode, store_snd_mode);

static int mcu_service_snd_probe(struct mcu_service_dev *dev)
{
	int ret = 0;

	printk(KERN_DEBUG "Sound service drv instance is starting\n");

	services = (struct sound_service *)(dev->service);
	serv_dev = dev;

	if(services == NULL) {
		printk(KERN_WARNING "no sound services\n");
		goto exit;
	}

	sndserv_class_dev = class_device_create (sndserv_class, NULL,
						 sndserv_dev_number, NULL,
						 "sndserv0");
	if (IS_ERR(sndserv_class_dev)) {
		printk( KERN_ERR
			"%s: ERROR: Creating sndserv0 class device: %ld)\n",
			__FUNCTION__, PTR_ERR(sndserv_class_dev));
		ret = PTR_ERR(sndserv_class_dev);
		goto exit;
	}

	ret = class_device_create_file(sndserv_class_dev,
				       &class_device_attr_mode);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creating sound service mode file (%d)\n",
			__FUNCTION__, ret);
		goto err1;
	}

	goto exit;

err1:
	class_device_del(sndserv_class_dev);
exit:
	return ret;
}

static int mcu_service_snd_remove(struct mcu_service_dev *dev)
{
	printk(KERN_DEBUG "Sound service drv instance is stopping\n");
	class_device_remove_file(sndserv_class_dev, &class_device_attr_mode);
	class_device_destroy(sndserv_class, sndserv_dev_number);
	class_destroy(sndserv_class);
	return 0;
}

struct mcu_service_driver mcu_service_snd_driver = {
	.driver = {
		.owner	= THIS_MODULE,
		.name	= "service_snd",
	},
	.service_type = SERVICE_SND,
	.probe = mcu_service_snd_probe,
	.remove = mcu_service_snd_remove,
};

static int __init snd_service_init(void)
{
	int ret;

	printk(KERN_DEBUG "Sound service is registering\n");

	ret = alloc_chrdev_region (&sndserv_dev_number, 0, 1, "sndservice");
	if (ret < 0) {
		printk (KERN_ERR
			"%s: sound service can't register device err: 0x%x\n",
			__FUNCTION__, ret);
		goto exit;
	}

	sndserv_class = class_create (THIS_MODULE, "sndserv");
	if (IS_ERR(sndserv_class)) {
		printk( KERN_ERR
			"%s: ERROR: Creating sndserv class: %ld)\n",
			__FUNCTION__, PTR_ERR(sndserv_class));
		ret = PTR_ERR(sndserv_class);
		goto err1;
	}

	ret = mcu_service_driver_register(&mcu_service_snd_driver);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't register the MCU service sound driver: 0x%x\n",
			__FUNCTION__, ret);
		goto err2;
	}

	goto exit;

err2:
	class_destroy(sndserv_class);
err1:
	unregister_chrdev_region(sndserv_dev_number, 1);
exit:
	return ret;
}

static void __exit snd_service_exit(void)
{
	printk(KERN_DEBUG "Sound service is unregistering\n");
	mcu_service_driver_unregister(&mcu_service_snd_driver);
}

module_init(snd_service_init);
module_exit(snd_service_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Sound service device support");
MODULE_AUTHOR("Frederic Mazuel <fmazuel@wyplay.com>");
