/*
 * (C) Copyright 2009 WyPlay SAS.
 * Frederic Mazuel <fmazuel@wyplay.com>
 * Jerry Zhang <zhangb@wyplayasia.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/ctype.h>
#include <linux/sysfs.h>
#include <linux/wymcu/mcu_service.h>

#define MAX_TIMER_VAL  (0x00ffffff - 1)

static struct timer_service *services;
static dev_t tmrserv_dev_number;
static struct class *tmrserv_class;
static struct class_device *tmrserv_class_dev;
/* buffer the alarm value to be written to the timer */
static int timer_alarm_value = 0;

/* timer state */
ssize_t store_tmr_state(struct class_device *cdev, const char *buf, size_t count)
{
	char status[10];
	int res, value;
	pkg_general_t pkg;
	char tx_buffer[5];
	smcu_mode_t	mode = MODE_NORMAL;

	res = sscanf(buf, "%s", status);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = timer_alarm_value;
	if(!strcmp("start", status)) {
		pkg.type = TMR_SET_ARM;
		tx_buffer[3] = value & 0xFF;
		tx_buffer[2] = (value >> 8) & 0xFF;
		tx_buffer[1] = (value >> 16) & 0xFF;
	}else if(!strcmp("stop", status)) {
		pkg.type = TMR_SET_STP;
	}else {
		res = -EINVAL;
		goto exit;
	}
	/* Call the MCU driver API to set the LED status */
	pkg.tx_buf = tx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = strnlen(buf, count);
exit:
	return res;
}

ssize_t show_tmr_state(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int res;
	pkg_general_t pkg;
	char tx_buffer[5], rx_buffer[5];
	smcu_mode_t	mode = MODE_NORMAL;

	/* Read boot status */
	pkg.type = TMR_GET_ARM;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		pos += sprintf(pos, "%s:%d\n", "Error", res);
		return pos - buf;
	}
	/* Update the status buffer */
	if((!rx_buffer[0]) && (!rx_buffer[1]) && (!rx_buffer[2])) {
		pos += sprintf(pos, "%s\n", "stopped");
	}else {
		pos += sprintf(pos, "%s\n", "running");
	}

	return pos - buf;
}

static CLASS_DEVICE_ATTR(state, S_IRUGO | S_IWUSR,
			 show_tmr_state, store_tmr_state);

/* timer value */
ssize_t store_tmr_value(struct class_device *cdev, const char *buf, size_t count)
{
	char status[10];
	int res;
	int value;

	res = sscanf(buf, "%s", status);
	if(res != 1) {
		res = -EINVAL;
		goto exit;
	}
	value = simple_strtoul(status, NULL, 10);
	if(value < 0 || value > MAX_TIMER_VAL) {
		res = -EINVAL;
		goto exit;
	}
	/* set the new value to timer and run it */
	timer_alarm_value = value;
	res = strnlen(buf, count);
exit:
	return res;
}

ssize_t show_tmr_value(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int res, value;
	pkg_general_t pkg;
	char tx_buffer[5], rx_buffer[5];
	smcu_mode_t	mode = MODE_NORMAL;

	/* Read boot status */
	pkg.type = TMR_GET_ARM;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		pos += sprintf(pos, "%s:%d\n", "Error", res);
		return pos - buf;
	}
	/* Update the status buffer */
	value = 0;
	value |= rx_buffer[0] << 16;
	value |= rx_buffer[1] << 8;
	value |= rx_buffer[2];

	pos += sprintf(pos, "%d\n", value);

	return pos - buf;
}
static CLASS_DEVICE_ATTR(value, S_IRUGO | S_IWUSR,
			 show_tmr_value, store_tmr_value);

/* timer maxvalue */
ssize_t show_tmr_maxval(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int value;

	/* Read boot status */
	value = MAX_TIMER_VAL;
	pos += sprintf(pos, "%d\n", value);

	return pos - buf;
}
static CLASS_DEVICE_ATTR(maxval, S_IRUGO,
			 show_tmr_maxval, store_tmr_value);

/* rtc interrupt state
 * 1:
 * 0:
 */
ssize_t show_tmr_rtc_int(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int value;

	value = services->get_rtc_state();
	pos += sprintf(pos, "%d\n", value);

	return pos - buf;
}

static CLASS_DEVICE_ATTR(rtc_int_state, S_IRUGO,
			 show_tmr_rtc_int, NULL);

static int mcu_service_tmr_probe(struct mcu_service_dev *dev)
{
	int ret = 0;

	printk(KERN_DEBUG "Timer service drv instance is starting\n");

	services = (struct timer_service *)(dev->service);
	if(services == NULL) {
		printk(KERN_WARNING "no timer services\n");
		goto exit;
	}

	tmrserv_class_dev = class_device_create (tmrserv_class, NULL,
						 tmrserv_dev_number, NULL,
						 "timerserv0");
	if (IS_ERR(tmrserv_class_dev)) {
		printk( KERN_ERR
			"%s: ERROR: Creating tmrserv0 class device: %ld)\n",
			__FUNCTION__, PTR_ERR(tmrserv_class_dev));
		ret = PTR_ERR(tmrserv_class_dev);
		goto exit;
	}

	ret = class_device_create_file(tmrserv_class_dev,
				       &class_device_attr_state);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creating timer service state file (%d)\n",
			__FUNCTION__, ret);
		goto err1;
	}
	ret = class_device_create_file(tmrserv_class_dev,
				       &class_device_attr_value);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creating timer service value file (%d)\n",
			__FUNCTION__, ret);
		goto err2;
	}
	ret = class_device_create_file(tmrserv_class_dev,
				       &class_device_attr_maxval);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creating timer service maxval file (%d)\n",
			__FUNCTION__, ret);
		goto err3;
	}
	ret = class_device_create_file(tmrserv_class_dev,
				       &class_device_attr_rtc_int_state);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creating timer service RTC state file (%d)\n",
			__FUNCTION__, ret);
		goto err4;
	}
	goto exit;

err4:
	class_device_remove_file(tmrserv_class_dev, &class_device_attr_maxval);
err3:
	class_device_remove_file(tmrserv_class_dev, &class_device_attr_value);
err2:
	class_device_remove_file(tmrserv_class_dev, &class_device_attr_state);
err1:
	class_device_del(tmrserv_class_dev);
exit:
	return ret;
}

static int mcu_service_tmr_remove(struct mcu_service_dev *dev)
{
	printk(KERN_DEBUG "Timer service drv instance is stopping\n");
	class_device_remove_file(tmrserv_class_dev,
				 &class_device_attr_rtc_int_state);
	class_device_remove_file(tmrserv_class_dev, &class_device_attr_maxval);
	class_device_remove_file(tmrserv_class_dev, &class_device_attr_value);
	class_device_remove_file(tmrserv_class_dev, &class_device_attr_state);
	class_device_destroy(tmrserv_class, tmrserv_dev_number);
	return 0;
}

struct mcu_service_driver mcu_service_timer_driver = {
	.driver = {
		.owner	= THIS_MODULE,
		.name	= "service_timer",
	},
	.service_type = SERVICE_TIMER,
	.probe = mcu_service_tmr_probe,
	.remove = mcu_service_tmr_remove,
};

static int __init timer_service_init(void)
{
	int ret;

	printk(KERN_DEBUG "Timer service is registering\n");

	ret = alloc_chrdev_region (&tmrserv_dev_number, 0, 1, "timerservice");
	if (ret < 0) {
		printk (KERN_ERR
			"%s: timer service can't register device err: 0x%x\n",
			__FUNCTION__, ret);
		goto exit;
	}

	tmrserv_class = class_create (THIS_MODULE, "timerserv");
	if (IS_ERR(tmrserv_class)) {
		printk( KERN_ERR
			"%s: ERROR: Creating tmrserv class: %ld)\n",
			__FUNCTION__, PTR_ERR(tmrserv_class));
		ret = PTR_ERR(tmrserv_class);
		goto err1;
	}

	ret = mcu_service_driver_register(&mcu_service_timer_driver);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't register the MCU service timer driver: 0x%x\n",
			__FUNCTION__, ret);
		goto err2;
	}

	goto exit;

err2:
	class_destroy(tmrserv_class);
err1:
	unregister_chrdev_region(tmrserv_dev_number, 1);
exit:
	return ret;
}

static void __exit timer_service_exit(void)
{
	printk(KERN_DEBUG "Timer service is unregistering\n");
	mcu_service_driver_unregister(&mcu_service_timer_driver);
	class_destroy(tmrserv_class);
	unregister_chrdev_region(tmrserv_dev_number, 1);
}

module_init(timer_service_init);
module_exit(timer_service_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Timer service device support");
MODULE_AUTHOR("Frederic Mazuel <fmazuel@wyplay.com>, Jerry Zhang <zhangb@wyplayasia.com>");
