/*
 * (C) Copyright 2009 WyPlay SAS.
 * Frederic Mazuel <fmazuel@wyplay.com>
 * Jerry Zhang <zhangb@wyplayasia.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/ctype.h>
#include <linux/sysfs.h>
#include <linux/wymcu/mcu_service.h>
#include <linux/wymcu/mcu_ldisc.h>

static struct vfd_service *services;
static dev_t vfdserv_dev_number;
static struct class *vfdserv_class;
static struct class_device *vfdserv_class_dev;

/* VFD message */
ssize_t store_vfd_message(struct class_device *cdev, const char *buf, size_t count)
{
	int res = 0;
	char tx_buffer[50], rx_buffer[5];
	pkg_general_t pkg;
	smcu_mode_t mode = MODE_NORMAL;

	if(count > 48) {
		res = -EINVAL;
		goto exit;
	}
	strncpy(tx_buffer +2, buf, count);
	tx_buffer[0] = 0x4F;
	tx_buffer[1] = count-1;
	pkg.type = INITIALIZED_PKG;
	pkg.tx_buf = tx_buffer;
	pkg.rx_buf = rx_buffer;
	pkg.tx_len = count + 1;
	pkg.rx_len = 0;
	res = services->smcu_transfer(&pkg, mode);
	if(res < 0) {
		goto exit;
	}
	res = count;
exit:
	return res;
}

ssize_t show_vfd_message(struct class_device *cdev, char *buf)
{
	char *pos = buf;
	int value;
	/* what should we show? There is no show message support in the mcu
	 * shall we buffer a former message here ?
	 */
	value = 0x10;

	pos += sprintf(pos, "%d\n", value);

	return pos - buf;
}
static CLASS_DEVICE_ATTR(message, S_IRUGO | S_IWUSR,
			 show_vfd_message, store_vfd_message);

static int mcu_service_vfd_probe(struct mcu_service_dev *dev)
{
	int ret = 0;

	printk(KERN_DEBUG "VFD service drv instance is starting\n");

	services = (struct vfd_service *)(dev->service);
	if(services == NULL) {
		printk(KERN_WARNING "no vfd services\n");
		goto exit;
	}

	vfdserv_class_dev = class_device_create (vfdserv_class, NULL,
						 vfdserv_dev_number, NULL,
						 "vfdserv0");
	if (IS_ERR(vfdserv_class_dev)) {
		printk( KERN_ERR
			"%s: ERROR: Creating displayserv0 class device: %ld)\n",
			__FUNCTION__, PTR_ERR(vfdserv_class_dev));
		ret = PTR_ERR(vfdserv_class_dev);
		goto exit;
	}

	ret = class_device_create_file(vfdserv_class_dev,
				       &class_device_attr_message);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't Creating display service vfd file (%d)\n",
			__FUNCTION__, ret);
		goto err1;
	}
	goto exit;

err1:
	class_device_del(vfdserv_class_dev);
exit:
	return ret;
}

static int mcu_service_vfd_remove(struct mcu_service_dev *dev)
{
	printk(KERN_DEBUG "VFD service drv instance is stopping\n");
	class_device_remove_file(vfdserv_class_dev, &class_device_attr_message);
	class_device_destroy(vfdserv_class, vfdserv_dev_number);
	return 0;
}

struct mcu_service_driver mcu_service_vfd_driver = {
	.driver = {
		.owner	= THIS_MODULE,
		.name	= "service_vfd",
	},
	.service_type = SERVICE_VFD,
	.probe = mcu_service_vfd_probe,
	.remove = mcu_service_vfd_remove,
};

static int __init vfd_service_init(void)
{
	int ret;
	printk(KERN_INFO "VFD service is registering\n");

	ret = alloc_chrdev_region (&vfdserv_dev_number, 0, 1, "vfdservice");
	if (ret < 0) {
		printk (KERN_ERR
			"%s: vfd service can't register device err: 0x%x\n",
			__FUNCTION__, ret);
		goto exit;
	}
	vfdserv_class = class_create (THIS_MODULE, "dispserv");
	if (IS_ERR(vfdserv_class)) {
		printk( KERN_ERR
			"%s: ERROR: Creating vfdserv class: %ld)\n",
			__FUNCTION__, PTR_ERR(vfdserv_class));
		ret = PTR_ERR(vfdserv_class);
		goto err1;
	}
	ret = mcu_service_driver_register(&mcu_service_vfd_driver);
	if (ret < 0) {
		printk (KERN_ERR
			"%s: Can't register the MCU service vfd driver: 0x%x\n",
			__FUNCTION__, ret);
		goto err2;
	}
	goto exit;

err2:
	class_destroy(vfdserv_class);
err1:
	unregister_chrdev_region(vfdserv_dev_number, 1);
exit:
	return ret;
}

static void __exit vfd_service_exit(void)
{
	printk(KERN_DEBUG "VFD service is unregistering\n");
	mcu_service_driver_unregister(&mcu_service_vfd_driver);
	class_destroy(vfdserv_class);
	unregister_chrdev_region(vfdserv_dev_number, 1);
}

module_init(vfd_service_init);
module_exit(vfd_service_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("VFD service device support");
MODULE_AUTHOR("Frederic Mazuel <fmazuel@wyplay.com> \
               Jerry Zhang <zhangb@wyplayasia.com>");
