/*
 * Direct UBI block device access
 *
 * Author: dmitry pervushin <dimka@embeddedalley.com>
 *
 * Copyright 2008 Freescale Semiconductor, Inc. All Rights Reserved.
 * Copyright 2008 Embedded Alley Solutions, Inc. All Rights Reserved.
 * Copyright 2010 Wyplay SAS. All Rights Reserved.
 *
 * The code contained herein is licensed under the GNU General Public
 * License. You may obtain a copy of the GNU General Public License
 * Version 2 or later at the following locations:
 *
 * http://www.opensource.org/licenses/gpl-license.html
 * http://www.gnu.org/copyleft/gpl.html
 *
 * Based on mtdblock by:
 * (C) 2000-2003 Nicolas Pitre <nico@cam.org>
 * (C) 1999-2003 David Woodhouse <dwmw2@infradead.org>
 */
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/vmalloc.h>
#include <linux/genhd.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/blktrans.h>
#include <linux/mtd/ubi.h>
#include <linux/mutex.h>
#include <linux/workqueue.h>
#include <asm/uaccess.h>
#include <mtd/ubiblock_ro-user.h>

extern struct mutex mtd_table_mutex;

static int vol_id = -1;

#if defined(MODULE)
module_param(vol_id, int, 0);
MODULE_PARM_DESC(vol_id, "volume id to be available as block device");
#else
static int __init vol_id_setup(char *str)
{
	vol_id = simple_strtoul(str, NULL, 0);
	return 1;
}

__setup("ubiblk.vol_id=", vol_id_setup);
#endif

static LIST_HEAD(ubiblk_devices);

/**
 * ubiblk_dev - the structure representing translation layer
 *
 * @m: interface to mtd_blktrans
 * @ubi_num: UBI device number
 * @ubi_vol: UBI volume ID
 * @usecount: reference count
 *
 * @cache_mutex: protects access to cache_data
 * @cache_data: content of the cached LEB
 * @cache_offset: offset of cached data on UBI volume, in bytes
 * @cache_size: cache size in bytes, usually equal to LEB size
 */
struct ubiblk_dev {
	struct mtd_blktrans_dev m;

	int 			ubi_num;
	int			ubi_vol;
	int			usecount;
	struct ubi_volume_desc *ubi;

	struct mutex 		cache_mutex;
	unsigned char 	       *cache_data;
	unsigned long 		cache_offset;
	unsigned int 		cache_size;

	enum {
		STATE_EMPTY,
		STATE_CLEAN,
		STATE_DIRTY
	} cache_state;

	struct list_head 	list;

	struct work_struct	unbind;
};

static int ubiblock_open(struct mtd_blktrans_dev *mbd);
static int ubiblock_release(struct mtd_blktrans_dev *mbd);
static int ubiblock_readsect(struct mtd_blktrans_dev *mbd,
			      unsigned long block, char *buf);
static int ubiblock_getgeo(struct mtd_blktrans_dev *mbd,
		struct hd_geometry *geo);
static void *ubiblk_add(int ubi_num, int ubi_vol_id);
static void *ubiblk_add_locked(int ubi_num, int ubi_vol_id);
static int ubiblk_del(struct ubiblk_dev *u);
static int ubiblk_del_locked(struct ubiblk_dev *u);

/*
 * These two routines are just to satify mtd_blkdev's requirements
 */
static void ubiblock_add_mtd(struct mtd_blktrans_ops *tr, struct mtd_info *mtd)
{
	return;
}

static void ubiblock_remove_dev(struct mtd_blktrans_dev *mbd)
{
	return;
}

static struct mtd_blktrans_ops ubiblock_tr = {

	.name		= "ubiblk",
	.major		= 33,			/* assign dynamically  */
	.part_bits	= 4,			/* allow up to 16 parts */
	.blksize 	= 512,

	.open		= ubiblock_open,
	.release	= ubiblock_release,
	.readsect	= ubiblock_readsect,
	.getgeo		= ubiblock_getgeo,

	.add_mtd	= ubiblock_add_mtd,
	.remove_dev	= ubiblock_remove_dev,
	.owner		= THIS_MODULE,
};

static int ubiblock_getgeo(struct mtd_blktrans_dev *bdev,
		struct hd_geometry *geo)
{
	return -ENOTTY;
}

/**
 * ubiblk_do_cached_read - cached read the data from ubi volume
 *
 * @u: ubiblk_dev
 * @pos: offset on the block device
 * @len: buffer length
 * @buf: preallocated buffer
 *
 * Cached LEB will be used if possible; otherwise data will be using
 * ubi_leb_read
 */
static int ubiblk_do_cached_read(struct ubiblk_dev *u, unsigned long pos,
			   int len, char *buf)
{
	unsigned int sect_size = u->cache_size;
	int err = 0;
	unsigned long sect_start;
	unsigned offset, size;
	int leb;

	pr_debug("%s: read at 0x%lx, size 0x%x\n",
			__func__, pos, len);

	while (len > 0) {

		leb = pos/sect_size;
		sect_start = leb*sect_size;
		offset = pos - sect_start;
		size = sect_size - offset;

		if (size > len)
			size = len;

		/*
		 * Check if the requested data is already cached
		 * Read the requested amount of data from our internal
		 * cache if it contains what we want, otherwise we read
		 * the data directly from flash.
		 */
		if (u->cache_state != STATE_EMPTY &&
		    u->cache_offset == sect_start) {
			pr_debug("%s: cached, returning back from cache\n",
					__func__);
			memcpy(buf, u->cache_data + offset, size);
			err = 0;
		} else {
			pr_debug("%s: pos = %ld, reading leb = %d\n",
					__func__, pos, leb);
			err = ubi_leb_read(u->ubi, leb, buf, offset, size, 0);
			if (err)
				goto out;
		}

		buf += size;
		pos += size;
		len -= size;
	}

out:
	return err;
}

/**
 * ubiblk_readsect - read the sector
 *
 * Allocate the cache, if necessary, and perform actual read using
 * ubiblk_do_cached_read
 */
static int ubiblock_readsect(struct mtd_blktrans_dev *mbd,
			      unsigned long block, char *buf)
{
	struct ubiblk_dev *u = container_of(mbd, struct ubiblk_dev, m);

	if (unlikely(!u->cache_data)) {
		u->cache_data = vmalloc(u->cache_size);
		if (!u->cache_data)
			return -EAGAIN;
	}
	return ubiblk_do_cached_read(u, block<<9, 512, buf);
}

static int ubiblock_open(struct mtd_blktrans_dev *mbd)
{
	struct ubiblk_dev *u = container_of(mbd, struct ubiblk_dev, m);

	if (u->usecount == 0) {
		u->ubi = ubi_open_volume(u->ubi_num, u->ubi_vol,
				/*UBI_READWRITE*/ UBI_READONLY);
		if (IS_ERR(u->ubi))
			return PTR_ERR(u->ubi);
	}
	u->usecount++;
	return 0;
}

static int ubiblock_release(struct mtd_blktrans_dev *mbd)
{
	struct ubiblk_dev *u = container_of(mbd, struct ubiblk_dev, m);

	if (--u->usecount == 0) {
		mutex_lock(&u->cache_mutex);
		vfree(u->cache_data);
		u->cache_data = NULL;
		mutex_unlock(&u->cache_mutex);
		ubi_close_volume(u->ubi);
		u->ubi = NULL;
	}
	return 0;
}
#if 0
/*
 * sysfs routines. The ubiblk creates two entries under /sys/block/ubiblkX:
 *  - volume, R/O, which is read like "ubi0:volume_name"
 *  - unbind, W/O; when user writes something here, the block device is
 *  removed
 *
 *  unbind schedules a work item to perform real unbind, because sysfs entry
 *  handler cannot delete itself :)
 */
ssize_t volume_show(struct device *dev,
		struct device_attribute *attr, char *buf)
{
	struct gendisk *gd = dev_to_disk(dev);
	struct mtd_blktrans_dev *m = gd->private_data;
	struct ubiblk_dev *u = container_of(m, struct ubiblk_dev, m);

	return sprintf(buf, "%d:%d\n", u->ubi_num, u->ubi_vol);
}

static void ubiblk_unbind(struct work_struct *ws)
{
	struct ubiblk_dev *u = container_of(ws, struct ubiblk_dev, unbind);

	ubiblk_del(u);
}

ssize_t unbind_store(struct device *dev, struct device_attribute *attr,
			 const char *buf, size_t count)
{
	struct gendisk *gd = dev_to_disk(dev);
	struct mtd_blktrans_dev *m = gd->private_data;
	struct ubiblk_dev *u = container_of(m, struct ubiblk_dev, m);

	INIT_WORK(&u->unbind, ubiblk_unbind);
	schedule_work(&u->unbind);
	return count;
}

DEVICE_ATTR(unbind, 0644, NULL, unbind_store);
DEVICE_ATTR(volume, 0644, volume_show, NULL);

static int ubiblk_sysfs(struct gendisk *hd, int add)
{
	int r = 0;

	if (add) {
		r = device_create_file(disk_to_dev(hd), &dev_attr_unbind);
		if (r < 0)
			goto out;
		r = device_create_file(disk_to_dev(hd), &dev_attr_volume);
		if (r < 0)
			goto out1;
		return 0;
	}

	device_remove_file(disk_to_dev(hd), &dev_attr_unbind);
out1:
	device_remove_file(disk_to_dev(hd), &dev_attr_volume);
out:
	return r;
}
#endif

/**
 * add the FTL by registering it with mtd_blkdevs
 */
static void *ubiblk_add(int ubi_num, int ubi_vol_id)
{
	void *p;

	mutex_lock(&mtd_table_mutex);
	p = ubiblk_add_locked(ubi_num, ubi_vol_id);
	mutex_unlock(&mtd_table_mutex);
	return p;
}

static void *ubiblk_add_locked(int ubi_num, int ubi_vol_id)
{
	struct ubiblk_dev *u = kzalloc(sizeof(*u), GFP_KERNEL);
	struct ubi_volume_info uvi;
	struct ubi_volume_desc *ubi;
	struct mtd_info *mtd;
	int ret;

	if (!u) {
		u = ERR_PTR(-ENOMEM);
		goto out;
	}

	printk(KERN_INFO "UBIBLK: adding volume %d:%d\n", ubi_num, ubi_vol_id);

	ubi = ubi_open_volume(ubi_num, ubi_vol_id, UBI_READONLY);
	if (IS_ERR(ubi)) {
		kfree(u);
		pr_err("cannot open the volume\n");
		u = (void *)ubi;
		goto out;
	}

	ubi_get_volume_info(ubi, &uvi);
	ubi_close_volume(ubi);

	pr_debug("adding volume of size %d (used_size %lld), LEB size %d\n",
		uvi.size, uvi.used_bytes, uvi.usable_leb_size);

	mtd = kzalloc(sizeof(*mtd), GFP_KERNEL);
	if (!mtd) {
		kfree(u);
		u = ERR_PTR(-ENOMEM);
		goto out;
	}

	u->m.mtd = mtd;
	u->m.devnum = -1;
	u->m.size = uvi.used_bytes >> 9;
	u->m.tr = &ubiblock_tr;
	u->m.readonly = 1;

	u->ubi_num = ubi_num;
	u->ubi_vol = ubi_vol_id;

	mutex_init(&u->cache_mutex);
	u->cache_state = STATE_EMPTY;
	u->cache_size = uvi.usable_leb_size;
	u->cache_data = NULL;
	u->usecount = 0;
	INIT_LIST_HEAD(&u->list);

	list_add_tail(&u->list, &ubiblk_devices);
	ret = add_mtd_blktrans_dev(&u->m);
//	ubiblk_sysfs(u->m.blkcore_priv, true);
out:
	return u;
}

static int ubiblk_del(struct ubiblk_dev *u)
{
	int r;
	mutex_lock(&mtd_table_mutex);
	r = ubiblk_del_locked(u);
	mutex_unlock(&mtd_table_mutex);
	return r;
}

static int ubiblk_del_locked(struct ubiblk_dev *u)
{
	if (u->usecount != 0)
		return -EBUSY;
//	ubiblk_sysfs(u->m.blkcore_priv, false);
	del_mtd_blktrans_dev(&u->m);

	if (u->m.mtd)
		kfree(u->m.mtd);

	list_del(&u->list);
	BUG_ON(u->cache_data != NULL); /* who did not free the cache ?! */
	kfree(u);
	return 0;
}

static struct ubiblk_dev *ubiblk_find(int num, int vol)
{
	struct ubiblk_dev *pos;

	list_for_each_entry(pos, &ubiblk_devices, list)
		if (pos->ubi_num == num && pos->ubi_vol == vol)
			return pos;
	return NULL;
}


static int ubiblock_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct ubiblk_req rq;
	struct ubiblk_dev *dev;
	int r;

	switch (cmd) {
		case UBIBLOCK_ADDVOL:
			if ((r = copy_from_user(&rq, arg, sizeof(rq))) < 0)
				return r;
			return (ubiblk_add(rq.ubi_dev_id, rq.ubi_vol_id) != NULL) ? 0 : -EIO;

		case UBIBLOCK_RMVOL:
			if ((r = copy_from_user(&rq, arg, sizeof(rq))) < 0)
				return r;

			mutex_lock(&mtd_table_mutex);
			dev = ubiblk_find(rq.ubi_dev_id, rq.ubi_vol_id);
			mutex_unlock(&mtd_table_mutex);

			if (dev)
				return ubiblk_del(dev);
			else
				return -ENODEV;
			break;
	}

	return -EINVAL;
}

static struct file_operations ubiblock_fops = {
	owner: THIS_MODULE,
	ioctl: ubiblock_ioctl,
};

static int __init ubiblock_init(void)
{
	int r;

	r = register_mtd_blktrans(&ubiblock_tr);
	if (r)
		return -EIO;

	r = register_chrdev(33, "ubiblk_ro", &ubiblock_fops);
	if (r < 0) {
		deregister_mtd_blktrans(&ubiblock_tr);
		return r;
	}

	if (vol_id >= 0)
		if(!ubiblk_add(0, vol_id))
			printk("UBIBLK: unable to add volume 0:%d as a block device\n", vol_id);

	return 0;
}

static void __exit ubiblock_exit(void)
{
	deregister_mtd_blktrans(&ubiblock_tr);
}

module_init(ubiblock_init);
module_exit(ubiblock_exit);
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Caching block device emulation access to UBI devices");
MODULE_AUTHOR("dmitry pervushin <dimka@embeddedalley.com>");
