/*
 *  Driver for the TKDMA
 *
 * Copyright (C) 2006 STMicroelectronics Limited
 * Copyright (C) 2009 Wyplay SAS
 *
 * May be copied or modified under the terms of the GNU General Public
 * License.  See linux/COPYING for more information.
 */

#include <linux/module.h>
#include <asm/cacheflush.h>
#include <linux/mm.h>
#include <linux/crypto.h>
#include <crypto/algapi.h>

#define AES_MIN_KEY_SIZE        16
#define AES_MAX_KEY_SIZE        16
#define AES_BLOCK_SIZE          16


extern int tkdma_aes_sck_physical(void *, void *, u8 *, u8 *, unsigned int, int);
extern void tkdma_lock(void);
extern void tkdma_unlock(void);

static int aes_set_key(struct crypto_tfm *tfm, const uint8_t *key, unsigned int key_len)
{
	uint8_t *ctx;

	if (key_len != 16)
		return -EINVAL;

	ctx = crypto_tfm_ctx(tfm);
	memcpy(ctx, key, 16);
	return 0;
}


static int aes_decrypt_cbc(struct blkcipher_desc *desc, struct scatterlist *dst,
			   struct scatterlist *src, unsigned int nbytes)
{
	int err;
	struct blkcipher_walk walk;

	memset(&walk, 0, sizeof(walk));
	blkcipher_walk_init(&walk, dst, src, nbytes);

	err = blkcipher_walk_phys(desc, &walk);

	while ((nbytes = walk.nbytes)) {
		int res;
		void *vsrc, *vdst, *iv;
		void *psrc, *pdst;

		tkdma_lock();

		vsrc = page_address(walk.src.phys.page) + walk.src.phys.offset;
		psrc = page_to_phys(walk.src.phys.page) + walk.src.phys.offset;
		vdst = page_address(walk.dst.phys.page) + walk.dst.phys.offset;
		pdst = page_to_phys(walk.dst.phys.page) + walk.dst.phys.offset;

		dma_cache_wback(vsrc, nbytes);
		res = tkdma_aes_sck_physical(psrc, pdst, crypto_blkcipher_ctx(desc->tfm),
				             walk.iv, nbytes, 0x43);
		dma_cache_inv(vdst, nbytes);

		if (res < 0) {
			tkdma_unlock();
			blkcipher_walk_done(desc, &walk, nbytes);
			return res;
		} else
			res = nbytes;

		iv = &walk.src.virt.addr;
		memcpy(walk.iv, iv + (nbytes - 16), 16);

		tkdma_unlock();

		nbytes -= res;
		err = blkcipher_walk_done(desc, &walk, nbytes);
	}

	return err;
}



static int dummy_encrypt(struct blkcipher_desc *desc, struct scatterlist *dst,
		         struct scatterlist *src, unsigned int nbytes)
{
	return -ENOSYS;
}


static struct crypto_alg aes_alg = {
	.cra_name = "cbc(aes)",
	.cra_driver_name = "tkdma-aes-cbc",
	.cra_priority = 300,
	.cra_flags = CRYPTO_ALG_TYPE_BLKCIPHER,
	.cra_blocksize = AES_BLOCK_SIZE,
	.cra_ctxsize = 16,
	.cra_alignmask = 15,
	.cra_type = &crypto_blkcipher_type,
	.cra_module = THIS_MODULE,
	.cra_list = LIST_HEAD_INIT(aes_alg.cra_list),
	.cra_u = {
		  .blkcipher = {
				.min_keysize = AES_MIN_KEY_SIZE,
				.max_keysize = AES_MAX_KEY_SIZE,
				.ivsize = AES_BLOCK_SIZE,
				.setkey = aes_set_key,
				.encrypt = dummy_encrypt,
				.decrypt = aes_decrypt_cbc,
		  }
	}
};


static __init int _init(void)
{
	return crypto_register_alg(&aes_alg);
}

static void __exit _cleanup(void)
{
	crypto_unregister_alg(&aes_alg);
}

MODULE_DESCRIPTION("Tkdma Driver");
MODULE_AUTHOR("STMicroelectronics Limited / Wyplay SAS");
MODULE_VERSION("20090828");
MODULE_LICENSE("GPL");

module_init(_init);
module_exit(_cleanup);

