/*

 Driver APIs for MxL5007 Tuner
 
 Copyright, Maxlinear, Inc.
 All Rights Reserved
 
 File Name:      MxL5007_API.c
 
 Version:    4.1.3
*/
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/string.h>
#include <linux/slab.h>
#include "dvb_frontend.h"


#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,17)
#include <linux/stm/pio.h>
#else	/* STLinux 2.2 or older */
#include <linux/stpio.h>        /*!< for stpio_set_pin */
#endif

#include "mxl5007t_lite.h"

struct i2c_adapter *mxl5007_adapter = NULL;
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//																		   //
//					I2C Functions (implement by customer)				   //
//																		   //
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//

/******************************************************************************
**
**  Name: mxl_i2c_write
**
**  Description:    I2C write operations
**
**  Parameters:    	
**					addr	- MxL5007 Device address
**					array		- Write data array pointer
**					count		- total number of array
**
**  Returns:        0 if success
**
**  Revision History:
**
**   SCR      Date      Author  Description
**  -------------------------------------------------------------------------
**   N/A   12-16-2007   khuang initial release.
**
******************************************************************************/
int mxl_i2c_write(u8 addr, u8* array, int count)
{
	struct i2c_msg msg = {.addr = addr, .flags = 0, .buf = array, .len = count};
	int ret=0;
	
	if (mxl5007_adapter == NULL)
		return -1;
	
	ret = i2c_transfer(mxl5007_adapter, &msg, 1);
	
	if (ret != 1)
		printk("%s :error in mxl_i2c_write, (addr = 0x%02x)\n",
					__FUNCTION__, addr);

	return (ret != 1) ? -1 : 0;
}

/******************************************************************************
**
**  Name: mxl_i2c_read
**
**  Description:    I2C read operations
**
**  Parameters:    	
**					addr	- MxL5007 Device address
**					reg		- register address for read
**					*dat		- data return
**
**  Returns:        0 if success
**
**  Revision History:
**
**   SCR      Date      Author  Description
**  -------------------------------------------------------------------------
**   N/A   12-16-2007   khuang initial release.
**
******************************************************************************/
int mxl_i2c_read(u8 addr, u8 reg, u8* val)
{
	int ret;
	u8 b0 [] = {0xFB, reg };
	u8 b1 [] = { 0 };
	struct i2c_msg msg [] = { { .addr = addr, .flags = 0, .buf = b0, .len = sizeof(b0)},
			   { .addr = addr, .flags = I2C_M_RD, .buf = b1, .len = 1 } };

	if (mxl5007_adapter == NULL)
		return -EIO;
		
	ret = i2c_transfer (mxl5007_adapter, msg, 1);
	ret |= i2c_transfer (mxl5007_adapter, msg+1, 1);

	if (ret != 1) {
		printk("%s: error (addr = 0x%02x, reg == 0x%02x,  ret == %i)\n",
			__FUNCTION__, addr, reg, ret);
	  return -EIO;
	}
	
	*val = b1[0];
	return 0;
}

/******************************************************************************
**
**  Name: mxl_delay
**
**  Description:    Delay function in milli-second
**
**  Parameters:    	
**					mSec		- milli-second to delay
**
**  Returns:        0
**
**  Revision History:
**
**   SCR      Date      Author  Description
**  -------------------------------------------------------------------------
**   N/A   12-16-2007   khuang initial release.
**
******************************************************************************/
void mxl_delay(u16 millisecond)
{
    msleep(millisecond);
}

typedef struct
{
	u8 number;	//register number
	u8 value;	//register value
} irv_type;

//local functions called by Init and RFTune
void set_irv_bit(irv_type *irv, u8 number, u8 mask, u8 value)
{
	while (irv->number || irv->value)
	{
		if (irv->number == number)
		{
			irv->value &= ~mask;
			irv->value |= value;
		}
		
		irv++;
	}	
}


int mxl5007_init(u8* array,				// a array pointer that store the addr and data pairs for I2C write
					int* array_size,			// a integer pointer that store the number of element in above array
					u8 mode,				
					int if_level,		 
					int xtal_freq,			
					int if_freq,				
					u8 invert_if,											
					u8 clk_out_en,    
					u8 clk_out_amp													
					)
{
	
	int reg_index=0;
	int array_index=0;

	irv_type irv_init[]=
	{
		//{ Addr, dat}	
		{ 0x02, 0x06}, 
		{ 0x03, 0x48},
		{ 0x05, 0x04},  
		{ 0x06, 0x10}, 
		{ 0x2E, 0x15},  //Override
		{ 0x30, 0x10},  //Override
		{ 0x45, 0x58},  //Override
		{ 0x48, 0x19},  //Override
		{ 0x52, 0x03},  //Override
		{ 0x53, 0x44},  //Override
		{ 0x6A, 0x4B},  //Override
		{ 0x76, 0x00},  //Override
		{ 0x78, 0x18},  //Override
		{ 0x7A, 0x17},  //Override
		{ 0x85, 0x06},  //Override		
		{ 0x01, 0x01}, //TOP_MASTER_ENABLE=1
		{ 0, 0}
	};


	irv_type irv_init_cable[]=
	{
		//{ Addr, dat}	
		{ 0x02, 0x06}, 
		{ 0x03, 0x48},
		{ 0x05, 0x04},  
		{ 0x06, 0x10},  
		{ 0x09, 0x3F},  
		{ 0x0A, 0x3F},  
		{ 0x0B, 0x3F},  
		{ 0x2E, 0x15},  //Override
		{ 0x30, 0x10},  //Override
		{ 0x45, 0x58},  //Override
		{ 0x48, 0x19},  //Override
		{ 0x52, 0x03},  //Override
		{ 0x53, 0x44},  //Override
		{ 0x6A, 0x4B},  //Override
		{ 0x76, 0x00},  //Override
		{ 0x78, 0x18},  //Override
		{ 0x7A, 0x17},  //Override
		{ 0x85, 0x06},  //Override	
		{ 0x01, 0x01}, //TOP_MASTER_ENABLE=1
		{ 0, 0}
	};
	//edit Init setting here

	irv_type *my_irv = NULL;

	switch(mode)
	{
	case MXL_MODE_ISDBT: //ISDB-T mode	
		my_irv = irv_init;
		set_irv_bit(my_irv, 0x06, 0x1F, 0x10);  
		break;
	case MXL_MODE_DVBT: //DVBT mode			
		my_irv = irv_init;
		set_irv_bit(my_irv, 0x06, 0x1F, 0x11);  
		break;
	case MXL_MODE_ATSC: //ATSC mode			
		my_irv = irv_init;
		set_irv_bit(my_irv, 0x06, 0x1F, 0x12);  
		break;	
	case MXL_MODE_CABLE:						
		my_irv = irv_init_cable;
		set_irv_bit(my_irv, 0x09, 0xFF, 0xC1);	
		set_irv_bit(my_irv, 0x0A, 0xFF, (u8)(8-if_level));	
		set_irv_bit(my_irv, 0x0B, 0xFF, 0x17);							
		break;
	default:
		printk(KERN_ERR "No valid mode selected\n");
		goto failed;
	}

	switch(if_freq)
	{
	case MXL_IF_4_MHZ:
		set_irv_bit(my_irv, 0x02, 0x0F, 0x00); 		
		break;
	case MXL_IF_4_5_MHZ: //4.5MHz
		set_irv_bit(my_irv, 0x02, 0x0F, 0x02); 		
		break;
	case MXL_IF_4_57_MHZ: //4.57MHz
		set_irv_bit(my_irv, 0x02, 0x0F, 0x03); 
		break;
	case MXL_IF_5_MHZ:
		set_irv_bit(my_irv, 0x02, 0x0F, 0x04); 
		break;
	case MXL_IF_5_38_MHZ: //5.38MHz
		set_irv_bit(my_irv, 0x02, 0x0F, 0x05); 
		break;
	case MXL_IF_6_MHZ: 
		set_irv_bit(my_irv, 0x02, 0x0F, 0x06); 
		break;
	case MXL_IF_6_28_MHZ: //6.28MHz
		set_irv_bit(my_irv, 0x02, 0x0F, 0x07); 
		break;
	case MXL_IF_9_1915_MHZ: //9.1915MHz
		set_irv_bit(my_irv, 0x02, 0x0F, 0x08); 
		break;
	case MXL_IF_35_25_MHZ:
		set_irv_bit(my_irv, 0x02, 0x0F, 0x09); 
		break;
	case MXL_IF_36_15_MHZ:
		set_irv_bit(my_irv, 0x02, 0x0F, 0x0a); 
		break;
	case MXL_IF_44_MHZ:
		set_irv_bit(my_irv, 0x02, 0x0F, 0x0B); 
		break;
	default:
		printk (KERN_ERR "no valid IF frequency selected\n");
		goto failed;
	}

	if(invert_if) 
		set_irv_bit(my_irv, 0x02, 0x10, 0x10);   //Invert IF
	else
		set_irv_bit(my_irv, 0x02, 0x10, 0x00);   //Normal IF


	switch(xtal_freq)
	{
	case MXL_XTAL_16_MHZ:
		set_irv_bit(my_irv, 0x03, 0xF0, 0x00);  //select xtal freq & Ref Freq
		set_irv_bit(my_irv, 0x05, 0x0F, 0x00);
		break;
	case MXL_XTAL_20_MHZ:
		set_irv_bit(my_irv, 0x03, 0xF0, 0x10);
		set_irv_bit(my_irv, 0x05, 0x0F, 0x01);
		break;
	case MXL_XTAL_20_25_MHZ:
		set_irv_bit(my_irv, 0x03, 0xF0, 0x20);
		set_irv_bit(my_irv, 0x05, 0x0F, 0x02);
		break;
	case MXL_XTAL_20_48_MHZ:
		set_irv_bit(my_irv, 0x03, 0xF0, 0x30);
		set_irv_bit(my_irv, 0x05, 0x0F, 0x03);
		break;
	case MXL_XTAL_24_MHZ:
		set_irv_bit(my_irv, 0x03, 0xF0, 0x40);
		set_irv_bit(my_irv, 0x05, 0x0F, 0x04);
		break;
	case MXL_XTAL_25_MHZ:
		set_irv_bit(my_irv, 0x03, 0xF0, 0x50);
		set_irv_bit(my_irv, 0x05, 0x0F, 0x05);
		break;
	case MXL_XTAL_25_14_MHZ:
		set_irv_bit(my_irv, 0x03, 0xF0, 0x60);
		set_irv_bit(my_irv, 0x05, 0x0F, 0x06);
		break;
	case MXL_XTAL_27_MHZ:
		set_irv_bit(my_irv, 0x03, 0xF0, 0x70);
		set_irv_bit(my_irv, 0x05, 0x0F, 0x07);
		break;
	case MXL_XTAL_28_8_MHZ: 
		set_irv_bit(my_irv, 0x03, 0xF0, 0x80);
		set_irv_bit(my_irv, 0x05, 0x0F, 0x08);
		break;
	case MXL_XTAL_32_MHZ:
		set_irv_bit(my_irv, 0x03, 0xF0, 0x90);
		set_irv_bit(my_irv, 0x05, 0x0F, 0x09);
		break;
	case MXL_XTAL_40_MHZ:
		set_irv_bit(my_irv, 0x03, 0xF0, 0xA0);
		set_irv_bit(my_irv, 0x05, 0x0F, 0x0A);
		break;
	case MXL_XTAL_44_MHZ:
		set_irv_bit(my_irv, 0x03, 0xF0, 0xB0);
		set_irv_bit(my_irv, 0x05, 0x0F, 0x0B);
		break;
	case MXL_XTAL_48_MHZ:
		set_irv_bit(my_irv, 0x03, 0xF0, 0xC0);
		set_irv_bit(my_irv, 0x05, 0x0F, 0x0C);
		break;
	case MXL_XTAL_49_3811_MHZ:
		set_irv_bit(my_irv, 0x03, 0xF0, 0xD0);
		set_irv_bit(my_irv, 0x05, 0x0F, 0x0D);
		break;	
	default:
		printk(KERN_ERR "no valid xtal frequency selected\n");
		goto failed;
	}

	if(!clk_out_en) //default is enable 
		set_irv_bit(my_irv, 0x03, 0x08, 0x00);   

	//clk_out_amp
	set_irv_bit(my_irv, 0x03, 0x07, clk_out_amp);   

	//Generate one Array that Contain dat, Address
	while (my_irv[reg_index].number || my_irv[reg_index].value)
	{
		array[array_index++] = my_irv[reg_index].number;
		array[array_index++] = my_irv[reg_index].value;
		reg_index++;
	}
	    
	*array_size = array_index;
	return 0;

failed:
	printk(KERN_ERR "%s failed\n", __FUNCTION__);
	return -1;
}

int mxl5007_rf_tune(u8* array, int* array_size, int rf_frequency, u8 bandwidth)
{
	irv_type irv_rf_tune[]=
	{
	//{ Addr, dat}
		{ 0x0F, 0x00},  //abort tune
		{ 0x0C, 0x15},
		{ 0x0D, 0x40},
		{ 0x0E, 0x0E},	
		{ 0x1F, 0x87},  //Override
		{ 0x20, 0x1F},  //Override
		{ 0x21, 0x87},  //Override
		{ 0x22, 0x1F},  //Override		
		{ 0x80, 0x01},  //Freq Dependent Setting
		{ 0x0F, 0x01},	//start tune
		{ 0, 0}
	};

	int dig_rf_freq=0;
	int temp;
	int reg_index=0;
	int array_index=0;
	int i;
	int frac_divider = 1000000;

	switch(bandwidth)
	{
	case MXL_BW_6MHz: //6MHz
		set_irv_bit(irv_rf_tune, 0x0C, 0x3F, 0x15);  //set DIG_MODEINDEX, DIG_MODEINDEX_A, and DIG_MODEINDEX_CSF
		break;
	case MXL_BW_7MHz: //7MHz
		set_irv_bit(irv_rf_tune, 0x0C, 0x3F, 0x2A);
		break;
	case MXL_BW_8MHz: //8MHz
		set_irv_bit(irv_rf_tune, 0x0C, 0x3F, 0x3F);
		break;
	}


	//Convert RF frequency into 16 bits => 10 bit integer (MHz) + 6 bit fraction
	dig_rf_freq = rf_frequency / MHz;
	temp = rf_frequency % MHz;
	for(i=0; i<6; i++)
	{
		dig_rf_freq <<= 1;
		frac_divider /=2;
		if(temp > frac_divider)
		{
			temp -= frac_divider;
			dig_rf_freq++;
		}
	}

	//add to have shift center point by 7.8124 kHz
	if(temp > 7812)
		dig_rf_freq ++;
    	
	set_irv_bit(irv_rf_tune, 0x0D, 0xFF, (u8)dig_rf_freq);
	set_irv_bit(irv_rf_tune, 0x0E, 0xFF, (u8)(dig_rf_freq>>8));

	if (rf_frequency >=333*MHz)
		set_irv_bit(irv_rf_tune, 0x80, 0x40, 0x40);

	//Generate one Array that Contain dat, Address 
	while (irv_rf_tune[reg_index].number || irv_rf_tune[reg_index].value)
	{
		array[array_index++] = irv_rf_tune[reg_index].number;
		array[array_index++] = irv_rf_tune[reg_index].value;
		reg_index++;
	}
    
	*array_size = array_index;
	
	return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//																		   //
//							Tuner Functions								   //
//																		   //
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
mxl_err_msg mxl_set_register(mxl5007_tuner_configs* configs, u8 addr, u8 val)
{
	int status=0;
	u8 array[2];
	array[0] = addr;
	array[1] = val;
	
	status = mxl_i2c_write((u8)configs->i2c_addr, array, 2);
	if(status)
	{
		printk("mxl_i2c_write in %s faild, register 0x%02x\n", __FUNCTION__, (int)addr);
		return MXL_ERR_SET_REG;
	}

	return MXL_OK;

}

mxl_err_msg mxl_get_register(mxl5007_tuner_configs* configs, u8 addr, u8 *val)
{
	if(mxl_i2c_read((u8)configs->i2c_addr, addr, val))
		return MXL_ERR_GET_REG;
	return MXL_OK;

}

mxl_err_msg mxl_soft_reset(mxl5007_tuner_configs* configs)
{
	u8 buf[] = {0xFF, 0xFF};
	
	if(mxl_i2c_write((u8)configs->i2c_addr, buf, 2))
	{
		printk("mxl_i2c_write in %s failed\n", __FUNCTION__ );
		return MXL_ERR_OTHERS;
	}

	return MXL_OK;
}

mxl_err_msg mxl_loop_through_on(mxl5007_tuner_configs* configs, mxl5007_lt isOn)
{	
	u8 array[2];	// a array pointer that store the addr and data pairs for I2C write	
	
	array[0]=0x04;
	if(isOn)
	 array[1]= 0x01;
	else
	 array[1]= 0x0;

	if(mxl_i2c_write((u8)configs->i2c_addr, array, 2))
	{
		printk("mxl_i2c_write in %s failed\n", __FUNCTION__);
		return MXL_ERR_OTHERS;
	}

	return MXL_OK;
}

mxl_err_msg mxl_standby(mxl5007_tuner_configs* configs)
{
	u8 array[4];	// a array pointer that store the addr and data pairs for I2C write	
	
	array[0] = 0x01;
	array[1] = 0x0;
	array[2] = 0x0F;
	array[3] = 0x0;

	if(mxl_i2c_write((u8)configs->i2c_addr, array, 4))
		return MXL_ERR_OTHERS;

	return MXL_OK;
}

mxl_err_msg mxl_wakeup(mxl5007_tuner_configs* configs)
{
	u8 array[2];	// a array pointer that store the addr and data pairs for I2C write	
	
	array[0] = 0x01;
	array[1] = 0x01;

	if(mxl_i2c_write((u8)configs->i2c_addr, array, 2))
		return MXL_ERR_OTHERS;

	if(mxl_tuner_rftune(configs, configs->rf_frequency, configs->bandwidth))
		return MXL_ERR_RFTUNE;

	return MXL_OK;
}

mxl_err_msg mxl_tuner_init(mxl5007_tuner_configs* configs)
{	
	u8 array[MAX_ARRAY_SIZE];	// a array pointer that store the addr and data pairs for I2C write
	int array_size;							// a integer pointer that store the number of element in above array

	//Soft reset tuner
	if(mxl_soft_reset(configs))
		return MXL_ERR_INIT;

	//perform initialization calculation
	mxl5007_init(array, &array_size, (u8)configs->mode, configs->if_level, (int)configs->xtal, 
				(int)configs->if_frequency, (u8)configs->if_spectrum, (u8)configs->clock_out, (u8)configs->amp);

	//perform I2C write here
	if(mxl_i2c_write((u8)configs->i2c_addr, array, array_size))
	{
		printk("mxl_i2c_write in %s failed\n", __FUNCTION__);
		return MXL_ERR_INIT;
	}

	return MXL_OK;
}


mxl_err_msg mxl_tuner_rftune(mxl5007_tuner_configs* configs, int rf_frequency, mxl5007_bandwidth bandwidth)
{
	u8 array[MAX_ARRAY_SIZE];	// a array pointer that store the addr and data pairs for I2C write
	int array_size;							// a integer pointer that store the number of element in above array

	//Store information into struc
	configs->rf_frequency = rf_frequency;
	configs->bandwidth = bandwidth;

	//perform Channel Change calculation
	mxl5007_rf_tune(array,&array_size,rf_frequency,(u8)bandwidth);

	//perform I2C write here
	if(mxl_i2c_write((u8)configs->i2c_addr, array, array_size))
	{
		printk("mxl_i2c_write in %s failed\n", __FUNCTION__);
		return MXL_ERR_RFTUNE;
	}

	//wait for 3ms
	mxl_delay(3); 

	return MXL_OK;
}

mxl5007_version mxl_check_version(mxl5007_tuner_configs* configs)
{	
	u8 dat;
	
	if(mxl_i2c_read((u8)configs->i2c_addr, 0xD9, &dat))
		return MXL_GET_ID_FAIL;
		
	switch(dat)
	{
	case 0x14: return MXL_5007T_V4; break;
	default: return MXL_UNKNOWN_ID;
	}	
}

mxl_err_msg mxl_rf_syn_lockstatus(mxl5007_tuner_configs* configs, u8* isLock)
{	
	u8 dat;
	*isLock = 0; 
	
	if(mxl_i2c_read((u8)configs->i2c_addr, 0xD8, &dat))
		return MXL_ERR_OTHERS;
	dat &= 0x0C;
	if (dat == 0x0C)
		*isLock = 1;  //RF Synthesizer is Lock	
	return MXL_OK;
}

mxl_err_msg mxl_ref_syn_lockstatus(mxl5007_tuner_configs* configs, u8* isLock)
{
	u8 dat;
	*isLock = 0; 
	
	if(mxl_i2c_read((u8)configs->i2c_addr, 0xD8, &dat))
		return MXL_ERR_OTHERS;
	dat &= 0x03;
	if (dat == 0x03)
		*isLock = 1;   //REF Synthesizer is Lock
	return MXL_OK;
}

EXPORT_SYMBOL(mxl5007_adapter);
