/*
 * ST-PTI DVB driver
 *
 * Copyright (c) STMicroelectronics 2005
 *
 *   Author:Peter Bennett <peter.bennett@st.com>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation; either version 2 of
 *	the License, or (at your option) any later version.
 */
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/wait.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <linux/delay.h>
#include <linux/time.h>
#include <linux/errno.h>
#include <asm/semaphore.h>
#include <linux/platform_device.h>

#include <asm/io.h>

#include <linux/bpa2.h>

#include "dvb_frontend.h"
#include "dmxdev.h"
#include "dvb_demux.h"
#include "dvb_net.h"

#include <linux/dvb/frontend.h>
#include <linux/dvb/dmx.h>

#if defined(CONFIG_CPU_SUBTYPE_STB5301)
#define ST5301
#endif

#include "pti.h"
#include "st-common.h"
#include "st-merger.h"
#include "st-tuner.h"

#undef SWTS_TEST

#define STPTI_MAXCHANNEL 32
#define STPTI_MAXADAPTER 3

struct stpti {
  struct stfe *stfe[STPTI_MAXADAPTER];
  int running_feed_count; 
  int mapping[16];

  struct semaphore sem;
  
  spinlock_t timer_lock;
  struct timer_list timer;	/* timer interrupts for outputs */

  char *back_buffer;
  unsigned int pti_io;
  
  short int *pidtable;
  short int *num_pids;
  short int *pidsearch;
  volatile unsigned short int *psize;
  volatile unsigned short int *pread;
  volatile unsigned short int *thrown_away;
  volatile unsigned short int *packet_size;
  volatile unsigned short int *header_size;
  volatile unsigned short int *pwrite;
  volatile unsigned short int *discard;

  unsigned int loop_count;
  unsigned int loop_count2;
  unsigned int packet_count;
};

#define HEADER_SIZE (6)
#define PACKET_SIZE (188+HEADER_SIZE)

#define PTI_BUFFER_SIZE (16*(PACKET_SIZE)*100) /*(188*1024)*/

/* PTI write ignores byte enables */
static void PtiWrite(volatile unsigned short int *addr,unsigned short int value)
{
  unsigned int *addr2 = (unsigned int*)(((unsigned int)addr) & ~3);
  unsigned int old_value  = *addr2;

  if (((unsigned)addr & 2)) old_value = (old_value & 0x0000ffff) | value << 16;
  else old_value = (old_value & 0xffff0000) | value;

  *addr2 = old_value;
}

/***********************************************************************************/

#define TCASM_LM_SUPPORT
#include "tc_code.h"

static void loadtc(struct stpti *pti)
{
  int n;

  for (n=0;n<=TCASM_LableMap[0].Value;n+=4)
    writel(transport_controller_code[n/4], pti->pti_io + PTI_IRAM_BASE + n );

  for (n=0;n<(sizeof(transport_controller_data) * 2);n+=4)
    writel(transport_controller_data[n/4], pti->pti_io + PTI_DRAM_BASE + n);
}

static void *getsymbol(struct stpti *pti, const char *symbol)
{
  char temp[128];
  void *result = NULL;
  int n;
  	
  temp[120] = 0;
  sprintf(temp,"::_%s",symbol);

  for (n=0;n<TRANSPORT_CONTROLLER_LABLE_MAP_SIZE;n++)
  	if (!strcmp(temp,TCASM_LableMap[n].Name))
  	result = (void*)(pti->pti_io + TCASM_LableMap[n].Value);

  return result;

}

/***************************************************************************************/

static void stpti_setup_dma(struct stpti *pti)
{
  /* Setup DMA0 to transfer the data to the buffer */
  writel( virt_to_phys(pti->back_buffer), pti->pti_io + PTI_DMA_0_BASE );
  writel( virt_to_phys(pti->back_buffer), pti->pti_io + PTI_DMA_0_WRITE);
  writel( virt_to_phys(pti->back_buffer), pti->pti_io + PTI_DMA_0_READ );
  writel( virt_to_phys(pti->back_buffer + PTI_BUFFER_SIZE - 1), pti->pti_io + PTI_DMA_0_TOP );

  writel( 0x8, pti->pti_io + PTI_DMA_0_SETUP ); /* 8 word burst */
}

static void stpti_start_dma(struct stpti *pti)
{
  /* Enable the DMA */
  writel( readl(pti->pti_io + PTI_DMA_ENABLE) | 0x1, pti->pti_io + PTI_DMA_ENABLE);

  /* Tell PTI we have lots of room in buffer */
  PtiWrite(pti->pread, 0);
  PtiWrite(pti->pwrite,0);
  PtiWrite(pti->discard, 0);

  /* Reset all the counts */
  pti->loop_count = 0;
  pti->loop_count2 = 0;
  pti->packet_count=0;
}

static void stpti_stop_dma(struct stpti *pti)
{
int loop_count = 0;

  /* Stop DMAing data */
  PtiWrite(pti->pread, 0);

  while (*pti->pwrite != *pti->psize +2)
  {
    PtiWrite(pti->pwrite, *pti->psize + 2);
    PtiWrite(pti->pwrite, *pti->psize + 2);
    dprintk("%s: Pwrite = %u Psize = %u\n",__FUNCTION__,*pti->pwrite,*pti->psize);
  };

  udelay(100);

  /* Ensure DMA is enabled */
  writel( readl(pti->pti_io + PTI_DMA_ENABLE) | 0x1, pti->pti_io + PTI_DMA_ENABLE );

  /* Do DMA Done to ensure all data has been transfered */
  writel( PTI_DMA_DONE , pti->pti_io + PTI_DMA_0_STATUS );

  while( (readl(pti->pti_io + PTI_DMA_0_STATUS) & PTI_DMA_DONE) && 
	 (loop_count<100)) 
    { udelay(1000); loop_count++; };
  
  /* Disable the DMA */
  writel(readl(pti->pti_io + PTI_DMA_ENABLE) & ~0x1, pti->pti_io + PTI_DMA_ENABLE);
}

static void stpti_reset_dma(struct stpti *pti)
{
  /* Stop the DMA */
  stpti_stop_dma(pti);

  /* Reset dma pointers */
  stpti_setup_dma(pti);
}


/*

0x10   ----------------- <--- PTI_DMA_x_BASE
       |               |   |
       |               |  \|/
       |               | <--- PTI_DMA_x_READ  (Last address sent to ADSC)
       |               |
       |               |
       |               | <--- PTI_DMA_x_WRITE (Last address written by PTI)
       |               |   ^
       |               |   |  (free space)
       |               |   |
0x1000 ----------------- <--- PTI_DMA_x_TOP

0x10   ----------------- <--- PTI_DMA_x_BASE
       |               |   
       |               |  
       |               | <--- PTI_DMA_x_WRITE (Last address written by PTI)
       |               |   |
       |               |   | (free space)
       |               | <--- PTI_DMA_x_READ  (Last address sent to ADSC)
       |               |   
       |               |   
       |               |   
0x1000 ----------------- <--- PTI_DMA_x_TOP

*/
/* We should make the PTI interrupt us every sensible packet period */
/* For now we shall just interrupt when ever the timer period is up */
/* And we assume buffer is a multiple of th packet length */

extern int read_tsm(int);

static void stpti_timer_interrupt(unsigned long astpti)
{
  struct stpti *pti = (struct stpti*)astpti;
  //  struct stpti *stpti = (struct stpti*)stfei2->driver_data;
  unsigned int pti_rp,pti_wp,pti_base,pti_top,size_first,num_packets,new_rp,pti_status,dma_status;
  static unsigned int pti_status_old=0;
  static unsigned int dma_status_old=0;
  int n;

  pti->timer.expires = 2 + jiffies;

  stm_tsm_interrupt();

  /* Load the read and write pointers, so we know where we are in the buffers */
  pti_rp   = readl(pti->pti_io + PTI_DMA_0_READ);
  pti_wp   = readl(pti->pti_io + PTI_DMA_0_WRITE);
  pti_base = readl(pti->pti_io + PTI_DMA_0_BASE);
  pti_top  = pti_base + PTI_BUFFER_SIZE;
  
  /* Read status registers */
  pti_status = readl(pti->pti_io + PTI_IIF_FIFO_COUNT);
  dma_status = readl(pti->pti_io + PTI_DMA_0_STATUS);

  /* Error if we overflow */
  if (pti_status & PTI_IIF_FIFO_FULL) printk(KERN_WARNING "%s: IIF Overflow\n",__FUNCTION__);
  if (dma_status != dma_status_old) dprintk("%s: DMA Status %x %x\n",__FUNCTION__,dma_status,pti->loop_count);

  /* If the PTI had to drop packets because we couldn't process in time error */
  if (*pti->discard) 
  { 
    printk(KERN_WARNING "%s: PTI had to discard %u packets %u %u\n",__FUNCTION__,*pti->discard,*pti->pread,*pti->pwrite);
    printk(KERN_WARNING "%s: Reseting DMA\n",__FUNCTION__);

    /* If this happens the PTI stops outputing so we need to reset it */
    stpti_reset_dma(pti);
    stpti_start_dma(pti);

    add_timer(&pti->timer);
    return;
  }
    
  pti_status_old = pti_status;
  dma_status_old = dma_status;
  
  /* If we get to the bottom of the buffer wrap the pointer back to the top */
  if ((pti_rp & ~0xf) == (pti_top & ~0xf)) pti_rp = pti_base;

  /* Calculate the amount of bytes used in the buffer */
  if (pti_rp <= pti_wp) size_first = pti_wp - pti_rp;
  else size_first = pti_top - pti_rp;
  
  /* Calculate the number of packets in the buffer */
  num_packets = size_first / PACKET_SIZE;

  /* If we have some packets */
  if (num_packets)
  {
    /* And the PTI has acknowledged the updated the packets */
    if (!*pti->pread)
    {
      /* Increment the loop counter */
      pti->loop_count++;

      /* This Macro helps us get the current bytes out of the buffer */
#define PR_A(x) (((unsigned int)pti->back_buffer[(pti_rp - pti_base) + (x)]) & 0xff)

      /* The read pointer should always be a multiple of the packet size */
      if ((pti_rp - pti_base) % PACKET_SIZE) 
	printk(KERN_WARNING "%s: 0x%x not multiple of %d\n",__FUNCTION__,(pti_rp - pti_base),PACKET_SIZE);

      /* The first bytes of the packet after the header should always be a 0x47 if not we got problems */
	if (PR_A(HEADER_SIZE) != 0x47) 
	  printk(KERN_INFO "%s: @ 0x%p %x %x %u %02x %02x %02x %02x %02x %02x %02x %02x\n",__FUNCTION__,
	                   &pti->back_buffer[(pti_rp - pti_base)],
			   pti_status,dma_status,pti->loop_count,
			   PR_A(0),PR_A(1),PR_A(2),PR_A(3),PR_A(4),PR_A(5),PR_A(6),PR_A(7));
#if 1		
      /* The start pointer should always be in the buffer, this is a bit of a useless test as it should never happen */
	if (&pti->back_buffer[pti_rp - pti_base] < &pti->back_buffer[0] || 
	    &pti->back_buffer[pti_rp - pti_base] > &pti->back_buffer[PTI_BUFFER_SIZE])
	  printk(KERN_WARNING "%s: Buffer pointers have overrun!\n",__FUNCTION__);

	/* The end pointer should always be in the buffer, this is a bit of a useless test as it should never happen */
	if (&pti->back_buffer[(pti_rp - pti_base) + num_packets * PACKET_SIZE] < &pti->back_buffer[0] ||
	    &pti->back_buffer[(pti_rp - pti_base) + num_packets * PACKET_SIZE] > &pti->back_buffer[PTI_BUFFER_SIZE])
	  printk(KERN_WARNING "%s: Buffer pointers have gone over the run %p<=%p<%p r:%x t:%x (%d)!\n",__FUNCTION__,
		 &pti->back_buffer[0],
		 &pti->back_buffer[(pti_rp - pti_base) + num_packets * PACKET_SIZE],
		 &pti->back_buffer[PTI_BUFFER_SIZE],pti_rp,pti_base,num_packets);
#endif

	/* We should only read this data so lets invalidate the cache so we can see what is behind it */
	dma_cache_wback((void*)&pti->back_buffer[pti_rp - pti_base],num_packets * PACKET_SIZE);
      
	/* Now go through each packet, check it's tag and place it in the correct demux buffer */
	for (n=0;n<num_packets;n++)
	  dvb_dmx_swfilter_packets(&pti->stfe[ pti->mapping[(PR_A(n * PACKET_SIZE) >> 2) & 0xf] ]->dvb_demux, 
				   &pti->back_buffer[(pti_rp - pti_base) + (n * PACKET_SIZE) + HEADER_SIZE] , 1);
      
	/* Update the read pointer based on the number of packets we have processed */
	new_rp = pti_rp + num_packets * PACKET_SIZE;

	/* Increment the packet_count, by the number of packets we have processed */
	pti->packet_count+=num_packets;
	if (new_rp > pti_top) dprintk("You b@*tard you killed kenny\n");
	
	/* If we have gone round the buffer */
	if (new_rp >= pti_top) 
	{ 
	  /* Update the read pointer so it now points back to the top */
	  new_rp = pti_base + (new_rp - pti_top); pti->loop_count2++; 

	  /* Print out some useful debug information when debugging is on */
	  dprintk("%s: round the buffer %u times %u=pwrite %u=packet_count %u=num_packets %lu=jiffies %u %u\n",__FUNCTION__, 
		  pti->loop_count2,*pti->pwrite,pti->packet_count,num_packets,jiffies,pti_rp,pti_wp); 

	  /* Update the packet count */
	  pti->packet_count = 0; 
	}

	/* Now update the read pointer in the DMA engine */
	writel(new_rp, pti->pti_io + PTI_DMA_0_READ );

	/* Now tell the firmware how many packets we have read */
	PtiWrite(pti->pread,num_packets);
    }
  }
  
  /* Schedule us for the next but one tick event (20ms time from when we entered the handler) */
  add_timer(&pti->timer);
}

static int stpti_start_feed(struct dvb_demux_feed *dvbdmxfeed)
{
  struct stfe  *stfe  = ((struct stfe *) dvbdmxfeed->demux);
  struct stpti *stpti = (struct stpti*)stfe->driver_data;
  struct stfe_channel *channel;

  dprintk("%s: entered %p %p\n",__FUNCTION__,stfe,stpti);
  
  switch (dvbdmxfeed->type) {
  case DMX_TYPE_TS:
    break;
  case DMX_TYPE_SEC:
    break;
  default:
    return -EINVAL;
  }

  if (dvbdmxfeed->type == DMX_TYPE_TS) {
    switch (dvbdmxfeed->pes_type) {
    case DMX_TS_PES_VIDEO:
    case DMX_TS_PES_AUDIO:
    case DMX_TS_PES_TELETEXT:
    case DMX_TS_PES_PCR:
    case DMX_TS_PES_OTHER:
      channel = stfe_channel_allocate(stfe);
      break;
    default:
      return -EINVAL;
    }
  } else {
    channel = stfe_channel_allocate(stfe);
  }
  
  if (!channel)
    return -EBUSY;

  dvbdmxfeed->priv    = channel;
  channel->dvbdmxfeed = dvbdmxfeed;
  channel->pid        = dvbdmxfeed->pid;
  channel->type       = 1;
  channel->id         = *stpti->num_pids;
  channel->active     = 1;

  PtiWrite(&stpti->pidtable[*stpti->num_pids],channel->pid);
  PtiWrite(stpti->num_pids,*stpti->num_pids + 1);

/*  if (stfe->running_feed_count==0)*/
  if (stpti->running_feed_count==0)
  {
    stpti->timer.expires = 1 + jiffies;
    add_timer(&stpti->timer);

    stm_tsm_reset_channel ( stfe->tsm_channel , -1 );

    //stpti_reset_dma();
    stpti_start_dma(stpti);

    dprintk("%s: Starting DMA feed\n",__FUNCTION__);
    
  }
  
/*  stfe->running_feed_count++;*/
  stpti->running_feed_count++;

  return 0;
}

static int stpti_stop_feed(struct dvb_demux_feed *dvbdmxfeed)
{
	struct stfe_channel *channel = (struct stfe_channel *) dvbdmxfeed->priv;
	struct dvb_demux *demux = dvbdmxfeed->demux;
	struct stfe *stfe  = container_of( demux, struct stfe, dvb_demux );
	struct stpti *stpti = (struct stpti*)stfe->driver_data;
	struct dvb_demux_feed *feed;
	int n;

	/* Now reallocate the pids, and update id information */
	for (n=channel->id;n<(*stpti->num_pids-1);n++) 
	{ 
	  PtiWrite(&stpti->pidtable[n], stpti->pidtable[n+1]);
	  //for (m=0;m<STPTI_MAXCHANNEL;m++) if (stpti->channel[m].id == (n+1)) stpti->channel[m].id = n;
	}

	/* Update PIDs positions in struct stfe_channel */
 	list_for_each_entry(feed, &demux->feed_list, list_head) {
 	  struct stfe_channel *channel_right = (struct stfe_channel *)feed->priv;
 	  
 	  if (channel_right->id > channel->id)
 	    channel_right->id--;
 	}

	PtiWrite(stpti->num_pids , *stpti->num_pids - 1);
	
/*    if (--stfe->running_feed_count == 0)*/
    if (--stpti->running_feed_count == 0)
	{
	  del_timer(&stpti->timer);
	  stpti_stop_dma(stpti);

	  dprintk("%s: Stopping DMA feed\n",__FUNCTION__);
	}

	channel->active = 0;

	return 0;
}

/********************************************************/

static struct platform_device *stpti_device_data;

static int stpti_probe(struct device *dev)
{
  struct stpti *pti;
  struct plat_frontend_config *config;
  int    n;
  unsigned long start;

  /* Get the platform device information (if it exists) */
  stpti_device_data = to_platform_device(dev);
  if (!stpti_device_data || !stpti_device_data->name) 
  {
    printk(KERN_ERR "%s: Device probe failed.  Check your kernel SoC config!!\n",__FUNCTION__);    
    return -ENODEV;
  }

  /* Allocate the stpti structure */
  pti = (struct stpti*)kmalloc(sizeof(struct stpti),GFP_KERNEL);
  if (!pti)
    return -ENOMEM;

  memset((void*)pti, 0, sizeof(struct stpti));
  
  /* Allocate the back buffer */
  pti->back_buffer = (char*)bigphysarea_alloc_pages((PTI_BUFFER_SIZE+PAGE_SIZE-1) / PAGE_SIZE,0,GFP_KERNEL);
  if (!pti->back_buffer)
    goto err1;

  /* ioremap the pti address space */
  start = platform_get_resource(stpti_device_data,IORESOURCE_MEM,0)->start;
  pti->pti_io = ioremap( (unsigned long)start, platform_get_resource(stpti_device_data,IORESOURCE_MEM,0)->end - start);
  if (!pti->pti_io)
    goto err2;

  /* Resolve pointers we need for later */
  pti->pidtable    = getsymbol(pti, "pidtable");
  pti->num_pids    = getsymbol(pti, "num_pids");	
  pti->pidsearch   = getsymbol(pti, "pidsearch");
  
  pti->psize       = getsymbol(pti, "psize");
  pti->pread       = getsymbol(pti, "pread");
  pti->pwrite      = getsymbol(pti, "pwrite");
  pti->packet_size = getsymbol(pti, "packet_size");
  
  pti->discard     = getsymbol(pti, "discard");
  pti->header_size = getsymbol(pti, "header_size");
  pti->thrown_away = getsymbol(pti, "thrown_away");

  /* Setup PTI */
  writel(0x1, pti->pti_io + PTI_DMA_PTI3_PROG);
  
  /* Write instructions and data */
  loadtc(pti);
  
  PtiWrite(pti->packet_size,PACKET_SIZE);
  PtiWrite(pti->header_size,HEADER_SIZE);
  
  /* Enable IIF */
  writel(0x0,         pti->pti_io + PTI_IIF_SYNC_LOCK);
  writel(0x0,         pti->pti_io + PTI_IIF_SYNC_DROP);
  writel(PACKET_SIZE, pti->pti_io + PTI_IIF_SYNC_PERIOD);
  writel(0x0,         pti->pti_io + PTI_IIF_SYNC_CONFIG);
  writel(0x1,         pti->pti_io + PTI_IIF_FIFO_ENABLE);
  
  /* Start the TC */
  writel(PTI_MODE_ENABLE, pti->pti_io + PTI_MODE );
  
  /* Setup packet count */
  PtiWrite(pti->psize, (PTI_BUFFER_SIZE / (PACKET_SIZE)) - 2);
  PtiWrite(pti->num_pids,0);
  
  /* Reset the DMA */
  stpti_setup_dma(pti);
  stpti_reset_dma(pti);
  
  /* Enable the DMA of data */
  writel( readl(pti->pti_io + PTI_DMA_ENABLE) | 0x1, pti->pti_io + PTI_DMA_ENABLE );  
  
  /* Get the configuration information about the tuners */
  config = (struct plat_frontend_config*)stpti_device_data->dev.platform_data;

  /* Setup the transport stream merger based on the configuration */
  stm_tsm_init( config );

  /* Initiliase DiseqC */
  if (config->diseqc_address) stm_diseqc_init( config );

  stm_tuner_register_frontend ( config, pti->stfe , (void*)pti, stpti_start_feed, stpti_stop_feed );

  for (n=0;n<3;n++)
    if (pti->stfe[n]) {
      pti->stfe[n]->tsm_channel = STM_GET_CHANNEL(config->channels[pti->stfe[n]->mapping].option);
      pti->mapping[ pti->stfe[n]->tsm_channel ] = n;
    }


  pti->timer.data     = (unsigned long)pti;
  /* Setup timer interrupt */
  spin_lock_init(&pti->timer_lock);
  pti->timer.function = stpti_timer_interrupt;
  init_timer(&pti->timer);
  pti->timer.function = stpti_timer_interrupt;  
  
  return 0;

 err3:
  iounmap( pti->pti_io );

 err2:
  kfree(pti->back_buffer);

 err1:
  kfree(pti);

  return -ENOMEM;
}

static int stpti_remove(struct device *dev)
{
#warning we need to do the deallocation, however in a STB this may not be too much of an issue
  return 0;
}

static struct device_driver stpti_driver = {
        .name           = "pti",
        .bus            = &platform_bus_type,
        .probe          = stpti_probe,
        .remove         = stpti_remove,
};

static __init int stpti_init(void)
{
  return driver_register(&stpti_driver);
}

static void stpti_cleanup(void)
{
        driver_unregister(&stpti_driver);
}

module_init(stpti_init);
module_exit(stpti_cleanup);

module_param(debug, int,0644);
MODULE_PARM_DESC(debug, "Debug or not");

MODULE_AUTHOR("Peter Bennett <peter.bennett@st.com>");
MODULE_DESCRIPTION("STPTI DVB Driver");
MODULE_LICENSE("GPL");

